package fluids;

/**
 * A <code>Tank</code> represents a container that can fill up with liquid. The amount of liquid can
 * be measured in two ways: the volume, in gallons, and the water level, the y coordinate on the screen
 * of the surface of the liquid. For completeness, it permits the volume to be negative, but negative
 * volume shows up in a different color (red, by default) to show that it is negative.
 * @author Russell Zahniser (russell@zahniser.net)
 */
public class Tank extends FluidContainer {
	private int[] rowSizes;
	private int volume;
	private Pipe overflow;
	
	private static int gallonsToPixels(double gallons) {
		return (int)(gallons * 2560000 + .5);
	}
	
	private static double pixelsToGallons(int pixels) {
		return pixels / 2560000.;
	}
	
	/**
	 * Create a new <code>Tank</code>. The dimensions of the tank will be set to reflect the area of
	 * the image where this tank's water may appear, and contains() will return true only for points
	 * within the liquid area of this tank.
	 * @param name Name of .png file to load from the model
	 * @param background <code>Background</code> which knows which model to load the image file from, 
	 *        and on which the water in this <code>Tank</code> will be drawn.
	 */
	public Tank(String name, Background background) {
		super(name, background);
		
		try {
			overflow = new Pipe(name + "Overflow", this, background.getTank(), background);
		} catch(IllegalArgumentException e) {
		}
		
		int sum = 0, width = getWidth(), height = getHeight();
		rowSizes = new int[height + 1];
		for(int y = height - 1; y >= 0; y--) {
			int i = y * width;
			for(int x = width; x > 0; x--, i++) {
				sum += mask[i] & 0xFF;
			}
			rowSizes[y] = sum;
		}
	}
	
	/**
	 * Set the volume of liquid in this <code>Tank</code>. If the volume is greater than the capacity, 
	 * or less than the negative of the capacity, the tank may overflow into the background.
	 * @param v new volume, in gallons
	 */
	public void setVolume(double v) {
		volume = gallonsToPixels(v);
		changed();
	}
	
	/**
	 * Get the volume of liquid present in this tank.
	 * @return The volume in this tank, in gallons
	 */
	public double getVolume() {
		return pixelsToGallons(volume);
	}
	
	/**
	 * Get the maximum volume that this tank can hold
	 * @return the maximum volume this tank can hold, in gallons
	 */
	public double getCapacity() {
		return pixelsToGallons(rowSizes[0]);
	}
	
	/**
	 * Get the water level in this tank. This will be the y coordinate on the screen of the water
	 * surface in the tank, and should be between <code>tank.getY()</code> (where volume = capacity)
	 * and <code>tank.getY() + tank.getHeight()</code> (where volume = 0), assuming the volume is 
	 * positive. If the volume is negative, a water level below the bottom of the tank will be
	 * returned, calculated as if the tank were mirrored over its bottom edge.
	 * @return the y coordinate of this tank's water surface
	 */
	public double getWaterLevel() {
		int pixels = Math.abs(volume);
		
		if(pixels >= rowSizes[0]) return getY();
		
		int min = 0, max = getHeight();
		while(min + 1 < max) {
			int mid = (min + max) / 2;
			if(pixels >= rowSizes[mid]) {
				max = mid;
			} else {
				min = mid;
			}
		}
		double y = getY() + min - (pixels - rowSizes[min + 1]) / (double)(rowSizes[min] - rowSizes[min + 1]);
		if(volume < 0) y = 2 * (getY() + getHeight()) - y;
		return y;
	}
	
	/**
	 * Set the water level of this tank. If the level provided is above the top of the tank, the tank will
	 * be set to be completely full, but will not overflow. If the water level is below the bottom of the
	 * tank, the tank will be given a negative volume.
	 * @param level New water level
	 */
	public void setWaterLevel(double level) {
		int sign = 1;
		int bottom = getY() + getHeight();
		if(level == bottom) {
			volume = 0;
			return;
		} else if(level > bottom) {
			level = 2 * bottom - level;
			sign = -1;
		}
		
		level -= getY();
		if(level < 0) {
			volume = sign * rowSizes[0];
		} else {
			double fpart = level - (int)level;
			volume = sign * ((int)(fpart * rowSizes[(int)level + 1] + (1 - fpart) * rowSizes[(int)level] + .5));
		}
		changed();
	}
	
	void draw(int[] pixels, int w, int h) {
		if(volume == 0 || isHidden()) return;
		double level;
		if(volume < 0) level = 2 * getHeight() - (getWaterLevel() - getY());
		else level = getWaterLevel() - getY();
		
		int rgb = positiveColor;
		if(volume < 0) rgb = negativeColor;
		
		int width = getWidth(), height = getHeight(), minX = getX(), minY = getY();

		int y = height - 1;
		// Fill in everywhere under the water level
		while(y >= level) {
			int d = (y + minY) * w + minX, i = y * width;
			for(int x = width; x > 0; x--, d++, i++) {
				pixels[d] = (pixels[d] & ~mask[i]) | (mask[i] & rgb);
			}
			y--;
		}
		if(y >= 0) {
			// Top row is antialiased
			int m = (int)(0xFF * ((int)(level + 1) - level)) * 0x010101;
			int d = (y + minY) * w + minX, i = y * width;
			for(int x = width; x > 0; x--, d++, i++) {
				pixels[d] = (pixels[d] & ~(mask[i] & m)) | (mask[i] & m & rgb);
			}
		}
		
		if(overflow != null) {
			if(volume > rowSizes[0]) {
				overflow.setFlowRate(volume - rowSizes[0]);
				if(overflow.getOutlet() != null) {
					overflow.getOutlet().volume += volume - rowSizes[0];
				}
				volume = rowSizes[0];
			} else if(volume < -rowSizes[0]) {
				overflow.setFlowRate(volume + rowSizes[0]);
				if(overflow.getOutlet() != null) {
					overflow.getOutlet().volume += volume + rowSizes[0];
				}
				volume = -rowSizes[0];
			} else {
				overflow.setFlowRate(0);
			}
		}
	}
}
