package legos;
import java.awt.Color;
import java.awt.Shape;
import java.awt.geom.AffineTransform;
import java.awt.geom.Arc2D;
import java.awt.geom.Ellipse2D;
import java.awt.geom.GeneralPath;
import java.awt.geom.Rectangle2D;

import objectdraw.DrawingCanvas;

/**
 * A <code>Legs</code> is a <code>LegoPiece</code> representing the legs of a lego person. It
 * has two <code>Leg</code> subunits.
 * 
 * @author Russell Zahniser (russell@zahniser.net)
*/
public class Legs extends LegoPiece {
	static {
		LegoPiece.registerShape(Legs.class, Color.BLUE, buildOutline(), buildFill(), new Attachment[] {
			new Attachment(-13, 28, true, Leg.class),
			new Attachment(13, 28, false, Leg.class),
		});
	}
	
	private static Shape buildFill() {
		GeneralPath path = new GeneralPath();
		path.moveTo(0, 20);
		path.lineTo(8, 20);
		path.lineTo(11, 0);
		path.lineTo(21, 0);
		path.lineTo(24, 20);
		path.lineTo(34, 20);
		path.lineTo(37, 0);
		path.lineTo(47, 0);
		path.lineTo(50, 20);
		path.lineTo(58, 20);
		path.lineTo(58, 28);
		path.append(new Arc2D.Double(26, 28, 12, 40, 90, -150, Arc2D.OPEN), true);
		path.lineTo(20, 48 + 10 * (float)Math.sqrt(3));
		path.lineTo(20, 53);
		path.append(new Arc2D.Double(15, 43, 3, 10, 270, -180, Arc2D.OPEN), true);
		path.lineTo(20, 43);
		path.lineTo(20, 28);
		path.lineTo(0, 28);
		path.closePath();
		return AffineTransform.getTranslateInstance(-29, -20).createTransformedShape(path);
	}
	
	private static Shape buildOutline() {
		GeneralPath path = new GeneralPath();
		
		path.append(new Arc2D.Double(26, 28, 12, 40, 90, -150, Arc2D.OPEN), false);
		path.append(new Arc2D.Double(20, 28, 12, 40, 300, 150, Arc2D.OPEN), true);
		path.closePath();
		path.append(new Rectangle2D.Double(0, 20, 58, 8), false);
		path.moveTo(12, 20);
		path.lineTo(12, 28);
		path.moveTo(29, 48 + 10 * (float)Math.sqrt(3));
		path.lineTo(20, 48 + 10 * (float)Math.sqrt(3));
		path.lineTo(20, 53);
		path.moveTo(20, 43);
		path.lineTo(20, 28);
		
		path.append(new Ellipse2D.Double(15, 43, 3, 10), false);
		path.moveTo(16.5f, 43);
		path.append(new Arc2D.Double(22, 43, 3, 10, 90, -180, Arc2D.OPEN), true);
		path.lineTo(16.5f, 53);
		
		// Two posts centered around x of 16 and 42, respectively
		path.moveTo(8, 20);
		path.lineTo(11, 0);
		path.lineTo(21, 0);
		path.lineTo(24, 20);
		
		path.moveTo(34, 20);
		path.lineTo(37, 0);
		path.lineTo(47, 0);
		path.lineTo(50, 20);
		return AffineTransform.getTranslateInstance(-29, -20).createTransformedShape(path);
	}
	
	/**
	 * Get the right <code>Leg</code>.
	 * @return the right <code>Leg</code>.
	 */
	public Leg getRightLeg() {
		return (Leg)getSubunit(0);
	}
	
	/**
	 * Get the left <code>Leg</code>.
	 * @return the left <code>Leg</code>.
	 */
	public Leg getLeftLeg() {
		return (Leg)getSubunit(1);
	}
	
	/**
	 * Replace the right <code>Leg</code> with a different <code>Leg</code>.
	 * @param leg new <code>Leg</code> to replace the old one with.
	 */
	public void setRightLeg(Leg leg) {
		setSubunit(0, leg);
	}
	
	/**
	 * Replace the left <code>Leg</code> with a different <code>Leg</code>.
	 * @param leg new <code>Leg</code> to replace the old one with.
	 */
	public void setLeftLeg(Leg leg) {
		setSubunit(1, leg);
	}
	
	/**
	 * Construct just the waist of a pair of legs, placing it on the given canvas. It will be
	 * colored blue.
	 * @param canvas Canvas to place this pair of legs on.
	 */
	public Legs(DrawingCanvas canvas) {
		this(null, null, null, canvas);
	}
	
	/**
	 * Construct just the waist of a pair of legs, placing it on the given canvas.
	 * @param color The color of the waist.
	 * @param canvas Canvas to place this pair of legs on.
	 */
	public Legs(Color color, DrawingCanvas canvas) {
		this(null, null, color, canvas);
	}
	
	/**
	 * Construct a pair of legs, placing them on the given canvas. The waist will be blue.
	 * @param rightLeg The right <code>Leg</code>.
	 * @param leftLeg The left <code>Leg</code>.
	 * @param canvas Canvas to place this pair of legs on.
	 */
	public Legs(Leg rightLeg, Leg leftLeg, DrawingCanvas canvas) {
		this(rightLeg, leftLeg, null, canvas);
	}
	
	/**
	 * Construct a pair of legs, placing them on the given canvas.
	 * @param rightLeg The right <code>Leg</code>.
	 * @param leftLeg The left <code>Leg</code>.
	 * @param color The color of the waist.
	 * @param canvas Canvas to place this pair of legs on.
	 */
	public Legs(Leg rightLeg, Leg leftLeg, Color color, DrawingCanvas canvas) {
		super(new LegoPiece[] { rightLeg, leftLeg }, color, canvas);
	}
	
	/**
	 * Construct just the waist of a pair of legs, but do not make it appear on any canvas
	 * yet. It will be colored blue. You can make it appear by adding it to a canvas, or 
	 * setting it as the legs of a <code>Body</code> that is already visible on a canvas.
	 */
	public Legs() {
		this(null, null, null, null);
	}
	
	/**
	 * Construct just the waist of a pair of legs, but do not make it appear on any canvas
	 * yet. You can make it appear by adding it to a canvas, or 
	 * setting it as the legs of a <code>Body</code> that is already visible on a canvas.
	 * @param color The color of the waist.
	 */
	public Legs(Color color) {
		this(null, null, color, null);
	}
	
	/**
	 * Construct a pair of legs, but do not make them appear on any canvas
	 * yet. The waist will be colored blue. You can make it appear by adding it to a canvas, or 
	 * setting it as the legs of a <code>Body</code> that is already visible on a canvas.
	 * @param rightLeg The right <code>Leg</code>.
	 * @param leftLeg The left <code>Leg</code>.
	 */
	public Legs(Leg rightLeg, Leg leftLeg) {
		this(rightLeg, leftLeg, null, null);
	}
	
	/**
	 * Construct a pair of legs, but do not make them appear on any canvas
	 * yet. You can make it appear by adding it to a canvas, or 
	 * setting it as the legs of a <code>Body</code> that is already visible on a canvas.
	 * @param rightLeg The right <code>Leg</code>.
	 * @param leftLeg The left <code>Leg</code>.
	 * @param color The color of the waist.
	 */
	public Legs(Leg rightLeg, Leg leftLeg, Color color) {
		this(rightLeg, leftLeg, color, null);
	}
}
