package fluids;

import java.awt.Component;
import java.awt.Image;
import java.awt.MediaTracker;
import java.awt.Toolkit;
import java.awt.image.BufferedImage;
import java.awt.image.ColorModel;
import java.awt.image.ImageObserver;
import java.awt.image.MemoryImageSource;
import java.awt.image.PixelGrabber;
import java.io.BufferedInputStream;
import java.net.URL;
import java.util.ArrayList;

import objectdraw.*;

/**
 * The <code>Background</code> class draws the background for a fluid mechanics simulation. All
 * other components of the simulation need to be added to this background in order to be drawn.
 * @author Russell Zahniser (russell@zahniser.net)
 */
public class Background extends VisibleImage {
	private String modelName;
	private BufferedImage backgroundImage;
	private int[] originalPixels;
	private int[] pixels;
	private MemoryImageSource source;
	private ArrayList containers = new ArrayList();
	private boolean needsRedraw = true;
	private Tank tank;
	private DrawingCanvas originalCanvas;

	private static Image loadImage(String fileName, DrawingCanvas canvas) {
		final URL url = Background.class.getResource(fileName);
		if(url == null) return null;
		
		final Image image = Toolkit.getDefaultToolkit().createImage(url);
		if(image == null) return null;
		
		final MediaTracker tracker = new MediaTracker((Component)canvas);

		tracker.addImage(image, 1);
		for(boolean loaded = false; !loaded; ) {
			try {
				tracker.waitForAll();
				loaded = true;
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}

		return image;
	}

	Image loadImage(String imageName, FluidContainer container) {
		containers.add(container);
		needsRedraw = true;
		return loadImage(modelName + "/" + imageName + ".png", originalCanvas);
	}

	void removeContainer(FluidContainer container) {
		containers.remove(container);
	}

	/**
	 * Get the tank that represents the background. This is where water flows into when it overflows from
	 * another tank. This method may return <code>null</code>, if there was no "backgroundWater.png" in 
	 * the directory for this model.
	 * @return The <code>Tank</code> into which other tanks overflow, or <code>null</code> if there is no
	 *         such tank.
	 */
	public Tank getTank() {
		return tank;
	}

	private static DrawingCanvas checkCanvas(DrawingCanvas canvas) {
		if(canvas == null) throw new IllegalArgumentException("Background cannot be created on a null canvas.");
		return canvas;
	}

	/**
	 * Create a new <code>Background</code>.
	 * @param modelName Name of the model. This should be the name of the directory in which
	 *        the images for this simulation are. There must be a file <code>background.png</code>
	 *        in that directory.
	 * @param canvas <code>DrawingCanvas</code> to add this <code>Background</code> to
	 */
	public Background(String modelName, DrawingCanvas canvas) {
		super(loadImage(modelName + "/background.png", checkCanvas(canvas)), 0, 0, canvas);

		this.modelName = modelName;
		this.originalCanvas = canvas;
		int w = getImageWidth(), h = getImageHeight(), len = w * h;
		if(w < 0 || h < 0) throw new IllegalArgumentException("Unable to load image \"" + modelName + "/background.png\".");

		try {
			tank = new Tank("backgroundWater", this);
		} catch(IllegalArgumentException e) {
		}

		originalPixels = new int[len];
		pixels = new int[len];
		try {
			new PixelGrabber(getImage(), 0, 0, w, h, originalPixels, 0, w).grabPixels();
		} catch (InterruptedException e) {
		}
		for(int i = 0; i < len; i++) {
			pixels[i] = originalPixels[i];
		}
		ColorModel rgbModel = new BufferedImage(1, 1, BufferedImage.TYPE_INT_RGB).getColorModel();
		source = new MemoryImageSource(w, h, rgbModel, pixels, 0, w);
		source.setAnimated(true);
		setImage(((Component)canvas).createImage(source));

		new EventTimer(this, "redraw").start(.05, .05);
	}

	void changed() {
		needsRedraw = true;
	}

	/**
	 * @deprecated for internal use only
	 */
	public void redraw() {
		if(!needsRedraw) return;

		needsRedraw = false;
		int w = getImageWidth(), h = getImageHeight(), len = w * h;
		for(int i = 0; i < len; i++) {
			pixels[i] = originalPixels[i];
		}
		for(int i = 0; i < containers.size(); i++) {
			FluidContainer p = (FluidContainer)containers.get(i);
			p.draw(pixels, w, h);
		}
		source.newPixels();
		getCanvas().repaint();
	}
}
