package fluids;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.image.PixelGrabber;

import objectdraw.*;

/**
 * <code>FluidContainer</code> is the shared superclass of both <code>Tank</code>and <code>Pipe</code>.
 * It contains method to set the color that that container will appear in if its volume or flow rate
 * is negative.
 * @author Russell Zahniser (russell@zahniser.net)
 *
 */
public abstract class FluidContainer extends FramedRect {
	protected int[] mask;
	protected int negativeColor, positiveColor;
	private Background background;
	
	FluidContainer(String name, Background background) {
		super(0, 0, 1, 1, background.getCanvas());
		this.background = background;
		
		int w = background.getImageWidth(), h = background.getImageHeight();
		Image image = background.loadImage(name, this);
		if(image == null) throw new IllegalArgumentException("Unable to load image \"" + name + "\".");
		int[] pixels = new int[w * h];
		try {
			new PixelGrabber(image, 0, 0, w, h, pixels, 0, w).grabPixels();
		} catch (InterruptedException e) {
		}
		
		// Find the bounds of our part of the image
		int minX = w, minY = h, maxX = 0, maxY = 0;
		for(int y = 0; y < h; y++) {
			for(int x = 0; x < w; x++) {
				if((pixels[x + y * w] & 0xFF000000) != 0) {
					if(x < minX) minX = x;
					if(x > maxX) maxX = x;
					if(y < minY) minY = y;
					if(y > maxY) maxY = y;
				}
			}
		}
		if(minX > maxX || minY > maxY) throw new IllegalArgumentException("Unable to load image \"" + name + "\".");
		
		int width = maxX - minX + 1, height = maxY - minY + 1;
		moveTo(minX, minY);
		setSize(width, height);
		
		mask = new int[getWidth() * getHeight()];
		for(int y = height - 1; y >= 0; y--) {
			int i = y * width, d = (y + minY) * w + minX;
			for(int x = width; x > 0; x--, i++, d++) {
				int t = ((pixels[d] & 0xFF000000) >>> 24);
				mask[i] = t * 0x010101;
			}
		}
		setColor(Color.BLUE);
		negativeColor = 0xFF0000;
	}
	
	/**
	 * Removing a fluid container from its canvas also removes it from the background.
	 */
	public void removeFromCanvas() {
		background.removeContainer(this);
		super.removeFromCanvas();
	}
	
	/**
	 * Return the <code>Background</code> object that this fluid container is placed on.
	 * @return the <code>Background</code>
	 */
	public Background getBackground() {
		return background;
	}
	
	/**
	 * @deprecated
	 */
	public void draw(Graphics2D g) {
	}
	
	/**
	 * Set the color that will be used to draw this container when its volume or flow rate is positive.
	 * @param c <code>Color</code> to draw with
	 */
	public void setColor(Color c) {
		positiveColor = (c.getRGB() & 0xFFFFFF);
		super.setColor(c);
		changed();
	}
	
	/**
	 * Set the color that will be used to draw this container when its volume or flow rate is negative.
	 * @param c <code>Color</code> to draw with
	 */
	public void setNegativeColor(Color c) {
		negativeColor = (c.getRGB() & 0xFFFFFF);
		changed();
	}
	
	/**
	 * Get the <code>Color</code> that will be used to draw this container when its volume 
	 * or flow rate is negative.
	 * @return the color negative flow is drawn with
	 */
	public Color getNegativeColor() {
		return new Color(negativeColor | 0xFF000000);
	}
	
	/**
	 * Determine whether a point is one of the pixels painted by this fluid container.
	 * @param point <code>Location</code> to check
	 */
	public boolean contains(Location point) {
		int x = (int)(point.getX() - getX() + .5);
		int y = (int)(point.getY() - getY() + .5);
		if(x < 0 || x >= getWidth() || y < 0 || y >= getHeight()) return false;
		return mask[x + y * getWidth()] != 0;
	}
	
	void changed() {
		background.changed();
	}
	
	abstract void draw(int[] pixels, int w, int h);
}
