package objectdraw;

import java.awt.Color;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.Shape;
import java.awt.geom.AffineTransform;
import java.awt.geom.GeneralPath;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;

/**
 * <code>Animation</code> is similar to <code>VisibleImage</code>, but rather 
 * than simply displaying an unchanging image, it displays an animation. After
 * playing through the animation, which typically takes only a few seconds, it 
 * automatically removes itself from the canvas.
 * <br><br>
 * Displaying an animation is a simple as creating an <code>Animation</code> object,
 * then telling it to <code>start</code>. WHen creating the object, you can choose
 * one of the constants in this class to specify the type of animation, and you
 * can also specify an object to center the animation around. So, for example, if
 * <code>player</code> is a <code>VisibleImage</code> representing the player,
 * this code would display a shower of gold sparks cascading down in front of him:
 * <br><br><code>
 * new Animation(Animation.SPARKS, player).start();
 * </code><br><br>
 * It is also possible to have the animation display text labels that appear over
 * an object and move upwards and fade out as the animation runs. These labels
 * should be set before starting the animation. So, for example, this code would 
 * make "Heal!" appear over the player as the "sparks" animation is running:
 * <br><br><code>
 * Animation animation = new Animation(Animation.SPARKS, player);
 * <br>animation.label(player, "Heal!");
 * <br>animation.start();
 * </code><br><br>
 * You can add as many labels as you wish to an animation in this same way.
 * <br><br>
 * Currently, <code>Animation</code> only contains animation types I produced while
 * designing an <a href="http://www.zahniser.net/~russell/computer/index.php?title=RPG">RPG</a>
 * project for an introductory Java course. If you have ideas for animation you 
 * would like to see added, send a detailed description to my email address below.
 * 
 * @author Russell Zahniser (russell@zahniser.net)
 */
public class Animation extends ObjectDrawShape {
	private final ArrayList labels = new ArrayList();
	private final ArrayList specialRegions = new ArrayList();
	private final AnimationType type;
	private final Color[] labelColors = new Color[10];
	private Particle[] particles;
	private final ObjectDrawShape basis;
	private int count = 1;
	private int frame;
	private boolean cancelled = false;
	private boolean firstLabelRender = true;
	// Offset from basis location to my location
	private final int dx, dy;
	private final DrawingCanvas oldCanvas;
	private boolean attached = defaultAttached;
	
	private static final List animations = Collections.synchronizedList(new ArrayList());
	private static final EventTimer timer = new EventTimer(new AnimationRunner(), "step");
	private static final Font LABEL_FONT = Text.DEFAULT_FONT.deriveFont(Font.BOLD, 24f);
	
	private static final Image[] blood;
	private static final Image[] smoke;
	private static final Image[] steam;
	private static final Image[] steam2;
	private static final Image[] explosion;
	private static final Image[] soot;
	private static final Image[] lightning;
	private static final Image[] spark;
	private static final Image[] green;
	private static final Image[] darkGreen;
	private static final Image[] limeGreen;
	private static final Image[] white;
	private static final Image[] purple;
	private static final Image[] rightArc;
	private static final Image[] leftArc;
	private static final Image[] pinkRightArc;
	private static final Image[] hailstone;
	private static final Image[] asteroid;
	
	private static Color defaultLabelColor = DEFAULT_COLOR;
	private static boolean defaultAttached = true;
	
	/**
	 * Set the default color for animation labels to appear in. You can set the
	 * color used by a particular Animation using its setColor(), but this method
	 * is convenient if you want all your animations to have the same color and
	 * that color isn't black.
	 * @param color Default color for labels
	 */
	public static void setDefaultLabelColor(final Color color) {
		defaultLabelColor = color;
	}
	
	/**
	 * Set whether animations are, by default, attached to the object they decorate.
	 * If true, moving that object will cause the animation to move with it.
	 * 
	 * @param attached Whether animations are attached to their target by default
	 */
	public static void setAttachedByDefault(final boolean attached) {
		defaultAttached = attached;
	}
	
	/**
	 * Animation type that surrounds the player with a cloud of orbiting meteors,
	 * which can be fired one by one at other targets. By default, six meteors
	 * will be shown; you can change this number by calling <code>setCount()</code>
	 * at any time <b>before</b> you <code>start()</code> the animation.
	 * 
	 * Calling <code>fireOne()</code> on an animation of this type will separate
	 * out a single meteor into a new animation, which you can then <code>start</code>
	 * as you would any other animation. This will also decrease the count by one.
	 * So, you can tell when all the meteors have been fired, because 
	 * <code>getCount()</code> will return 0.
	 */
	public static final AnimationType METEORS = new MeteorAnimationType();
	/**
	 * Animation type that shows a single meteor moving from one object to strike
	 * another. The source object is the one passed to the constructor of
	 * <code>Asteroid</code>; the destination object should be set with 
	 * addSpecialRegion(). This is also the animation type used for the meteors 
	 * fired by a <code>METEORS</code> animation.
	 */
	public static final AnimationType METEOR = new SingleMeteorAnimationType();
	/**
	 * Animation type that shows a sword slash, moving from left to right, which 
	 * leaves behind it a trail of blood that then drips down to the floor.
	 */
	public static final AnimationType SLASH_RIGHT = new BloodAnimationType(1);
	/**
	 * Animation type that shows a sword slash, moving from right to left, which 
	 * leaves behind it a trail of blood that then drips down to the floor.
	 */
	public static final AnimationType SLASH_LEFT = new BloodAnimationType(-1);
	/**
	 * Animation type that shows a cloud of smoke erupting around its target.
	 */
	public static final AnimationType SMOKE_BOMB = new SimpleAnimationType(40) {
		Particle[] createParticles(Animation animation) {
			Particle[] particles = new Particle[100];
			
			for(int i = 0; i < particles.length; ) {
				double x = -25 + Math.random() * 50;
				double y = -35 + Math.random() * 50;
				if(x * x + y * y > 625) continue;
				
				particles[i++] = new Particle(x, y, 
						2.5 - 5 * Math.random(), 1 - 7 * Math.random(),
						(int)(Math.random() * 20), smoke[0]);
			}
			
			return particles;
		}

		void update(Particle particle, int frame) {
			if(frame >= 40 || particle.y > 30) particle.image = null;
			else {
				particle.vx = .8 * particle.vx + 2 * Math.random() - 1;
				particle.vy = .8 * particle.vy + 2 * Math.random() - .75;
				if(frame > 30) particle.image = smoke[frame - 30];
			}
		}
	};
	/**
	 * Animation type that halos the target in green spots, which then shoot off
	 * upward.
	 */
	public static final AnimationType GLOW = new SimpleAnimationType(60) {
		Location getLocation(DrawableInterface object) {
			return getOrigin(object);
		}
		
		Particle[] createParticles(Animation animation) {
			final Particle[] particles = outline(animation.basis, 1.2, green[0]);
			for(int i = 0; i < particles.length; i++) {
				final Particle p = particles[i];
				p.t = (int)(Math.random() * 30);
				p.vx = p.vy = 0;
			}
			return particles;
		}

		void update(Particle particle, int frame) {
			frame -= particle.t;
			
			if(frame < 0 || frame >= 30) {
				particle.image = null;
				return;
			}
			
			else if(frame < 20) {
				particle.image = green[frame < 9 ? 9 - frame : 0];
			} else if(frame == 20) {
				particle.vx = Math.random() - .5;
				particle.vy = -4 - 2 * Math.random();
				particle.image = green[0];
			} else {
				particle.vx = .8 * particle.vx + .5 * Math.random() - .25;
				particle.vy = .8 * particle.vy + .5 * Math.random() - .25;
				particle.image = green[frame - 20];
			}
		}
	};
	/**
	 * Animation type that outlines the target in dark clouds of dust
	 */
	public static final AnimationType DUSTY = new SimpleAnimationType(90) {
		Location getLocation(DrawableInterface object) {
			return getOrigin(object);
		}
		
		Particle[] createParticles(Animation animation) {
			final Particle[] particles = outline(animation.basis, .75, soot[0]);
			for(int i = 0; i < particles.length; i++) {
				final Particle p = particles[i];
				p.t = (int)(Math.random() * 20 + Math.random() * 20 + Math.random() * 20);
				p.vx *= .3;
				p.vy *= .3;
			}
			return particles;
		}

		void update(Particle particle, int frame) {
			frame -= particle.t;
			
			if(frame < 0 || frame >= 30) {
				particle.image = null;
				return;
			} else {
				particle.image = soot[frame / 3];
			}
		}
	};
	/**
	 * Animation type that outlines the target with bright yellow sparks.
	 */
	public static final AnimationType SPARKLY = new SimpleAnimationType(90) {
		Location getLocation(DrawableInterface object) {
			return getOrigin(object);
		}
		
		Particle[] createParticles(Animation animation) {
			final Particle[] particles = outline(animation.basis, .75, spark[0]);
			for(int i = 0; i < particles.length; i++) {
				final Particle p = particles[i];
				p.t = (int)(Math.random() * 20 + Math.random() * 20 + Math.random() * 20);
				p.vx *= .3;
				p.vy *= .3;
			}
			return particles;
		}

		void update(Particle particle, int frame) {
			frame -= particle.t;
			
			if(frame < 0 || frame >= 30) {
				particle.image = null;
				return;
			} else {
				particle.image = spark[frame / 3];
			}
		}
	};
	/**
	 * Animation type that covers the target in white layers of frost.
	 */
	public static final AnimationType FROSTY = new SimpleAnimationType(90) {
		Location getLocation(DrawableInterface object) {
			return getOrigin(object);
		}
		
		Particle[] createParticles(Animation animation) {
			final Particle[] particles = outline(animation.basis, .75, white[0]);
			for(int i = 0; i < particles.length; i++) {
				final Particle p = particles[i];
				p.t = (int)(Math.random() * 20 + Math.random() * 20 + Math.random() * 20);
				p.vx *= .3;
				p.vy *= .3;
			}
			return particles;
		}

		void update(Particle particle, int frame) {
			frame -= particle.t;
			
			if(frame < 0 || frame >= 30) {
				particle.image = null;
				return;
			} else {
				particle.image = white[frame / 3];
			}
		}
	};
	
	/**
	 * Animation type that shows large hailstones falling around its target. The
	 * hailstones, upon hitting the ground, shatter into smaller pieces. It is
	 * also possible for hailstones to turn into steam when they hit the ground;
	 * to accomplish this, call <code>addSpecialRegion()</code> for each object
	 * within whose bounds steam ought to be produced.
	 */
	public static final AnimationType ICE_STORM = new SimpleAnimationType(70) {
		Particle[] createParticles(Animation animation) {
			Particle[] particles = new Particle[600];
			
			for(int i = 0; i < particles.length; ) {
				double angle = Math.random() * Math.PI * 2;
				double r = .5 + Math.random();
				double target_x = 50 * Math.cos(angle) * r;
				double target_y = 60 * Math.sin(angle) * r;
				double vx = 8 * Math.random() - 4;
				double vy = 8 + 4 * Math.random();
				int t = (int)(Math.random() * 10) * 4;
				double x = target_x - 20 * vx;
				double y = target_y - 20 * vy;
				
				particles[i++] = new Particle(x, y, vx, vy, t, hailstone[0]);
				if(animation.isSpecial(target_x, target_y)) {
					for(int e = 0; e < 9; e++) {
						double rad = Math.random() * Math.PI * 2;
						double speed = Math.random() * 6 + 1;
						particles[i++] = new Particle(target_x, target_y, speed * Math.sin(rad), speed * Math.cos(rad), t + 22 + (int)Math.random() * 2, steam[0]);
					}
				} else {
					for(int e = 0; e < 9; e++) {
						particles[i++] = new Particle(target_x, target_y, Math.random() * 6 - 3, -5 - 2 * Math.random(), t + 21, white[0]);
					}
				}
			}
			
			return particles;
		}

		void update(Particle particle, int frame) {
			frame = frame - (particle.t & ~3);
			if(frame < 0 || frame >= 20) {
				particle.image = null;
				return;
			}
			
			switch(particle.t & 3) {
			case 1:
				if(frame >= 10) {
					particle.vy = 0;
					particle.vx *= .85;
				} else {
					particle.vy += 1;
				}
				
				particle.image = frame > 10 ? white[frame - 10] : white[0];
				return;
			case 2: case 3:
				particle.vx = .8 * particle.vx + Math.random() * 1 - .5;
				particle.vy = .8 * particle.vy + Math.random() * 1 - .5;
				
				particle.image = ((particle.t & 3) == 2 ? steam : steam2)[frame  / 2];
				return;
			default:
				if(frame < 0 || frame >= 20) particle.image = null;
				else particle.image = frame < 9 ? hailstone[9 - frame] : hailstone[0];
				return;
			}
		}
	};
	/**
	 * Animation type that shows gold glittering in a open treasure chest, then
	 * gradually fading away. This animation is designed to work only with one
	 * particular open treasure chest image.
	 */
	public static final AnimationType TREASURE = new SimpleAnimationType(50) {
		Particle[] createParticles(Animation animation) {
			Particle[] particles = new Particle[100];
			
			for(int i = 0; i < particles.length; ) {
				double x = Math.random() * 22;
				double y = Math.random() * 10;
				
				particles[i++] = new Particle(-10 + x - .5 * y, -8 + y + .35 * x, (int)(Math.random() * 50), spark[0]);
			}
			
			return particles;
		}

		void update(Particle particle, int frame) {
			int dt = particle.t - frame;
			if(dt < 0) {
				particle.image = null;
			} else {
				particle.image = dt < 9 ? spark[9 - dt] : spark[0];
			}
		}
	};
	/**
	 * Animation type that shows a rotating sphere of green particles forming
	 * around its target.
	 */
	public static final AnimationType SHIELD = new SimpleAnimationType(70, -5) {
		Particle[] createParticles(Animation animation) {
			Particle[] particles = new Particle[200];
			
			for(int i = 0; i < particles.length; ) {
				double phi = (Math.random() - .5) * Math.PI;
				double circ = Math.random();
				double cos = Math.cos(phi);
				if(circ > cos) continue;
				
				double y = 25 * Math.sin(phi);
				double x = 25 * cos * Math.sin(circ * Math.PI * 2 / cos);
				
				particles[i++] = new Particle(x, y, (int)(Math.random() * 40), darkGreen[0]);
			}
			
			return particles;
		}

		void update(Particle particle, int frame) {
			frame -= particle.t;
			if(frame <= 0 || frame >= 30) {
				particle.image = null;
				return;
			}
			
			double sin = (particle.y) / 25.;
			double cos = Math.sqrt(1 - sin * sin);
			double dCirc = 2 * cos * Math.cos(Math.PI * particle.x / (50 * cos));
			
			particle.x += dCirc;
			
			if(frame <= 10) {
				particle.image = darkGreen[10 - frame];
			} else if(frame >= 20) {
				particle.image = darkGreen[frame - 20];
			}
		}
	};
	/**
	 * Animation type that shows green particles collecting around the center of
	 * the target and then bursting outward.
	 */
	public static final AnimationType BURST = new SimpleAnimationType(60, -5) {
		Particle[] createParticles(Animation animation) {
			Particle[] particles = new Particle[100];
			
			for(int i = 0; i < particles.length; ) {
				double phi = (Math.random() - .5) * Math.PI;
				double circ = Math.random();
				double cos = Math.cos(phi);
				if(circ > cos) continue;
				
				double y = 20 * Math.sin(phi);
				double x = 20 * cos * Math.sin(circ * Math.PI * 2 / cos);
				
				particles[i++] = new Particle(x, y, (int)(Math.random() * 30), limeGreen[0]);
			}
			
			return particles;
		}

		void update(Particle particle, int frame) {
			frame -= particle.t;
			if(frame <= 0 || frame >= 30) {
				particle.image = null;
				return;
			}
			
			if(frame <= 10) {
				particle.image = limeGreen[10 - frame];
			} else if(frame > 15) {
				particle.vx = .05 * particle.x;
				particle.vy = .05 * particle.y;
				particle.image = limeGreen[frame <= 20 ? 0 : frame - 20];
			}
		}
	};
	/**
	 * Animation type that shows sparks rising up from the base of the target,
	 * as if it had stepped on hot coals.
	 */
	public static final AnimationType SPARKS = new SimpleAnimationType(40) {
		Particle[] createParticles(Animation animation) {
			Particle[] particles = new Particle[50];
			
			for(int i = 0; i < particles.length; ) {
				double x = -20 + Math.random() * 40;
				double y = 25 + Math.random() * 10;
				
				particles[i++] = new Particle(x, y, (int)(Math.random() * 20), spark[0]);
			}
			
			return particles;
		}

		void update(Particle particle, int frame) {
			frame -= particle.t;
			if(frame <= 0 || frame >= 20) {
				particle.image = null;
				return;
			}
			
			particle.vx = .8 * particle.vx + 2 * Math.random() - 1;
			particle.vy = .8 * particle.vy - 3 * Math.random() + 1;
			
			particle.image = spark[frame <= 10 ? 0 : frame - 10];
		}
	};
	
	/**
	 * Animation type that shows a shower of golden sparks falling in front of
	 * the target.
	 */
	public static final AnimationType GOLD_SHOWER = new SimpleAnimationType(60) {
		Particle[] createParticles(Animation animation) {
			Particle[] particles = new Particle[100];
			
			for(int i = 0; i < particles.length; ) {
				double x = -25 + Math.random() * 50;
				double y = -45 + Math.random() * 50;
				
				particles[i++] = new Particle(x, y, (int)(Math.random() * 30), spark[0]);
			}
			
			return particles;
		}

		void update(Particle particle, int frame) {
			frame -= particle.t;
			if(frame <= 0 || frame >= 30) {
				particle.image = null;
				return;
			}
			
			particle.vx = .8 * particle.vx + .5 * Math.random() - .25;
			particle.vy = .8 * particle.vy + .2;
			
			if(frame <= 10) {
				particle.image = spark[10 - frame];
			} else if(frame > 20) {
				particle.image = spark[frame - 20];
			}
		}
	};
	/**
	 * Animation type that shows a vertical line of blue arcs forming at the
	 * middle of the target and then splitting out to the two sides.
	 */
	public static final AnimationType SPLIT = new SimpleAnimationType(90) {
		Particle[] createParticles(Animation animation) {
			Particle[] particles = new Particle[120];
			
			for(int i = 0; i < particles.length; ) {
				double y = 4 + Math.random() * 62;
				particles[i++] = new Particle(-1, y - 40, 10 * (i / 40) + (int)((70 - y) / 7), leftArc[0]);
				particles[i++] = new Particle(1, y - 40, 10 * (i / 40) + (int)((70 - y) / 7), rightArc[0]);
			}
			
			return particles;
		}

		void update(Particle particle, int frame) {
			frame -= particle.t;
			if(frame <= 0 || frame >= 40) {
				particle.image = null;
				return;
			}
			
			if(particle.x < 0) {
				if(frame <= 10) {
					particle.image = leftArc[10 - frame];
				} else if(frame > 30) {
					particle.image = leftArc[frame - 30];
				} else {
					particle.vx -= .2 * Math.random();
				}
			} else {
				if(frame <= 10) {
					particle.image = rightArc[10 - frame];
				} else if(frame > 30) {
					particle.image = rightArc[frame - 30];
				} else {
					particle.vx += .2 * Math.random();
				}
			}
		}
	};
	/**
	 * Animation type that shows a swirl of pink particles wrapping upwards around
	 * its target.
	 */
	public static final AnimationType SWIRL = new SimpleAnimationType(100) {
		Particle[] createParticles(Animation animation) {
			Particle[] particles = new Particle[150];
			
			for(int i = 0; i < particles.length; ) {
				double y = -6 + Math.random() * 32;
				double x = -24 + Math.random() * 10;
				int t = (int)(Math.random() * 5) * 10;
				for(int p = 0; p < 10; p++, t++) {
					particles[i++] = new Particle(x, y, 0, -1, t++, pinkRightArc[0]);
				}
			}
			
			return particles;
		}

		void update(Particle particle, int frame) {
			frame -= particle.t;
			if(frame <= 0) {
				particle.image = null;
				return;
			}
			
			int f = particle.t % 10;
			if(frame > 40) {
				f += (frame - 40);
				if(f > 9) {
					particle.image = null;
					return;
				}
			}
			
			if(particle.x > 20) particle.x = -20;
			particle.vx = 2 * Math.cos(particle.x * Math.PI / 50);
			
			particle.image = pinkRightArc[f];
		}
	};
	/**
	 * Animation type that shows an angry swarm of black particles gathering on
	 * its target.
	 */
	public static final AnimationType SWARM = new SimpleAnimationType(70) {
		Particle[] createParticles(Animation animation) {
			Particle[] particles = new Particle[80];
			
			for(int i = 0; i < particles.length; ) {
				double x = -25 + Math.random() * 50;
				double y = -30 + Math.random() * 60;
				
				particles[i++] = new Particle(x, y, (int)(Math.random() * 30), soot[0]);
			}
			
			return particles;
		}

		void update(Particle particle, int frame) {
			frame -= particle.t;
			if(frame < 0) {
				particle.image = null;
				return;
			}
			
			particle.vx = .7 * particle.vx;
			particle.vy = .7 * particle.vy + .1;
			
			if(Math.random() < .1) {
				particle.vx += 6 * Math.random() - 3 - particle.x / 20;
				particle.vy += 6 * Math.random() - 5 - particle.y / 20;
			}
			
			if(frame <= 9) {
				particle.image = soot[9 - frame];
			} else if(frame <= 30) {
			} else if(frame < 40) {
				particle.image = soot[frame - 30];
			} else if(frame == 40) {
				particle.image = null;
			}
		}
	};
	/**
	 * Animation type that shows purple lightning striking its target, followed
	 * by a white explosion.
	 */
	public static final AnimationType LIGHTNING = new SimpleAnimationType(30) {
		Particle[] createParticles(Animation animation) {
			ArrayList particles = new ArrayList();
			
			particles.add(new Particle(0, -30, 9, purple[0]));
			int lineStart = 0;
			for(int y = -3; y > -207; y -= 3) {
				int lastLineStart = lineStart;
				lineStart = particles.size();
				double branchProb = .2 / (lineStart - lastLineStart);
				
				for(int j = lastLineStart; j < lineStart; j++) {
					Particle p = (Particle)particles.get(j);
					int t = p.t - (int)(Math.random() * 1.4);
					if(t >= 0) {
						if(Math.random() < branchProb) {
							double vx;
							vx = Math.random() + 1;
							particles.add(new Particle(p.x + vx, y, vx, 0, t, purple[0]));
							vx = -Math.random() - 1;
							particles.add(new Particle(p.x + vx, y, vx, 0, t, purple[0]));
						} else {
							double vx = p.vx + Math.random() - .5;
							particles.add(new Particle(p.x + vx, y, vx, 0, t, purple[0]));
						}
					}
					p.vx = 0;
				}
			}
			for(int j = lineStart; j < particles.size(); j++) {
				((Particle)particles.get(j)).vx = 0;
			}			
			
			for(int i = 0; i < 30; i++) {
				double angle = Math.random() * Math.PI * 2;
				double speed = 5 + Math.random() * 2;
				
				particles.add(new Particle(0, 0, speed * Math.sin(angle), speed * Math.cos(angle), 10, white[0]));
			}
			
			return (Particle[])particles.toArray(new Particle[particles.size()]);
		}

		void update(Particle particle, int frame) {
			frame -= particle.t;
			
			if(frame < 0) {
				particle.image = null;
			} else if(particle.t == 10) {
				// Part of explosion
				particle.vx = .8 * particle.vx + .5 * Math.random() - .25;
				particle.vy = .8 * particle.vy + .5 * Math.random() - .15;
				particle.image = frame <= 10 ? white[0] : (frame > 19 ? null : white[frame - 10]);
			} else {
				// Part of lightning strike
				particle.image = frame < 10 ? purple[frame] : null;
			}
		}
	};
		
	private static class BloodAnimationType extends SimpleAnimationType {
		private int dir;
		
		private BloodAnimationType(int dir) {
			super(40);
			this.dir = dir;
		}
		
		Particle[] createParticles(Animation animation) {
			Particle[] particles = new Particle[50];
			double slope = .5 * Math.cos(Math.random() * Math.PI);
			final double intercept = 20 * (1 - slope + Math.random()) - 30;
			final double offset = Math.random() * 8 + 1 - 25;
			
			for(int frame = 0; frame < 10; frame++) {
				for(int i = 0; i < 5; i++) {
					double x = 20 + dir * (Math.random() * 4 + 4 * frame - 20);
					double y = x * slope + intercept + Math.random() * 3 - 1.5;
					
					particles[frame * 5 + i] = new Particle(x + offset, y, frame, blood[0]);
				}
			}
			return particles;
		}

		void update(Particle particle, int frame) {
			if(particle.y >= 30) {
				particle.image = null;
				return;
			}
			
			int dt = frame - particle.t;
			
			if(dt < 0) particle.image = null;
			else if(dt <= 9) {
				particle.image = blood[9 - dt];
			} else {
				particle.vy += .4 * Math.random();
			}
		}
	};
	
	/**
	 * Cancel every animation that is currently running.
	 */
	public static void cancelAll() {
		synchronized(animations) {
			for(Iterator iter = animations.iterator(); iter.hasNext(); ) {
				((Animation)iter.next()).cancel();
			}
		}
	}
	
	/**
	 * Cancel this animation.
	 */
	public void cancel() {
		cancelled = true;
	}
//	
//	private static void readImage(Image image) {
//		BufferedImage buff = VisibleImage.createBufferedCopy(image);
//		int w = buff.getWidth(null);
//		int h = buff.getHeight(null);
//		int[] pixels = new int[w * h];
//		buff.getRGB(0, 0, w, h, pixels, 0, w);
//		
//		System.out.print("new int[]{\n\t");
//		for(int i = 0; i < pixels.length; i++) {
//			int p = pixels[i];
//			
//			//p = (p & 0xFF0000FF) | ((p & 0x00FF0000) >> 8) | ((p & 0x0000FF00) << 8);
//			System.out.print(Integer.toString(p));
//			if((i + 1) % w == 0) {
//				if(i == pixels.length - 1) {
//					System.out.print("};\n");
//				} else {
//					System.out.print(",\n\t");
//				}
//			} else {
//				System.out.print(", ");
//			}
//		}
//	}
	
	private static Image[] makeImages(int[] pixels, int width, int height) {
		Image[] images = new Image[10];
		
		images[0] = new BufferedImage(width, height, BufferedImage.TYPE_INT_ARGB);
		((BufferedImage)images[0]).setRGB(0, 0, width, height, pixels, 0, width);
		
		for(int i = 1; i < 10; i++) {
			for(int p = 0; p < pixels.length; p++) {
				int alpha = ((((pixels[p] & 0xFF000000) >>> 4) * 7) << 1) & 0xFF000000;
				pixels[p] = (pixels[p] & 0xFFFFFF) | alpha;
			}
			images[i] = new BufferedImage(width, height, BufferedImage.TYPE_INT_ARGB);
			((BufferedImage)images[i]).setRGB(0, 0, width, height, pixels, 0, width);
		}
		
		return images;
	}
	
	static {
		int[] pixels;
		
		pixels = new int[]{
				0, 1727529217, 1071120898, 0,
				1073676802, -1711341568, -1716781056, 1726218497,
				1725104385, -1715142656, -1717436416, 1724121345,
				0, 1726218497, 1722941697, 1066074626};
		blood = makeImages(pixels, 4, 4);
		
		pixels = new int[]{
				0, 0, 863993727, 0, 1283555713, 423114808, 431799484, 423114808, 863993727, 0,
				436207615, 1288819153, 0, 863993727, 1291845631, 431799484, 872415231, 1283555713, 431799484, 436207615,
				427128181, 0, 1291845631, 1286187433, 423114808, 1284871573, 860375112, 855769602, 863993727, 1288819153,
				0, 869915097, 1284871573, 863993727, 872415231, 867612342, 867612342, 1279542340, 869915097, 427128181,
				427128181, 1286187433, 0, 427128181, 867612342, 1288819153, 419759365, 867612342, 860375112, 419759365,
				436207615, 427128181, 1291845631, 872415231, 1283555713, 1291845631, 1279542340, 869915097, 1286187433, 855769602,
				436207615, 427128181, 1286187433, 423114808, 872415231, 1284871573, 869915097, 423114808, 0, 423114808,
				0, 0, 855769602, 872415231, 860375112, 0, 867612342, 1284871573, 1291845631, 1284871573,
				1288819153, 0, 0, 1279542340, 431799484, 1283555713, 860375112, 860375112, 855769602, 436207615,
				0, 0, 1284871573, 0, 0, 419759365, 0, 1286187433, 0, 0};
		smoke = makeImages(pixels, 10, 10);
		
		pixels = new int[]{
				0, 0, 871826175, 872352985, 872352985, 0, 0,
				0, 872350400, -2098459905, -2097214742, -2097214273, 872350400, 0,
				871826175, -2097214273, -906025777, -47920, -906027617, -2097214742, 872350400,
				872352985, -2097214273, -57698, -41258, -57880, -2097214273, 872352985,
				872350400, -2097214742, -906032448, -55614, -906031384, -2098459905, 871826175,
				0, 872350400, -2098459905, -2097214273, -2097214742, 872350400, 0,
				0, 0, 871826175, 872352985, 872350400, 0, 0};
		lightning = makeImages(pixels, 7, 7);
		
		pixels = new int[]{
				37699391, 872546818, -2046688766, 1342243073, -2063465982, 838992386,
				889324034, -1610481150, -1258291200, -754974720, -1526595070, 1996554497,
				319161862, -2046688766, -218103808, -33554432, -754974720, 1828782337,
				1275134209, -1610481150, -536870912, -134217728, -1258291200, 1291911425,
				252184584, -1761476094, -1342177280, -1576926718, -1996357118, 2097217793,
				0, 235473161, 838992386, 302450439, 822215170, 135204623};
		soot = makeImages(pixels, 6, 6);
		
		pixels = new int[]{
				0, 0, 872396345, 872397590, 872401415, 0, 0,
				0, 872398890, 2147483393, 2147475745, 2147474689, 872401415, 0,
				872397590, 2147475745, -855642610, -953, -856031470, 2147483393, 872397590,
				872401415, 2147483393, -393454, -65671, -2276, 2147475745, 872401415,
				872398890, 2147474689, -855643337, -4806, -855644667, 2147474689, 872396345,
				0, 872401415, 2147483393, 2147475745, 2147483393, 872401415, 0,
				0, 0, 872396345, 872397590, 872401415, 0, 0};
		spark = makeImages(pixels, 7, 7);
		
		pixels = new int[]{
				0, 2138439427, 2137259793, 0,
				2132934411, -16711903, -443220223, 2134638343,
				2136080153, -866451712, -10551552, 2133589767,
				0, 2141454105, 2138701575, 0};
		green = makeImages(pixels, 4, 4);
		
		pixels = new int[]{
				0, 1275181579, 1275181579, 1275174667, 0,
				1275181579, -1308379635, -452746485, -1308385782, 1275174667,
				1275181579, -452746485, -16477173, -452820471, 1275165192,
				1275174667, -1308385782, -452820471, -1308395767, 1275165192,
				0, 1275174667, 1275165192, 1275165192, 0};
		darkGreen = makeImages(pixels, 5, 5);
		
		pixels = new int[]{
				0, 1282473729, 1282473729, 1279393537, 0,
				1282473729, -1299448064, -445645048, -1302266112, 1279393537,
				1282473729, -445645048, -9896185, -447873278, 1275657992,
				1279393537, -1302266112, -447873278, -1304232192, 1275657992,
				0, 1279393537, 1275657992, 1275657992, 0};
		limeGreen = makeImages(pixels, 5, 5);
		
		pixels = new int[]{
				0, 1291843318, 1291843318, 1291843318, 0,
				1291843318, -1291848203, -436210187, -1291848203, 1291843318,
				1291843318, -436210187, -2571, -436210187, 1291843318,
				1291843318, -1291848203, -436210187, -1291848203, 1291843318,
				0, 1291843318, 1291843318, 1291843318, 0};
		white = makeImages(pixels, 5, 5);
		
		pixels = new int[]{
				0, 0, 864618495, 864618495, 861995775, 0, 0,
				0, 864618495, 2140342271, 2140342271, 2137850367, 861995775, 0,
				864618495, 2140342271, -1298920961, -445840897, -865205249, 2137850367, 861995775,
				864618495, 2140342271, -445840897, -9762561, -445840897, 2136540671, 860357375,
				861995775, 2137850367, -865205249, -445840897, -1302593281, 2136540671, 860357375,
				0, 861995775, 2137850367, 2136540671, 2136540671, 860357375, 0,
				0, 0, 861995775, 860357375, 860357375, 0, 0};
		purple = makeImages(pixels, 7, 7);
		
		pixels = new int[]{
				-1308560897, -16583937, -16583937, -1308560897, 0, 0,
				1711535103, -1727859969, -872156161, -16583937, -1308560897, 0,
				855828722, 1711532531, -1727862542, -872159760, -16589593, -1308560897,
				0, 855826152, 1711528676, -1727866397, -872163873, -16589593,
				0, 855826152, 1711528676, -1727866397, -872163873, -16594991,
				855823582, 1711526106, -1727870766, -872166443, -16594991, -1308569379,
				1711523022, -1727874107, -872170298, -16599104, -1308573494, 0,
				-1308576324, -16599104, -16599104, -1308576064, 0, 0};
		rightArc = makeImages(pixels, 6, 8);
		
		pixels = new int[]{
				0, 0, -1308560897, -16583937, -16583937, -1308560897,
				0, -1308560897, -16583937, -872156161, -1727859969, 1711535103,
				-1308560897, -16589593, -872159760, -1727862542, 1711532531, 855828722,
				-16589593, -872163873, -1727866397, 1711528676, 855826152, 0,
				-16594991, -872163873, -1727866397, 1711528676, 855826152, 0,
				-1308569379, -16594991, -872166443, -1727870766, 1711526106, 855823582,
				0, -1308573494, -16599104, -872170298, -1727874107, 1711523022,
				0, 0, -1308576064, -16599104, -16599104, -1308576324};
		leftArc = makeImages(pixels, 6, 8);
		
		pixels = new int[]{
				-1292828417, -916737, -916737, -1292828417, 0, 0,
				1727202303, -1712192769, -856488961, -916737, -1292828417, 0,
				870843122, 1726546931, -1712848142, -857406480, -2358553, -1292828417,
				0, 870187752, 1725563876, -1713831197, -858455073, -2358553,
				0, 870187752, 1725563876, -1713831197, -858455073, -3734831,
				869532382, 1724908506, -1714945326, -859110443, -3734831, -1294991139,
				1724122062, -1715797307, -860093498, -4783424, -1296039734, 0,
				-1296760644, -4783424, -4783424, -1296695104, 0, 0};
		pinkRightArc = makeImages(pixels, 6, 8);
		
		pixels = new int[]{
				0, 0, 2147481075, -855640589, -855640589, 2147481075, 0, 0,
				0, 2147481075, -855640589, -2574, -2574, -860562689, 2142624255, 0,
				2147481075, -855640589, -2574, -3014670, -3014670, -4859393, -860562689, 2142615551,
				-855640589, -2574, -3014670, -5177345, -4462849, -5650433, -4868353, -860571649,
				-855640589, -2574, -3014670, -4462849, -5645057, -5650433, -4868353, -860571649,
				2147481075, -860562689, -4859393, -5650433, -5650433, -4868353, -860571649, 2142615551,
				0, 2142624255, -860562689, -4868353, -4868353, -860571649, 2142615551, 0,
				0, 0, 2142615551, -860571649, -860571649, 2142615551, 0, 0};
		hailstone = makeImages(pixels, 8, 8);
		
		pixels = new int[]{
				0, 0, 0, 855638015, 855638015, 855638015, 855638015, 855638015, 0, 0, 0, 0,
				0, 0, 855638015, 1711276031, 2147483647, 2147483647, 1711276031, 1291845631, 855638015, 0, 0, 0,
				0, 855638015, 1711276031, -1728053249, -855638017, -1728053249, -1728053249, 1711276031, 1291845631, 855638015, 0, 0,
				855638015, 1711276031, 2147483647, -1291845633, -855638017, -1291845633, -855638017, 2147483647, 1711276031, 1711276031, 855638015, 0,
				855638015, 1711276031, 2147483647, -1728053249, -855638017, -855638017, -1291845633, -1728053249, -1728053249, 2147483647, 1711276031, 855638015,
				0, 855638015, 1711276031, 2147483647, -855638017, -1291845633, -1291845633, -855638017, -1291845633, -1728053249, 2147483647, 855638015,
				0, 855638015, 1711276031, 2147483647, -1728053249, 2147483647, -1728053249, -855638017, -855638017, -1291845633, -1728053249, 855638015,
				855638015, 2147483647, -1291845633, -855638017, -1728053249, 1711276031, 2147483647, -1728053249, -1291845633, -1728053249, 2147483647, 855638015,
				855638015, 2147483647, -1291845633, -1728053249, 2147483647, 1291845631, 1711276031, 2147483647, -1728053249, 2147483647, 1711276031, 855638015,
				855638015, 1711276031, 2147483647, 2147483647, 1711276031, 855638015, 1291845631, 1711276031, 1711276031, 1711276031, 855638015, 0,
				0, 855638015, 1711276031, 1711276031, 1291845631, 855638015, 855638015, 1291845631, 1291845631, 855638015, 0, 0,
				0, 0, 855638015, 855638015, 855638015, 0, 0, 855638015, 855638015, 0, 0, 0};
		steam = makeImages(pixels, 12, 12);
		
		pixels = new int[]{
				855638015, 855638015, 0, 0, 0, 855638015, 855638015, 855638015, 855638015, 0, 0, 0,
				855638015, 0, 0, 855638015, 855638015, 1711276031, 2147483647, 2147483647, 1711276031, 855638015, 0, 0,
				0, 855638015, 855638015, 1291845631, 1711276031, 2147483647, -1728053249, -1728053249, 2147483647, 1711276031, 855638015, 0,
				0, 855638015, 1291845631, 1711276031, 2147483647, -1728053249, -1291845633, -1291845633, -1728053249, 2147483647, 1711276031, 855638015,
				855638015, 1711276031, 2147483647, 2147483647, -1291845633, -1291845633, -855638017, -855638017, -1291845633, 2147483647, 1711276031, 855638015,
				855638015, 2147483647, -1728053249, -1291845633, -1728053249, -1728053249, -1728053249, -1291845633, -1728053249, 1711276031, 855638015, 0,
				855638015, 2147483647, -1291845633, -855638017, 2147483647, 2147483647, 2147483647, -1291845633, -855638017, -1728053249, 1711276031, 855638015,
				855638015, 1711276031, 2147483647, 2147483647, 1711276031, 1711276031, -1728053249, -1728053249, -1291845633, -1291845633, 2147483647, 855638015,
				855638015, 855638015, 855638015, 1291845631, 1711276031, 1711276031, 2147483647, -1291845633, -855638017, -855638017, -1728053249, 855638015,
				0, 855638015, 0, 855638015, 1291845631, 1711276031, 2147483647, -1291845633, -855638017, -1291845633, 2147483647, 855638015,
				855638015, 855638015, 0, 0, 855638015, 1291845631, 1711276031, 2147483647, -1728053249, -1728053249, 1711276031, 855638015,
				0, 0, 855638015, 0, 0, 855638015, 855638015, 855638015, 855638015, 855638015, 855638015, 0};
		steam2 = makeImages(pixels, 12, 12);
		
		pixels = new int[]{
				0, 0, 0, 0, 118624786, 308042844, 0, 0, 0, 0, 0, 0,
				0, 0, 50315199, 1432048978, -884457409, -431669443, -1405076163, 625952840, 0, 0, 0, 0,
				0, 0, -1956752564, -11251892, -9937060, -10134696, -11450815, -648589495, 659245886, 0, 0, 0,
				46120895, 1834635340, -10791855, -9345181, -8555669, -8292752, -11451586, -11450560, -499831252, 1294543138, 0, 0,
				456998205, -194814888, -9410204, -9344925, -8161168, -9806248, -11713987, -12633803, -13817303, -81386719, 841359649, 0,
				506930717, -193696667, -8226444, -9016476, -10267057, -11581630, -12305094, -13489110, -13554646, -14804199, -2061691372, 0,
				508049207, -177643686, -10201005, -10528942, -10726579, -11450300, -12370886, -12896972, -13423059, -14606562, -937090784, 73359199,
				171917119, -649445060, -10199463, -10528684, -12305351, -12304579, -13225425, -13686230, -14343390, -14737891, -417325540, 236657435,
				0, 1597978935, -61715640, -13094867, -14080476, -13817303, -13817560, -14541282, -14672355, -14738148, -1927801324, 0,
				0, 0, 891560996, 1797069081, -887218916, -316794343, -266726124, -618849252, -1525081834, 1562122778, 186780427, 0,
				0, 0, 0, 0, 69148447, 387649552, 454498071, 185273099, 0, 0, 0, 0};
		asteroid = makeImages(pixels, 12, 11);
		
		pixels = new int[]{
				0, 0, 0, 855612162, 872369425, 855601666, 872398870, 0, 0, 0,
				0, 0, 872369425, 1711249411, 1711249411, 1711240961, 1711240961, 855601666, 0, 0,
				0, 855612162, 1711249411, 2147457797, -1291872250, -1291881472, 2147460873, 1728001288, 855601666, 0,
				855612162, 1728035344, 2147466517, -1291872250, -436233722, -436226026, -1291867126, 2147448065, 1728001288, 872398870,
				855612162, 1711249411, -1291863024, -436233722, -9713, -35328, -436243200, -1291863024, 1711240961, 855601666,
				872369425, 1711240961, -1291881472, -436247534, -35328, -39415, -436247534, -1291885047, 1728035344, 855599111,
				855601666, 1728035344, 2147433747, -1291891703, -436243200, -436226026, -1291896048, 2147433747, 1711236616, 855599111,
				0, 855601666, 1711240961, 2147448065, -1291889142, -1291885047, 2147443975, 1728001288, 872369425, 0,
				0, 0, 855601666, 1728035344, 1728035344, 1711236616, 1728001288, 855599111, 0, 0,
				0, 0, 0, 872398870, 855601666, 872369425, 855599111, 0, 0, 0};
		explosion = makeImages(pixels, 10, 10);
	}
	
	boolean step() {
		return !cancelled && (type.step(particles, frame, this) | checkLabels(frame++));
	}
	
	private static abstract class AnimationType {
		Animation fireOne(Animation animation, DrawableInterface object) {
			return null;
		}
		
		private void accum(final int[] pixels, final int[] accum, final int offset) {
			if(offset < 0) {
				for(int source = 0, dest = -offset; dest < pixels.length; ) {
					accum[dest++] += pixels[source++] >>> 24;
				}
			} else {
				for(int source = offset, dest = 0; source < pixels.length; ) {
					accum[dest++] += pixels[source++] >>> 24;
				}
			}
		}
		
		private BufferedImage getImage(VisibleImage image) {
			final Image im = image.getImage();
			return (im instanceof BufferedImage ? (BufferedImage)im : VisibleImage.createBufferedCopy(im));
		}
		
		Particle[] outline(final ObjectDrawShape basis, final double separation, final Image image) {
			ArrayList particles = new ArrayList();
			
			if(basis instanceof VisibleImage) {
				// Trace the shape using edge detection
				final BufferedImage target = getImage((VisibleImage)basis);
				final int w = target.getWidth(null);
				final int h = target.getHeight(null);
				final int[] pixels = new int[w * h];
				final int[] accum = new int[w * h];
				target.getRGB(0, 0, w, h, pixels, 0, w);
				
				accum(pixels, accum, 0);
				accum(pixels, accum, 1);
				accum(pixels, accum, w);
				accum(pixels, accum, w + 1);
				
				final double prob = .5 / separation;
				
				for(int y = 1; y < h - 1; y++) {
					for(int x = 1; x < w - 1; x++) {
						final int i = x + w * y;
						final int a = accum[i];
						if(a >= 20 && a <= 1000 && Math.random() < prob) {
							final double gx = .7 * (-accum[i - w - 1] - accum[i + w - 1] + accum[i - w + 1] + accum[i + w + 1]) - accum[i - 1] + accum[i + 1];
							final double gy = .7 * (-accum[i - w - 1] + accum[i + w - 1] - accum[i - w + 1] + accum[i + w + 1]) - accum[i - w] + accum[i + w];
							final double g = -1 / Math.sqrt(gx * gx + gy * gy);
							particles.add(new Particle(x, y, g * gx, g * gy, 0, image));
						}
					}
				}
				for(int y = 0; y < h; y++) {
					final int a = accum[y * w];
					if(a >= 20 && a <= 1000 && Math.random() < prob) {
						particles.add(new Particle(0, y, -1, 0, 0, image));
					}
					final int b = accum[y * w + w - 1];
					if(b >= 20 && b <= 1000 && Math.random() < prob) {
						particles.add(new Particle(w - 1, y, 1, 0, 0, image));
					}
				}
				for(int x = 0; x < w; x++) {
					final int a = accum[x];
					if(a >= 20 && a <= 1000 && Math.random() < prob) {
						particles.add(new Particle(x, 0, 0, -1, 0, image));
					}
					final int b = accum[(h - 1) * w + x];
					if(b >= 20 && b <= 1000 && Math.random() < prob) {
						particles.add(new Particle(x, h - 1, 0, 1, 0, image));
					}
				}
			} else {
				final Location origin = getOrigin(basis);
				final double dx = -origin.getDoubleX();
				final double dy = -origin.getDoubleY();
				
				PathMeasurer.Point[] points = PathMeasurer.getPoints(basis.getShape(), separation);
				for(int i = 0; i < points.length; i++) {
					final PathMeasurer.Point p = points[i]; 
					particles.add(new Particle(p.x + dx, p.y + dy, p.nx, p.ny, 0, image));
				}
			}
			return (Particle[])particles.toArray(new Particle[particles.size()]);
		}

		void draw(Graphics2D g, Particle particle) {
			if(particle != null && particle.image != null) g.drawImage(particle.image, 
					(int)(particle.x - particle.dx + .5), 
					(int)(particle.y - particle.dy + .5), 
					null);
		}
		
		abstract void draw(final Particle[] particles, final Graphics2D g, final Animation animation);
		abstract Location getLocation(DrawableInterface object);
		abstract Particle[] createParticles(Animation animation);
		abstract boolean step(Particle[] particles, int frame, Animation animation);
	}
	
	// Because Meteor moves across the screen, the VisibleImage actually moves with it.
	// The image will stay centered about the centroid of the meteor particles until
	// it hits, at which point it will stop moving and stay centered about the explosion.
	private static class SingleMeteorAnimationType extends SimpleAnimationType {
		public SingleMeteorAnimationType() {
			super(40);
			// Meteor has its own step() method.
		}

		Particle[] createParticles(Animation animation) {
			final Particle[] particles = new Particle[100];
			
			if(animation.specialRegions.isEmpty()) {
				throw new IllegalStateException("You must use addSpecialRegion() to specify a meteor's target.");
			}
			final Location target = getCenter((DrawableInterface)animation.specialRegions.get(0));
			final double tx = target.getX() - animation.getMyLocation().getX();
			final double ty = target.getY() - animation.getMyLocation().getY();
			final Particle accelerateParticle = particles[0] = new Particle(tx, ty, 1, null);
			
			// The meteor starts at (0, 0) (or its current position) and goes to (tx, ty)
			
			if(animation.particles != null) {
				// Copy the particles from the existing animation
				System.arraycopy(animation.particles, 0, particles, 1, 11);
			} else {
				// Create a new set of particles, moving (roughly) toward the target
				final double t = Math.sqrt(tx * tx + ty * ty);
				final double ox = -20 * tx / t + Math.random() * 40 - 20;
				final double oy = -20 * ty / t + Math.random() * 40 - 20;
				final double o = Math.sqrt(ox * ox + oy * oy);
				final double ovx = -3 * ox / o;
				final double ovy = -3 * oy / o;
				
				Particle prev = particles[1] = new Particle(ox, oy, ovx, ovy, 0, asteroid[0]);
				for(int i = 2; i <= 11; i++) {
					prev = particles[i] = new Particle(prev.x - ovx, prev.y - ovy, 0, explosion[i - 2]);
				}
			}
			
			final Particle head = particles[1];
			// Set acceleration such that meteor will hit in 20 frames
			double vAvgX = (tx - head.x) / 20;
			double vAvgY = (ty - head.y) / 20;
			
			// a = (vavg - vi) / (.5dt)
			accelerateParticle.vx = (vAvgX - head.vx) / 10;
			accelerateParticle.vy = (vAvgY - head.vy) / 10;
			
			return particles;
		}

		boolean step(final Particle[] particles, final int frame, final Animation animation) {
			if(frame < 20) {
				for(int i = 11; i > 1; ) {
					final Particle p = particles[i--];
					p.x = particles[i].x;
					p.y = particles[i].y;
				}
				final Particle head = particles[1];
				head.x += (head.vx += particles[0].vx);
				head.y += (head.vy += particles[0].vy);
			} else if(frame >= 40) {
				return false;
			} else {
				if(frame == 20) {
					final double px = particles[0].x;
					final double py = particles[0].y;
					for(int i = 0; i < particles.length; i++) {
						double angle = Math.random() * Math.PI * 2;
						double speed = Math.random() * 2 + 5;
						particles[i] = new Particle(px, py, speed * Math.sin(angle), speed * Math.cos(angle), 0, explosion[0]);
					}
				}
				
				final Image image = explosion[frame <= 30 ? 0 : frame - 30];
				for(int i = 0; i < particles.length; i++) {
					Particle p = particles[i];
					p.x += (p.vx = p.vx * .9 + .5 * Math.random() - .25);
					p.y += (p.vy = p.vy * .9 + .5 * Math.random() - .25);
					p.image = image;
				}
			}
			return true;
		}

		void update(Particle particle, int frame) { }
	}
	
	private static class MeteorAnimationType extends SimpleAnimationType {
		public MeteorAnimationType() {
			super(0);
			// Length is infinite; we deal with this by overriding step().
		}

		Animation fireOne(Animation animation, DrawableInterface object) {
			int c = animation.getCount();
			animation.setCount(--c);
			Particle[] p = new Particle[11];
			System.arraycopy(animation.particles, 11 * c, p, 0, 11);
			
			Animation anim = new Animation(p, METEOR, animation.basis);
			anim.addSpecialRegion(object);
			return anim;
		}

		Particle[] createParticles(Animation animation) {
			Particle[] particles = new Particle[11 * animation.count];
			
			for(int p = 0, i = 0; p < animation.count; p++) {
				double angle = Math.random() * Math.PI * 2;
				double r = Math.random() * 40;
				double speed = Math.sqrt(16 - .01 * r * r);
				double direction = Math.random() * Math.PI * 2;
				
				particles[i++] = new Particle(r * Math.sin(angle), r * Math.cos(angle), 
						speed * Math.sin(direction), speed * Math.cos(direction), (int)(Math.random() * 4), asteroid[0]);
				for(int c = 0; c < 10; c++) {
					particles[i++] = new Particle(0, 0, c, explosion[0]);
				}
			}
			return particles;
		}
		
		// m = 1
		// F = -kr
		// k = .01
		// PE = .5 kr^2
		// KE = .5v^2
		// PE + KE = .5k(40)^2 = 800k
		// KE = 800k - .5 kr^2
		// .5v^2 = 8 - .005r^2
		// v = sqrt(16 - .01r^2)
		
		boolean step(final Particle[] particles, final int frame, final Animation animation) {
			final int drawLength = 11 * animation.getCount();
			if(drawLength <= 0) return false;
			
			for(int i = 0; i < drawLength; i += 11) {
				Particle particle = particles[i];
				if(particle == null) continue;
				
				particle.vx += -.01 * particle.x;
				particle.vy += -.01 * particle.y;
				
				double d = particle.vx * particle.x + particle.vy * particle.y;
				
				particle = particles[i + 10];
				// Move contrail to old position of asteroid
				for(int c = 9; c >= 0; c--) {
					Particle prev = particles[i + c];
					particle.x = prev.x;
					particle.y = prev.y;
					particle.t = prev.t;
					particle = prev;
				}
				
				particle.x += particle.vx;
				particle.y += particle.vy;
				
				if(((particle.t & 1) == 1) != (d > 0)) particle.t = (particle.t + 1) & 3;
			}
			
			for(int i = 0; i < particles.length; i += 11) {
				for(int c = 1; c <= 10; c++) {
					final Particle particle = particles[i + c];
					int t = c - 1;
					if(frame < 10) {
						t += (10 - frame);
						particle.image = (t < 10 ? explosion[t] : null);
					} else {
						particle.image = explosion[t];
					}
				}
				particles[i].image = asteroid[frame < 9 ? 9 - frame : 0];
			}
			
			return true;
		}
		
		// Draw all (0 <= t < 2, then basis, then t >= 2
		public void draw(final Particle[] particles, final Graphics2D g, final Animation animation) {
			final int drawLength = 11 * animation.getCount();
			for(int i = 0; i < drawLength; i++) {
				final Particle particle = particles[i];
				if(particle.t < 2) draw(g, particle);
			}
			final int dx = animation.getMyLocation().getX();
			final int dy = animation.getMyLocation().getY();
			
			g.translate(-dx, -dy);
			animation.basis.draw(g);
			g.translate(dx, dy);
			
			for(int i = 0; i < drawLength; i++) {
				final Particle particle = particles[i];
				if(particle.t >= 2) draw(g, particle);
			}
		}

		void update(Particle particle, int frame) { }
	}
	
	private static abstract class SimpleAnimationType extends AnimationType {
		private final int yOffset, length;
		
		public SimpleAnimationType(int length) {
			this(length, 0);
		}
		
		public SimpleAnimationType(int length, int yOffset) {
			this.length = length;
			this.yOffset = yOffset;
		}
		
		// Create a buffer sufficient for this animation, and add an image to the object's canvas
		Location getLocation(DrawableInterface object) {
			Location loc = getCenter(object);
			loc.translate(0, yOffset);
			return loc;
		}
		
		public void draw(final Particle[] particles, final Graphics2D g, final Animation animation) {
			if(animation.frame >= length) return;
			
			for(int i = 0; i < particles.length; i++) draw(g, particles[i]);
		}
		
		abstract Particle[] createParticles(Animation animation);
		abstract void update(Particle particle, int frame);
		
		boolean step(final Particle[] particles, final int frame, final Animation animation) {
			for(int i = 0; i < particles.length; i++) {
				final Particle particle = particles[i];
				update(particle, frame);
				if(particle.image != null) {
					particle.x += particle.vx;
					particle.y += particle.vy;
				}
			}
			return frame < length;
		}
	}
	
	private static class Label {
		private final String text;
		private final int t;
		private final DrawableInterface object;
		private int offset;
		
		public Label(String text, DrawableInterface object, int t) {
			this.text = text;
			this.t = t;
			this.object = object;
		}
	}
	
	private static class Particle {
		private double x;
		private double y;
		private double vx;
		private double vy;
		private int t;
		private double dx;
		private double dy;
		private Image image;
		
		public Particle(double x, double y, int t, Image image) {
			this.x = x;
			this.y = y;
			this.t = t;
			setImage(image);
		}
		
		public Particle(double x, double y, double vx, double vy, int t, Image image) {
			this.x = x;
			this.y = y;
			this.vx = vx;
			this.vy = vy;
			this.t = t;
			setImage(image);
		}
		
		public void setImage(Image image) {
			if(image == null) {
				dx = dy = 0;
			} else {
				dx = .5 * image.getWidth(null);
				dy = .5 * image.getHeight(null);
			}
			this.image = image;
		}
	}
	
	private boolean isSpecial(double x, double y) {
		Location screenLoc = new Location(x + getMyLocation().getX(), y + getMyLocation().getY());
		for(int i = 0; i < specialRegions.size(); i++) {
			DrawableInterface region = (DrawableInterface)specialRegions.get(i);
			if(region.contains(screenLoc)) return true;
		}
		return false;
	}
	
	/**
	 * Start (or restart) this animation. You should be sure to set the count,
	 * labels, and any special regions before calling this method if you want them
	 * to appear correctly. If the animation is restarted before it has finished,
	 * it will jump back to the start of the animation without completing the
	 * sequence it was in.
	 * 
	 * @see #setCount(int)
	 * @see #label(DrawableInterface, String)
	 * @see #addSpecialRegion(DrawableInterface)
	 */
	public void start() {
		particles = type.createParticles(this);
		frame = 0;
		step();
		if(!animations.contains(this)) animations.add(this);
		if(getCanvas() == null) addToCanvas(oldCanvas);
		
		if(!timer.isRunning()) timer.start(.02, .02);
	}
	
	/**
	 * Split off another animation targeting the given object. This only works for
	 * certain animation types, such as <code>METEORS</code>, that have several pieces
	 * that can be targeted separately.
	 * <br><br>
	 * The returned animation will need to be <code>start()</code>'ed
	 * just as any other newly created animation would. It can have labels or
	 * special regions added to it first, like any other animation.
	 * <br><br>
	 * This method reduces the <code>count</code> of this animation by one, and
	 * if the count gets to zero, the animation is stopped and removed from
	 * the canvas.
	 * 
	 * @param object Target of the new animation
	 * @return newly created animation
	 */
	public Animation fireOne(DrawableInterface object) {
		return type.fireOne(this, object);
	}
	
	/**
	 * Get this animation's count. This is a number that has meaning only for
	 * animations like <code>METEORS</code> that consist of several discreet
	 * pieces. The count will be 1 unless you set it to something else.
	 * @return The number of pieces in this animation.
	 */
	public int getCount() {
		return count;
	}
	
	/**
	 * Set this animation's count. This is a number that has meaning only for
	 * animations like <code>METEORS</code> that consist of several discreet
	 * pieces. The count will be 1 unless you set it to something else. This
	 * method must be called before <code>start()</code> to have the proper effect.
	 * The count is automatically reduced by <code>fireOne()</code>, and the 
	 * animation is automatically removed from the canvas when the count gets 
	 * to zero.
	 * 
	 * @param n The number of pieces in this animation.
	 */
	public void setCount(int n) {
		this.count = n;
	}
	
	/**
	 * Mark the region enclosed by the given object to have some special behavior.
	 * @param object Objects whose bounding rectangle will be marked.
	 */
	public void addSpecialRegion(DrawableInterface object) {
		specialRegions.add(object);
	}
	
	/**
	 * Produce a label that will appear in front of the given object, move
	 * upwards, and then fade out. 
	 * 
	 * @param object Object to label
	 * @param label Text to display
	 */
	public void label(DrawableInterface object, String label) {
		label(object, label, 0);
	}
	
	/**
	 * Produce a label that will appear in front of the given object, move
	 * upwards, and then fade out. 
	 * 
	 * @param object Object to label
	 * @param label Text to display
	 * @param dt Number of frames to wait, before the start of the animation,
	 *        before displaying the label.
	 */
	public void label(final DrawableInterface object, String label, int dt) {
		int i;
		while((i = label.indexOf('\n')) != -1) {
			labels.add(new Label(label.substring(0, i), object, dt));
			dt += LABEL_FONT.getSize();
			label = label.substring(i + 1);
		}
		labels.add(new Label(label, object, dt));
	}
	
	/**
	 * Set whether this animation moves along with the object it decorates. If this
	 * is set to false, you can move the animation or its target and have them
	 * appear in separate places.
	 * @param attached
	 */
	public void setAttached(final boolean attached) {
		if(this.attached = attached) update();
	}
	
	// If basis is moved, move me as well.
	/**
	 * @deprecated for internal use only
	 */
	public void update() {
		super.update();
		
		if(!attached) return;
		
		final Location myLocation = getMyLocation();
		final Location basisLocation = basis.getMyLocation();
		if(!myLocation.equals(basisLocation)) {
			myLocation.translate(
					basisLocation.getX() - myLocation.getX() + dx, 
					basisLocation.getY() - myLocation.getY() + dy);
		}
	}
	
	/**
	 * Create an animation.
	 * @param type Type of animation. This will be one of the constants, like
	 *        <code>Animation.SPARKS</code>, that are defined in this class.
	 * @param object Target of the animation. The animation will be centered
	 *        around this object.
	 */
	public Animation(AnimationType type, DrawableInterface object) {
		this(null, type, object);
	}
	
	private Animation(Particle[] particles, AnimationType type, DrawableInterface object) {
		super(type.getLocation(object), false, null, object.getCanvas());
		this.type = type;
		if(object == null) throw new IllegalArgumentException("Object to decorate cannot be null.");
		
		oldCanvas = object.getCanvas();
		
		this.basis = (ObjectDrawShape)object;
		final Location basisLoc = basis.getMyLocation();
		final Location myLoc = getMyLocation();
		dx = myLoc.getX() - basisLoc.getX();
		dy = myLoc.getY() - basisLoc.getY();
		basisLoc.depend(this);
		
		this.particles = particles;
		setColor(defaultLabelColor);
		ready();
	}
	
	static class AnimationRunner {
		public void step() {
			final HashSet canvases = new HashSet();
			
			synchronized(animations) {
				for(Iterator iter = animations.iterator(); iter.hasNext(); ) {
					Animation animation = (Animation)iter.next();
					if(!animation.step()) {
						iter.remove();
						animation.removeFromCanvas();
					} else canvases.add(animation.getCanvas());
				}
				
				if(animations.isEmpty()) timer.stop();
			}
			
			for(Iterator iter = canvases.iterator(); iter.hasNext(); ) {
				final DrawingCanvas canvas = (DrawingCanvas)iter.next();
				if(canvas != null) canvas.update();
			}
		}
	}
	
	private boolean checkLabels(final int frame) {
		final int earliestFrame = frame - 40;
		for(int i = 0; i < labels.size(); i++) {
			if(((Label)labels.get(i)).t >= earliestFrame) return true;
		}
		return false;
	}
	
	private static Location getCenter(DrawableInterface object) {
		final Rectangle2D bounds = ((ObjectDrawShape)object).getShape().getBounds2D();
		return new Location(bounds.getCenterX(), bounds.getCenterY());
	}
	
	private static Location getOrigin(DrawableInterface object) {
		return new Location(((ObjectDrawShape)object).getMyLocation());
	}
	
	private void renderLabels(final Graphics2D g, final int frame) {
		g.setFont(LABEL_FONT);
		
		if(firstLabelRender) {
			firstLabelRender = false;
			
			FontMetrics fm = g.getFontMetrics();
			
			for(int i = 0; i < labels.size(); i++) {
				final Label label = (Label)labels.get(i);
				label.offset = -fm.stringWidth(label.text) / 2;
			}
		}
		
		for(int i = 0; i < labels.size(); i++) {
			final Label label = (Label)labels.get(i);
			final int dt = frame - label.t;
			
			if(dt < 0 || dt >= 40) continue;
			
			g.setColor(labelColors[dt > 30 ? dt - 30 : 0]);
			
			final Location origin = getCenter(label.object);
			final int x = origin.getX() + label.offset;
			final int y = origin.getY() - dt;
			
			g.drawString(label.text, x, y);
		}
	}
	
	/* * * * Methods from DrawableInterface * * * */
	
	/**
	 * Make this object invisible. It stays in position on its canvas; it is simply not drawn.
	 * This method has no effect if the object was already hidden.
	 */
	public void hide() { super.hide(); }
	/**
	 * Make an object visible again after it had been hidden. This method has no effect if the
	 * object was already visible.
	 */
	public void show() { super.show(); }
	/**
	 * Return <code>true</code> if this object has been rendered invisible with a call to its
	 * <code>hide()</code> method, <code>false</code> if it is still being drawn.
	 * @return whether this object is hidden
	 */
	public boolean isHidden() { return super.isHidden(); }

	/**
	 * Place this object on the specified canvas. The object will be removed from its current
	 * canvas, if it is on a canvas that is not the one it is being moved to. The object will
	 * be at the front of the canvas, as if it had just been drawn. If the given canvas is 
	 * <code>null</code> the object is simply removed from its current canvas.
	 * @param c <code>DrawingCanvas</code> to place this object on
	 */
	public void addToCanvas(DrawingCanvas c) { super.addToCanvas(c); }
	/**
	 * Remove this object from its canvas.
	 */
	public void removeFromCanvas() { super.removeFromCanvas(); }
	/**
	 * Return the canvas that this object is on. If it has been removed from its canvas, this
	 * method will return <code>null</code>.
	 * @return the <code>DrawingCanvas</code> that this object is on.
	 */
	public DrawingCanvas getCanvas() { return super.getCanvas(); }

	/**
	 * Move the center of this animation to the given location. If this animation 
	 * is attached to its target, the target will move with it.
	 * 
	 * @see #setAttached(boolean)
	 * @param point <code>Location</code> to move to
	 */
	public void moveTo(Location point) { 
		if(attached) basis.moveTo(point.getDoubleX() - dx, point.getDoubleY() - dy);
		else super.moveTo(point); 
	}
	/**
	 * Move the center of this animation to the given (code>x</code>, <code>y</code>). 
	 * If this animation is attached to its target, the target will move with it.
	 * @see #setAttached(boolean)
	 * @param x x coordinate to move to
	 * @param y y coordinate to move to
	 */
	public void moveTo(double x, double y) { 
		if(attached) basis.moveTo(x - dx, y - dy);
		else super.moveTo(x, y); 
	}
	/**
	 * Shift (translate) this object left or right by an amount <code>dx</code>, and up or
	 * down by an amount <code>dy</code>. A negative <code>dx</code> indicates a shift to the
	 * left; a negative <code>dy</code> indicates a shift up. If this animation is 
	 * attached to its target, the target will move with it.
	 * @see #setAttached(boolean)
	 * @param dx amount of x translation
	 * @param dy amount of y translation
	 */
	public void move(double dx, double dy) { 
		if(attached) basis.move(dx, dy);
		else super.move(dx, dy); 
	}

	/**
	 * Return the color that this animation's labels are drawn with.
	 * @return the <code>Color</code> that this animation's labels are drawn with
	 */
	public Color getColor() { return super.getColor(); }
	/**
	 * Change the color that this animation's labels are drawn with.
	 * @param c new <code>Color</code> for this animation's labels
	 */
	public void setColor(Color c) { 
		super.setColor(c); 
		
		final int RGB = (c.getRGB() & 0xFFFFFF);
		final int alpha = c.getAlpha();
		
		labelColors[0] = c;
		for(int i = 0; i < 10; i++) {
			labelColors[i] = new Color(RGB | ((alpha * (10 - i) / 10) << 24), true);
		}
	}

	/**
	 * Move this object one step forward in the draw order, causing it to be drawn on top of
	 * the objects that are behind it.
	 */
	public void sendForward() { super.sendForward(); }
	/**
	 * Move this object one step backward in the draw order, causing it to be drawn underneath
	 * the objects that are in front of it.
	 */
	public void sendBackward() { super.sendBackward(); }
	/**
	 * Move this object to the very front of the draw order, causing it to be drawn on top of
	 * all the other objects.
	 */
	public void sendToFront() { super.sendToFront(); }
	/**
	 * Move this object to the very back of the draw order, causing it to be drawn underneath
	 * all the other objects.
	 */
	public void sendToBack() { super.sendToBack(); }

	/**
	 * Return <code>true</code> if the given location is inside this object, <code>false</code>
	 * otherwise. Because <code>Line</code>s and <code>FramedArc</code>s have no "inside", a
	 * point is considered to be contained in one of them as long as it is within a few pixels
	 * of the line.
	 * @param point <code>Location</code> to test
	 * @return whether the given location is inside this shape
	 */
	public boolean contains(Location point) { return false; }

	/**
	 * Return a string describing how this object might be constructed.
	 */
	public String toString() {
		return "new Animation()";
	}
	
	/**
	 * @deprecated for internal use only
	 */
	public void draw(Graphics2D g) {
		if(frame == 0) return;
		
		AffineTransform saveAT = g.getTransform();
		g.translate(getMyLocation().getX(), getMyLocation().getY());
		type.draw(particles, g, this);
		g.setTransform(saveAT);
		renderLabels(g, frame);
	}

	/**
	 * @deprecated for internal use only
	 */
	public Shape makeShape() {
		return new GeneralPath();
	}
	
	/* * * * Methods from LocatableInterface * * * */
	
	/**
	 * Get this object's x coordinate.
	 * @return this object's x coordinate
	 */
	public int getX() { return getMyLocation().getX(); }
	/**
	 * Get this object's y coordinate.
	 * @return this object's y coordinate
	 */
	public int getY() { return getMyLocation().getY(); }
	/**
	 * Get this object's x coordinate, in double precision.
	 * @return this object's x coordinate, in double precision
	 */
	public double getDoubleX() { return getMyLocation().getDoubleX(); }
	/**
	 * Get this object's y coordinate, in double precision.
	 * @return this object's y coordinate, in double precision
	 */
	public double getDoubleY() { return getMyLocation().getDoubleY(); }
	/**
	 * Get this object's location. Translating this <code>Location</code> object
	 * will have the same effect as moving the object. So, the following two lines
	 * of code are equivalent:
	 * <br><br><code>
	 * object.move(10, 20);<br>
	 * object.getLocation().translate(10, 20);
	 * </code><br><br>If you want to create a <code>Location</code> offset by some
	 * amount from the <code>Location</code> of this object, without actually
	 * moving this object, use the <code>offset()</code> method in <code>Location</code>,
	 * which creates a new <code>Location</code>.
	 * 
	 * @return the <code>Location</code> of this object.
	 */
	public Location getLocation() { return getMyLocation(); }
}
