package objectdraw;

import java.io.IOException;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.nio.channels.SelectionKey;
import java.nio.channels.SocketChannel;

import javax.swing.SwingUtilities;

/**
 * A <code>Client</code> makes a connection to another computer (the server) and can
 * send messages to the server and receive messages back. You will never create a
 * <code>Client</code> directly. Instead, create an object that implements the 
 * <code>ClientListener</code> interface, and pass that object to <code>Network.configure()</code>
 * so that it will be notified when the user has connected to a server and when 
 * messages come in from that server.
 * <br><br>
 * The <code>Client</code> object is passed to all the <code>ClientListener</code> 
 * methods. You can use that object to send a reply to the server right away, using
 * the <code>sendMessage()</code> method, or you can keep the <code>Client</code>
 * in an instance variable so that you can send messages later when you need to. The
 * <code>Client</code> also has methods to determine the name the user entered for
 * their <code>Client</code>, and the name and address of the <code>Server</code>
 * that they joined.
 * @see ClientListener
 * @see Network#configure(String, ServerListener, ClientListener)
 * @author Russell Zahniser (russell@zahniser.net)
 */
public class Client extends NetworkThread {
	private final InetSocketAddress serverAddress;
	private final ClientListener listener;
	private int channelID = -1;
	private String serverName = null;
	
	Client(final String name, final InetAddress server) throws IOException {
		super(name);
	    
		serverAddress = new InetSocketAddress(server, Network.gamePort);
	    listener = Network.clientListener;
	    
	    start();
	}
	
	/**
	 * Send a message to the server. This will result in the server listener getting
	 * a <code>messageReceived()</code> event.
	 * @param message The message to send.
	 */
	public void sendMessage(final String message) {
		enqueueMessage(message, channelID);
	}
	
	/**
	 * Get the client name. This is the name that the user entered in the
	 * "Join network game" window.
	 * @return the client name
	 */
	public String getName() {
		return super.getName();
	}
	
	/**
	 * Get the server name. This is the name that the user selected from the
	 * list in the "Join network game" window.
	 * @return the server name
	 */
	public String getServerName() {
		return serverName;
	}
	
	/**
	 * Get the server address.
	 * @return the server address
	 */
	public InetSocketAddress getServerAddress() {
		return serverAddress;
	}
	
	/**
	 * Check if the client is still connected.
	 * @return <code>true</code> if the client is running, <code>false</code> otherwise.
	 */
	public boolean isRunning() {
		return super.isRunning();
	}
	
	/**
	 * Shut down the client, disconnecting it from the server. The client listener
	 * will receive a <code>clientDisconnected()</code> message, and the server
	 * listener will receive a <code>channelClosed()</code> event. Quitting the
	 * program will automatically shut down the client.
	 */
	public void shutDown() {
		super.shutDown();
	}
	
	void messageReceived(final int channel, final String channelName, final String message) {
		SwingUtilities.invokeLater(new Runnable() {
			public void run() {
				listener.messageReceived(Client.this, message);
			}
		});
	}
	
	void channelClosed(final int channel, final String channelName) {
		if(channel == channelID) {
			channelID = -1;
			SwingUtilities.invokeLater(new Runnable() {
				public void run() {
					listener.clientDisconnected(Client.this);
				}
			});
		}
		shutDown();
	}
	
	void channelReady(final int channel, final String channelName) {
		channelID = channel;
		SwingUtilities.invokeLater(new Runnable() {
			public void run() {
				listener.clientConnected(Client.this);
			}
		});
	}

	boolean setup() {
		// This initial configuration, plus the way pending messages are handled,
		// is essentially all that is different between a client and a server.
		// We should extract the shared functionality into a superclass.
		if(!super.setup()) return false;

        try {
        	final SocketChannel channel = SocketChannel.open();
        	channel.configureBlocking(false);
        	channel.connect(serverAddress);
        	
        	channel.register(selector, SelectionKey.OP_CONNECT, null);      
		} catch (IOException e) {
			e.printStackTrace();
			closeAll();
			return false;
		}
		
		return true;
	}
	
	void tearDown() {
		Network.closeClient(this);
		
		super.tearDown();
	}
}
