package objectdraw;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Rectangle;
import java.awt.event.ComponentListener;
import java.awt.event.KeyListener;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.image.ImageObserver;

/**
 * <code>DrawingCanvas</code> is the interface listing the methods of the canvas on which
 * objects are placed. It has methods dealing with the list of objects, its size, its background
 * color, the events it receives, and whether it repaints automatically when a change is made.
 * It is also possible to use a special magnifying-glass feature called the "loupe", which 
 * provides a zoomed-in view of a circular portion of the canvas.
 * 
 * @author Russell Zahniser (russell@zahniser.net)
 */
public interface DrawingCanvas extends Dependent {
	/**
	 * @deprecated for internal use only
	 */
	public Graphics getGraphics();
	/**
	 * @deprecated for internal use only
	 */
	public boolean prepareImage(Image im, ImageObserver observer);
	/**
	 * @deprecated for internal use only
	 */
	public Image createImage(int width, int height);
	/**
	 * @deprecated for internal use only
	 */
	public boolean requestFocusInWindow();
	/**
	 * @deprecated for internal use only
	 */
	public void requestFocus();
	/**
	 * @deprecated for internal use only
	 */
	public Rectangle getBounds();
	
	/**
	 * Remove all objects from this canvas.
	 */
	public void clear();
	
	/**
	 * Repaint the canvas. Normally, the canvas is repainted automatically when any change is 
	 * made, so using this method is only necessary if you have disabled auto-repainting.
	 */
	public void repaint();
	/**
	 * Set the canvas to repaint only when you explicitly call its <code>repaint()</code>
	 * method. This can be useful if you want to be able to make a lot of changes at once
	 * and not have the canvas draw in a state where the changes are incomplete.
	 */
	public void disableAutoRepaint();
	/**
	 * Set the canvas to repaint automatically when a change is made. This does not mean that 
	 * it has to redraw the whole canvas for every change; if a string of changes are made, it 
	 * remembers that it needs to redraw, and does so once the changes are done.
	 */
	public void enableAutoRepaint();
	
	/**
	 * Return the size of the canvas as a <code>Dimension</code> object.
	 * @return the size of the canvas
	 */
	public Dimension getSize();
	/**
	 * Return the width of the canvas
	 * @return the width of the canvas
	 */
	public int getWidth();
	/**
	 * Return the height of the canvas
	 * @return the height of the canvas
	 */
	public int getHeight();
	
	/**
	 * Add a mouse listener to the canvas. This is an alternative to using the mouse event
	 * methods in <code>Controller</code>.
	 * @param l an object that will react to mouse events in the canvas
	 */
	public void addMouseListener(MouseListener l);
	/**
	 * Add a mouse motion listener to the canvas. This is an alternative to using the mouse event
	 * methods in <code>WindowController</code>.
	 * @param l an object that will react to mouse motion events in the canvas
	 */
	public void addMouseMotionListener(MouseMotionListener l);
	/**
	 * Add a component listener to the canvas. This is a way to be notified when the canvas is
	 * resized, moved, hidden, or shown.
	 * @param l an object that will react to component events in the canvas
	 */
	public void addComponentListener(ComponentListener l);
	/**
	 * Add a key listener to the canvas. This is an alternative to using the key event
	 * methods in <code>WindowController</code>.
	 * @param l an object that will react to key events in the canvas
	 */
	public void addKeyListener(KeyListener l);
	
	/**
	 * Return a <code>DrawableIterator</code> that will list off, one by one, the 
	 * <code>DrawableInterface</code> objects that exist on this canvas.
	 * @return a <code>DrawableIterator</code> for all the objects on this canvas
	 */
	public DrawableIterator getDrawableIterator();
	
	/**
	 * Return the background color of this canvas
	 * @return the background <code>Color</code> of the canvas
	 */
	public Color getBackground();
	/**
	 * Change the background color of the canvas
	 * @param c new <code>Color</code> for the background of the canvas
	 */
	public void setBackground(Color c);
	
	/**
	 * Move the loupe to the given location, and change its size and zoom factor. This will 
	 * also cause it to be displayed, if it was not displayed previously. The
	 * <code>size</code> is the diameter of the circle to be magnified, and <code>zoom</code>
	 * is how many times bigger to make it. So, the circle you will actually see on the
	 * canvas has a diameter of <code>size * zoom</code>. 
	 * @param point center <code>Location</code> to magnify
	 * @param size diameter of the circle to be magnified
	 * @param zoom how many times bigger to show the image
	 */
	public void setLoupe(Location point, double size, double zoom);
	/**
	 * Move the loupe to center on the given location. This will also cause it to be displayed,
	 * if it was not displayed previously.
	 * @param point <code>Location</code> to move the loupe to
	 */
	public void setLoupe(Location point);
	/**
	 * Hide the loupe.
	 */
	public void clearLoupe();
}
