package objectdraw;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import javax.swing.Timer;

/**
 * <code>EventTimer</code> will repeatedly send a message to an object by calling
 * one of its methods with a preset list of parameters. The target object, method
 * to call, and values for the parameters are set when the <code>EventTimer</code>
 * is created. It can then be told to <code>start()</code> sending events, with
 * some delay before the first event and then a different delay between each later
 * event. The timer will then continue to call that method until it is told to
 * <code>stop()</code>. It is also possible to have the timer fire once and stop
 * automatically, by giving it a delay that is less than or equal to zero.
 * 
 * @author Russell Zahniser (russell@zahniser.net)
 */
public class EventTimer {
	/**
	 * This is a special object that may be included among the target parameters 
	 * to have this timer pass itself as a parameter. The parameter which will
	 * receive it should have type <code>EventTimer</code>.
	 */
	public static final Object SELF = new Object();
	
	private static final Set activeTimers = Collections.synchronizedSet(new HashSet());
	
	/**
	 * Stop all active timers. If your program is running as an applet in a web browser, 
	 * this method is automatically called when the user navigates away from the 
	 * page containing your applet. This prevents all your <code>EventTimer</code>s
	 * from just merrily churning away with no applet left to draw on.
	 */
	public static void stopAll() {
		synchronized(activeTimers) {
			for(Iterator iter = activeTimers.iterator(); iter.hasNext(); ) {
				((EventTimer)iter.next()).timer.stop();
			}
			activeTimers.clear();
		}
	}
	
	private Timer timer;
	private Object[] args;
	private Object target;
	private Method method;
	private long lastFireTime = -1;
	private long startTime = -1;
	private long initialFireTime = -1;
	private ActionListener listener = new ActionListener() {
		public void actionPerformed(ActionEvent e) {
			fire();
		}
	};
	
	/**
	 * Create a timer that will call <code>target.methodName(args[0], args[1], ...)</code> 
	 * every time it fires. The method to be called must have the same number of parameters
	 * as there are objects in <code>args</code>, and the types must match. If one of the 
	 * parameters is a primitive type (<code>int</code>, <code>double</code>, etc.) you 
	 * can use the appropriate "wrapper" class for its value (as in, 
	 * <code>new Integer(2)</code> instead of just 2).
	 * 
	 * @param target Object whose method will be called
	 * @param methodName Name of method to be called
	 * @param args Array of objects to pass for the method's parameters
	 */
	public EventTimer(Object target, String methodName, Object[] args) {
		setTarget(target, methodName, args);
	}
	
	/**
	 * Set this timer to call <code>target.methodName(args[0], args[1], ...)</code> 
	 * every time it fires. The method to be called must have the same number of parameters
	 * as there are objects in <code>args</code>, and the types must match. If one of the 
	 * parameters is a primitive type (<code>int</code>, <code>double</code>, etc.) you 
	 * can use the appropriate "wrapper" class for its value (as in, 
	 * <code>new Integer(2)</code> instead of just 2).
	 * 
	 * @param target Object whose method will be called
	 * @param methodName Name of method to be called
	 * @param args Array of objects to pass for the method's parameters
	 */
	public void setTarget(Object target, String methodName, Object[] args) {
		Class[] types = new Class[args.length];
		for(int i = 0; i < args.length; i++) {
			if(args[i] == SELF) args[i] = this;
			
			types[i] = (args[i] == null ? null : args[i].getClass());
		}
		
		if(target == null) throw new IllegalArgumentException("Timer target cannot be null.");
		
		try {
			Class targetClass = (target instanceof Class ? (Class)target : target.getClass());
			Method[] methods = targetClass.getMethods();
			
			for(int i = 0; i < methods.length; i++) {
				if(matches(methods[i], methodName, types)) {
					method = methods[i];
					this.target = target;
					this.args = args;
					timer = new Timer(1000, listener);
					
					checkTypes(args, method);
					break;
				}
			}
		} catch (SecurityException e) {
			throw new IllegalArgumentException("The security manager does not permit access to the methods of this target.");
		}
		
		if(method == null) {
			StringBuffer error = new StringBuffer();
			error.append("Unable to locate method: public void ");
			error.append(methodName);
			error.append("(");
			for(int i = 0; i < types.length; i++) {
				if(i != 0) error.append(", ");
				error.append(simpleClassName(types[i]));
			}
			error.append(")");
			throw new IllegalArgumentException(error.toString());
		}
	}
	
	private static boolean matches(Method method, String methodName, Class[] types) {
		if(!method.getName().equals(methodName)) return false;
		
		Class[] params = method.getParameterTypes();
		if(params.length != types.length) return false;
		
		for(int i = 0; i < params.length; i++) {
			if(!typeMatch(types[i], params[i])) return false;
		}
		return true;
	}
	
	private static boolean typeMatch(Class valueClass, Class paramClass) {
		if(valueClass == null) {
			// If the value given is null, it can be passed for any non-primitive parameter
			return !paramClass.isPrimitive();
		} else if(paramClass.isPrimitive()) {
			// If the parameter is primitive, the value needs to be convertible to the
			// wrapper for that primitive type
			if(paramClass == boolean.class) {
				return valueClass == Boolean.class;
			} else if(paramClass == char.class) {
				return valueClass == Character.class;
			} else {
				return Number.class.isAssignableFrom(valueClass);
			}
		} else {
			return paramClass.isAssignableFrom(valueClass);
		}
	}
	
	private static void checkTypes(Object[] values, Method method) {
		Class[] params = method.getParameterTypes();
		for(int i = 0; i < params.length; i++) {
			if(params[i].isPrimitive() && params[i] != boolean.class && params[i] != char.class) {
				// In this case, we might need to do a number value conversion to the appropriate class
				// The value class is guaranteed to be a Number subclass
				Number n = (Number)values[i];
				if(params[i] == byte.class) {
					if(!(values[i] instanceof Byte)) values[i] = new Byte(n.byteValue());
				} else if(params[i] == short.class) {
					if(!(values[i] instanceof Short)) values[i] = new Short(n.shortValue());
				} else if(params[i] == int.class) {
					if(!(values[i] instanceof Integer)) values[i] = new Integer(n.intValue());
				} else if(params[i] == long.class) {
					if(!(values[i] instanceof Long)) values[i] = new Long(n.longValue());
				} else if(params[i] == float.class) {
					if(!(values[i] instanceof Float)) values[i] = new Float(n.floatValue());
				} else if(params[i] == double.class) {
					if(!(values[i] instanceof Double)) values[i] = new Double(n.doubleValue());
				}
			}
		}
	}
	
	private static String simpleClassName(Class c) {
		String s = c.getName();
		return s.substring(s.lastIndexOf(".") + 1);
	}
	
	/**
	 * Create a timer that will call <code>target.methodName()</code> every time it 
	 * fires. The method must not take any parameters.
	 * 
	 * @param target Object whose method will be called
	 * @param methodName Name of method to be called
	 */
	public EventTimer(Object target, String methodName) { 
		setTarget(target, methodName, new Object[]{});
	}
	
	/**
	 * Set this timer to call <code>target.methodName()</code> every time it 
	 * fires. The method must not take any parameters.
	 * 
	 * @param target Object whose method will be called
	 * @param methodName Name of method to be called
	 */
	public void setTarget(Object target, String methodName) {
		setTarget(target, methodName, new Object[]{});
	}
	
	/**
	 * Create a timer that will call <code>target.methodName(arg1)</code> every time it
	 * fires. The method to be called must have a single parameter matching the type of 
	 * <code>arg1</code>. If the parameter is a primitive type (<code>int</code>, 
	 * <code>double</code>, etc.) you can use the appropriate "wrapper" class for its 
	 * value (as in, <code>new Integer(2)</code> instead of just 2).
	 * 
	 * @param target Object whose method will be called
	 * @param methodName Name of method to be called
	 * @param arg1 Object to pass as a parameter. 
	 */
	public EventTimer(Object target, String methodName, Object arg1) { 
		setTarget(target, methodName, new Object[]{arg1}); 
	}
	
	/**
	 * Set this timer to call <code>target.methodName(arg1)</code> every time it
	 * fires. The method to be called must have a single parameter matching the type of 
	 * <code>arg1</code>. If the parameter is a primitive type (<code>int</code>, 
	 * <code>double</code>, etc.) you can use the appropriate "wrapper" class for its 
	 * value (as in, <code>new Integer(2)</code> instead of just 2).
	 * 
	 * @param target Object whose method will be called
	 * @param methodName Name of method to be called
	 * @param arg1 Object to pass as a parameter. 
	 */
	public void setTarget(Object target, String methodName, Object arg1) { 
		setTarget(target, methodName, new Object[]{arg1}); 
	}
	
	/**
	 * Create a timer that will call <code>target.methodName(arg1, arg2)</code> every time it
	 * fires. The method to be called must have exactly two parameters, and they must match 
	 * the types of <code>arg1</code> and <code>arg2</code>. If one of the parameters is a 
	 * primitive type (<code>int</code>, <code>double</code>, etc.) you can use the appropriate 
	 * "wrapper" class for its value (as in, <code>new Integer(2)</code> instead of just 2).
	 * 
	 * @param target Object whose method will be called
	 * @param methodName Name of method to be called
	 * @param arg1 Object to pass as the first parameter. 
	 * @param arg2 Object to pass as the second parameter. 
	 */
	public EventTimer(Object target, String methodName, Object arg1, Object arg2) { 
		setTarget(target, methodName, new Object[]{arg1, arg2}); 
	}
	
	/**
	 * Set this timer to call <code>target.methodName(arg1, arg2)</code> every time it
	 * fires. The method to be called must have exactly two parameters, and they must match 
	 * the types of <code>arg1</code> and <code>arg2</code>. If one of the parameters is a 
	 * primitive type (<code>int</code>, <code>double</code>, etc.) you can use the appropriate 
	 * "wrapper" class for its value (as in, <code>new Integer(2)</code> instead of just 2).
	 * 
	 * @param target Object whose method will be called
	 * @param methodName Name of method to be called
	 * @param arg1 Object to pass as the first parameter. 
	 * @param arg2 Object to pass as the second parameter. 
	 */
	public void setTarget(Object target, String methodName, Object arg1, Object arg2) { 
		setTarget(target, methodName, new Object[]{arg1, arg2}); 
	}
	
	/**
	 * Create a timer that will call <code>target.methodName(arg1, arg2, arg3)</code> every 
	 * time it fires. The method to be called must have exactly three parameters, and they 
	 * must match the types of <code>arg1</code>, <code>arg2</code> and <code>arg3</code>. 
	 * If one of the parameters is a primitive type (<code>int</code>, <code>double</code>, 
	 * etc.) you can use the appropriate "wrapper" class for its value (as in, 
	 * <code>new Integer(2)</code> instead of just 2).
	 * 
	 * @param target Object whose method will be called
	 * @param methodName Name of method to be called
	 * @param arg1 Object to pass as the first parameter. 
	 * @param arg2 Object to pass as the second parameter. 
	 * @param arg3 Object to pass as the third parameter. 
	 */
	public EventTimer(Object target, String methodName, Object arg1, Object arg2, Object arg3) { 
		setTarget(target, methodName, new Object[]{arg1, arg2, arg3}); 
	}
	
	/**
	 * Set this timer to call <code>target.methodName(arg1, arg2, arg3)</code> every 
	 * time it fires. The method to be called must have exactly three parameters, and they 
	 * must match the types of <code>arg1</code>, <code>arg2</code> and <code>arg3</code>. 
	 * If one of the parameters is a primitive type (<code>int</code>, <code>double</code>, 
	 * etc.) you can use the appropriate "wrapper" class for its value (as in, 
	 * <code>new Integer(2)</code> instead of just 2).
	 * 
	 * @param target Object whose method will be called
	 * @param methodName Name of method to be called
	 * @param arg1 Object to pass as the first parameter. 
	 * @param arg2 Object to pass as the second parameter. 
	 * @param arg3 Object to pass as the third parameter. 
	 */
	public void setTarget(Object target, String methodName, Object arg1, Object arg2, Object arg3) { 
		setTarget(target, methodName, new Object[]{arg1, arg2, arg3}); 
	}
	
	/**
	 * Set this timer to pause for <code>startDelay</code> before it first 
	 * fires, and then fire then and every <code>delay</code> seconds
	 * after that. If <code>delay</code> is less than or equal to zero, the
	 * timer will fire only once, after <code>startDelay</code>. If the timer
	 * is currently running, it will try to adjust its schedule so that its
	 * next firing will be as close as possible to <code>delay</code> seconds
	 * after when it last fired. If it is stopped, this method will <b>not</b>
	 * start it.
	 * 
	 * @param startDelay Number of seconds to wait before first firing
	 * @param delay Number of seconds to wait in between firing
	 */
	public void setDelay(double startDelay, double delay) {
		if(timer.isRunning()) {
			timer.stop();
			internalSetDelay(lastFireTime < 0 ? delay : delay - .001 * (System.currentTimeMillis() - lastFireTime), delay);
			timer.start();
			timer.setInitialDelay((int)(startDelay * 1000 + .5));
		} else {
			internalSetDelay(startDelay, delay);
		}
	}
	
	private void internalSetDelay(double startDelay, double delay) {
		if(delay > 0) {
			timer.setRepeats(true);
			timer.setDelay((int)(delay * 1000 + .5));
		} else {
			timer.setRepeats(false);
		}
		
		if(startDelay < 0) startDelay = 0;
		timer.setInitialDelay((int)(startDelay * 1000 + .5));
	}
	
	/**
	 * Start the timer, using whatever delay values have been set. The default is
	 * to fire every second, starting one second after the timer is started. If the 
	 * timer was currently running, it will restart, pausing for the initial delay 
	 * amount all over again.
	 */
	public void start() {
		timer.stop();
		internalStart();
	}
	
	private void internalStart() {
		timer.start();
		activeTimers.add(this);
		
		startTime = System.currentTimeMillis();
		lastFireTime = initialFireTime = -1;
	}
	
	/**
	 * Start the timer. It will fire first after <code>startDelay</code> seconds,
	 * then every <code>delay</code> seconds after that. If <code>delay</code> is
	 * less than or equal to zero, the timer will only fire once, and then will 
	 * stop running. If the timer was currently running, it will restart, pausing
	 * for the initial delay amount all over again.
	 * 
	 * @param startDelay Number of seconds to delay before first firing
	 * @param delay Number of seconds to delay between subsequent firings
	 */
	public void start(double startDelay, double delay) {
		if(timer.isRunning()) timer.stop();
		
		internalSetDelay(startDelay, delay);
		
		internalStart();
	}
	
	/**
	 * Return <code>true</code> if the timer is currently running, <code>false<code>
	 * otherwise.
	 * @return whether the timer is running
	 */
	public boolean isRunning() {
		return timer.isRunning();
	}
	
	/**
	 * Stop the timer. This method has no effect if the timer is not running.
	 */
	public void stop() {
		activeTimers.remove(this);
		timer.stop();
	}
	
	/**
	 * Get the timer's start delay.
	 * @return the start delay
	 */
	public double getStartDelay() {
		return .001 * timer.getInitialDelay();
	}
	
	/**
	 * Get the timer's delay. This method will return 0 if the timer was set
	 * to fire only once.
	 * @return the delay
	 */
	public double getDelay() {
		return (timer.isRepeats() ? .001 * timer.getDelay() : 0);
	}
	
	/**
	 * Return the number of seconds since this timer was started. This will report
	 * time since the <b>most recent</b> call to <code>start()</code>. If <code>start()</code>
	 * has never been called, this method will return 0.
	 * @return Time since this timer was started.
	 */
	public double getTimeSinceStart() {
		return (startTime < 0 ? 0 : .001 * (System.currentTimeMillis() - startTime));
	}
	
	/**
	 * Return the number of seconds since this timer's initial firing. If the timer
	 * is currently performing its initial firing, or if it has not fired since
	 * <code>start()</code> was last called, then 0 will be returned.
	 * @return Time since initial firing
	 */
	public double getTimeSinceInitialFiring() {
		return initialFireTime < 0 ? 0 : .001 * (System.currentTimeMillis() - initialFireTime);
	}
	
	/**
	 * Return the number of seconds since this timer's previous firing. If the timer
	 * is currently performing its initial firing, or if it has not fired since
	 * <code>start()</code> was last called, then 0 will be returned.
	 * @return Time since previous firing
	 */
	public double getTimeSincePreviousFiring() {
		return lastFireTime < 0 ? 0 : .001 * (System.currentTimeMillis() - lastFireTime);
	}
	
	/**
	 * Check whether this is the initial firing of this timer. This method will 
	 * return true if the timer has not yet completed a firing since it was last
	 * <code>start()</code>ed.
	 * @return Whether this is the initial firing
	 */
	public boolean isInitialFiring() {
		return initialFireTime < 0;
	}
	
	/**
	 * Manually fire the timer. This will call the target method, just as if the
	 * timer had fired normally.
	 */
	public void fire() {
		try {
			if(method != null) method.invoke(target, args);
		} catch (IllegalArgumentException e) {
		} catch (IllegalAccessException e) {
		} catch (InvocationTargetException e) {
			e.getCause().printStackTrace();
		} finally {
			lastFireTime = System.currentTimeMillis();
			if(initialFireTime < 0) initialFireTime = lastFireTime;
			if(!timer.isRepeats()) activeTimers.remove(this);
		}
	}
}
