package objectdraw;

import java.awt.Color;
import java.awt.Shape;
import java.awt.geom.RoundRectangle2D;

/**
 * A <code>FilledRoundedRect</code> is a rectangle with rounded corners, filled in with some
 * color. Like other shapes that implement the <code>Drawable2DInterface</code>, its shape is 
 * defined by a top left corner, plus a width and height. You can also control the size of the
 * arcs that form the corners. 
 * @author Russell Zahniser (russell@zahniser.net)
 *
 */
public class FilledRoundedRect extends ObjectDrawRectangularShape {
	private double arcWidth, arcHeight;
	
	/**
	 * Construct a rounded rectangle starting at (<code>x</code>, <code>y</code>) at its
	 * top left corner, and extending the given <code>width</code> and <code>height</code> from
	 * there, filled in with the given <code>color</code>. Its corners will be rounded by arcs
	 * that have a total width of <code>arcWidth</code> and a total height of <code>arcHeight</code>.
	 * So, the straight parts of the edges have lengths <code>width - arcWidth</code> and
	 * <code>height - arcHeight</code>.
	 * @param x x coordinate of the top left corner
	 * @param y y coordinate of the top left corner
	 * @param width width of the oval
	 * @param height height of the oval
	 * @param color <code>Color</code> to draw the oval in
	 * @param canvas <code>DrawingCanvas</code> on which to place the oval
	 */
	public FilledRoundedRect(double x, double y, double width, double height, double arcWidth, double arcHeight, Color color, DrawingCanvas canvas) {
		super(x, y, width, height, true, color, canvas);
		this.arcWidth = arcWidth;
		this.arcHeight = arcHeight;
		ready();
	}
	
	/**
	 * Construct a rounded rectangle starting at (<code>x</code>, <code>y</code>) at its
	 * top left corner, and extending the given <code>width</code> and <code>height</code> from
	 * there, filled in with black. Its corners will be rounded by arcs
	 * that have a total width of <code>arcWidth</code> and a total height of <code>arcHeight</code>.
	 * So, the straight parts of the edges have lengths <code>width - arcWidth</code> and
	 * <code>height - arcHeight</code>.
	 * @param x x coordinate of the top left corner
	 * @param y y coordinate of the top left corner
	 * @param width width of the oval
	 * @param height height of the oval
	 * @param canvas <code>DrawingCanvas</code> on which to place the oval
	 */
	public FilledRoundedRect(double x, double y, double width, double height, double arcWidth, double arcHeight, DrawingCanvas canvas) {
		this(x, y, width, height, arcWidth, arcHeight, null, canvas);
	}
	
	/**
	 * Construct a rounded rectangle starting at <code>point</code> at its
	 * top left corner, and extending the given <code>width</code> and <code>height</code> from
	 * there, filled in with the given <code>color</code>. Its corners will be rounded by arcs
	 * that have a total width of <code>arcWidth</code> and a total height of <code>arcHeight</code>.
	 * So, the straight parts of the edges have lengths <code>width - arcWidth</code> and
	 * <code>height - arcHeight</code>.
	 * @param point <code>Location</code> of the top left corner
	 * @param width width of the oval
	 * @param height height of the oval
	 * @param color <code>Color</code> to draw the oval in
	 * @param canvas <code>DrawingCanvas</code> on which to place the oval
	 */
	public FilledRoundedRect(Location point, double width, double height, double arcWidth, double arcHeight, Color color, DrawingCanvas canvas) {
		this(point.getDoubleX(), point.getDoubleY(), width, height, arcWidth, arcHeight, color, canvas);
	}
	
	/**
	 * Construct a rounded rectangle starting at <code>point</code> at its
	 * top left corner, and extending the given <code>width</code> and <code>height</code> from
	 * there, filled in with black. Its corners will be rounded by arcs
	 * that have a total width of <code>arcWidth</code> and a total height of <code>arcHeight</code>.
	 * So, the straight parts of the edges have lengths <code>width - arcWidth</code> and
	 * <code>height - arcHeight</code>.
	 * @param point <code>Location</code> of the top left corner
	 * @param width width of the oval
	 * @param height height of the oval
	 * @param canvas <code>DrawingCanvas</code> on which to place the oval
	 */
	public FilledRoundedRect(Location point, double width, double height, double arcWidth, double arcHeight, DrawingCanvas canvas) {
		this(point, width, height, arcWidth, arcHeight, null, canvas);
	}
	
	/**
	 * Construct a rounded rectangle stretching from <code>corner1</code> to
	 * <code>corner2</code>, filled in with the given <code>color</code>. Its corners will be 
	 * rounded by arcs that have a total width of <code>arcWidth</code> and a total height of 
	 * <code>arcHeight</code>. So, the straight parts of the edges have lengths 
	 * <code>width - arcWidth</code> and <code>height - arcHeight</code>.
	 * @param corner1 <code>Location</code> of one corner
	 * @param corner2 <code>Location</code> of the opposite corner
	 * @param color <code>Color</code> to draw the oval in
	 * @param canvas <code>DrawingCanvas</code> on which to place the oval
	 */
	public FilledRoundedRect(Location corner1, Location corner2, double arcWidth, double arcHeight, Color color, DrawingCanvas canvas) {
		this(corner1, corner2.getDoubleX() - corner1.getDoubleX(), corner2.getDoubleY() - corner1.getDoubleY(), arcWidth, arcHeight, color, canvas);
	}
	
	/**
	 * Construct a rounded rectangle stretching from <code>corner1</code> to
	 * <code>corner2</code>, filled in with black. Its corners will be 
	 * rounded by arcs that have a total width of <code>arcWidth</code> and a total height of 
	 * <code>arcHeight</code>. So, the straight parts of the edges have lengths 
	 * <code>width - arcWidth</code> and <code>height - arcHeight</code>.
	 * @param corner1 <code>Location</code> of one corner
	 * @param corner2 <code>Location</code> of the opposite corner
	 * @param canvas <code>DrawingCanvas</code> on which to place the oval
	 */
	public FilledRoundedRect(Location corner1, Location corner2, double arcWidth, double arcHeight, DrawingCanvas canvas) {
		this(corner1, corner2, arcWidth, arcHeight, null, canvas);
	}
	
	/**
	 * Return the height of the arcs of this rounded rectangle, in <code>int</code> precision.
	 * Since each corner is a quarter oval, the height of each corner is actually half this.
	 * @return the arc height of this rounded rectangle, in <code>int</code> precision.
	 */
	public int getArcHeight() {
		return (int)(arcHeight + .5);
	}
	
	/**
	 * Return the height of the arcs of this rounded rectangle, in <code>double</code> precision.
	 * Since each corner is a quarter oval, the height of each corner is actually half this.
	 * @return the arc height of this rounded rectangle, in <code>double</code> precision.
	 */
	public double getDoubleArcHeight() {
		return arcHeight;
	}
	
	/**
	 * Return the width of the arcs of this rounded rectangle, in <code>int</code> precision.
	 * Since each corner is a quarter oval, the width of each corner is actually half this.
	 * @return the arc width of this rounded rectangle, in <code>int</code> precision.
	 */
	public int getArcWidth() {
		return (int)(arcWidth + .5);
	}
	
	/**
	 * Return the width of the arcs of this rounded rectangle, in <code>double</code> precision.
	 * Since each corner is a quarter oval, the width of each corner is actually half this.
	 * @return the arc width of this rounded rectangle, in <code>double</code> precision.
	 */
	public double getDoubleArcWidth() {
		return arcWidth;
	}
	
	/**
	 * Set the width of the arcs of this rounded rectangle.
	 * Since each corner is a quarter oval, the width of each corner is actually half this.
	 * @param a new arc width
	 */
	public void setArcWidth(double a) {
		arcWidth = a;
		update();
	}
	
	/**
	 * Set the height of the arcs of this rounded rectangle.
	 * Since each corner is a quarter oval, the height of each corner is actually half this.
	 * @param a new arc height
	 */
	public void setArcHeight(double a) {
		arcHeight = a;
		update();
	}

	/**
	 * @deprecated for internal use only
	 */
	public Shape makeShape() {
		return new RoundRectangle2D.Double(getDoubleX(), getDoubleY(), 
				getDoubleWidth(), getDoubleHeight(), arcWidth, arcHeight);
	}
	
	/* * * * Methods from Resizable2DInterface * * * */
	
	/**
	 * Set the width and height of this object. Its upper left corner will stay fixed in place,
	 * and the opposite corner will move to make it the new size.
	 * @param width new width
	 * @param height new height
	 */
	public void setSize(double width, double height) { super.setSize(width, height); }
	/**
	 * Set the width of this object, leaving its left edge in place and stretching the right.
	 * @param w new width
	 */
	public void setWidth(double w) { super.setWidth(w); }
	/**
	 * Set the height of this object, leaving its top edge in place and stretching the bottom.
	 * @param h new height
	 */
	public void setHeight(double h) { super.setHeight(h); }
	
	/* * * * Methods from LocatableInterface * * * */
	
	/**
	 * Get this object's x coordinate.
	 * @return this object's x coordinate
	 */
	public int getX() { return super.getX(); }
	/**
	 * Get this object's y coordinate.
	 * @return this object's y coordinate
	 */
	public int getY() { return super.getY(); }
	/**
	 * Get this object's x coordinate, in double precision.
	 * @return this object's x coordinate, in double precision
	 */
	public double getDoubleX() { return super.getDoubleX(); }
	/**
	 * Get this object's y coordinate, in double precision.
	 * @return this object's y coordinate, in double precision
	 */
	public double getDoubleY() { return super.getDoubleY(); }
	/**
	 * Get this object's location. Translating this <code>Location</code> object
	 * will have the same effect as moving the object. So, the following two lines
	 * of code are equivalent:
	 * <br><br><code>
	 * object.move(10, 20);<br>
	 * object.getLocation().translate(10, 20);
	 * </code><br><br>If you want to create a <code>Location</code> offset by some
	 * amount from the <code>Location</code> of this object, without actually
	 * moving this object, use the <code>offset()</code> method in <code>Location</code>,
	 * which creates a new <code>Location</code>.
	 * @return the <code>Location</code> of this object.
	 */
	public Location getLocation() { return super.getLocation(); }

	/* * * * Methods from Drawable2DInterface * * * */

	/**
	 * Return the height of the rectangle bounding this object.
	 * @return the height of the rectangle bounding this object.
	 */
	public int getWidth() { return super.getWidth(); }
	/**
	 * Return the width of the rectangle bounding this object.
	 * @return the width of the rectangle bounding this object.
	 */
	public int getHeight() { return super.getHeight(); }
	/**
	 * Return the height of the rectangle bounding this object, in double precision.
	 * @return the height of the rectangle bounding this object, in double precision.
	 */
	public double getDoubleWidth() { return super.getDoubleWidth(); }
	/**
	 * Return the height of the rectangle bounding this object, in double precision.
	 * @return the height of the rectangle bounding this object, in double precision.
	 */
	public double getDoubleHeight() { return super.getDoubleHeight(); }
	/**
	 * Return <code>true</code> if the rectangle bounding this object overlaps with the
	 * rectangle bounding another object; <code> false</code> otherwise.
	 * @param other another two dimensional object to check for overlapping with
	 * @return whether the bounding rectangles of the two objects overlap
	 */
	public boolean overlaps(Drawable2DInterface other) { return super.overlaps(other); }

	/* * * * Methods from DrawableInterface * * * */
	
	/**
	 * Make this object invisible. It stays in position on its canvas; it is simply not drawn.
	 * This method has no effect if the object was already hidden.
	 */
	public void hide() { super.hide(); }
	/**
	 * Make an object visible again after it had been hidden. This method has no effect if the
	 * object was already visible.
	 */
	public void show() { super.show(); }
	/**
	 * Return <code>true</code> if this object has been rendered invisible with a call to its
	 * <code>hide()</code> method, <code>false</code> if it is still being drawn.
	 * @return whether this object is hidden
	 */
	public boolean isHidden() { return super.isHidden(); }

	/**
	 * Place this object on the specified canvas. The object will be removed from its current
	 * canvas, if it is on a canvas that is not the one it is being moved to. The object will
	 * be at the front of the canvas, as if it had just been drawn. If the given canvas is 
	 * <code>null</code> the object is simply removed from its current canvas.
	 * @param c <code>DrawingCanvas</code> to place this object on
	 */
	public void addToCanvas(DrawingCanvas c) { super.addToCanvas(c); }
	/**
	 * Remove this object from its canvas.
	 */
	public void removeFromCanvas() { super.removeFromCanvas(); }
	/**
	 * Return the canvas that this object is on. If it has been removed from its canvas, this
	 * method will return <code>null</code>.
	 * @return the <code>DrawingCanvas</code> that this object is on.
	 */
	public DrawingCanvas getCanvas() { return super.getCanvas(); }

	/**
	 * Move the reference point of this object to the given location. For a Drawable2D object,
	 * the reference point is the top left corner; for a line, it is the start point.
	 * @param point <code>Location</code> to move to
	 */
	public void moveTo(Location point) { super.moveTo(point); }
	/**
	 * Move the reference point of this object to the given location. For a Drawable2D object,
	 * the reference point is the top left corner; for a line, it is the start point.
	 * @param x x coordinate to move to
	 * @param y y coordinate to move to
	 */
	public void moveTo(double x, double y) { super.moveTo(x, y); }
	/**
	 * Shift (translate) this object left or right by an amount <code>dx</code>, and up or
	 * down by an amount <code>dy</code>. A negative <code>dx</code> indicates a shift to the
	 * left; a negative <code>dy</code> indicates a shift up
	 * @param dx amount of x translation
	 * @param dy amount of y translation
	 */
	public void move(double dx, double dy) { super.move(dx, dy); }

	/**
	 * Return the color of this object
	 * @return the <code>Color</code> of this object
	 */
	public Color getColor() { return super.getColor(); }
	/**
	 * Change the color of this object.
	 * @param c new <code>Color</code> for this object
	 */
	public void setColor(Color c) { super.setColor(c); }

	/**
	 * Move this object one step forward in the draw order, causing it to be drawn on top of
	 * the objects that are behind it.
	 */
	public void sendForward() { super.sendForward(); }
	/**
	 * Move this object one step backward in the draw order, causing it to be drawn underneath
	 * the objects that are in front of it.
	 */
	public void sendBackward() { super.sendBackward(); }
	/**
	 * Move this object to the very front of the draw order, causing it to be drawn on top of
	 * all the other objects.
	 */
	public void sendToFront() { super.sendToFront(); }
	/**
	 * Move this object to the very back of the draw order, causing it to be drawwn underneath
	 * all the other objects.
	 */
	public void sendToBack() { super.sendToBack(); }

	/**
	 * Return <code>true</code> if the given location is inside this object, <code>false</code>
	 * otherwise. Because <code>Line</code>s and <code>FramedArc</code>s have no "inside", a
	 * point is considered to be contained in one of them as long as it is within a few pixels
	 * of the line.
	 * @param point <code>Location</code> to test
	 * @return whether the given location is inside this shape
	 */
	public boolean contains(Location point) { return super.contains(point); }

	/**
	 * Return a string describing how this object might be constructed.
	 */
	public String toString() {
		return toString(", " + Text.formatDecimal(arcWidth) + ", " + Text.formatDecimal(arcHeight));
	}
}
