package objectdraw;

import java.awt.Color;
import java.awt.BasicStroke;
import java.awt.Shape;
import java.awt.geom.Arc2D;

/**
 * A <code>FramedArc</code> is a curved line that forms some portion of an oval. Like other
 * shapes that implement the <code>Drawable2DInterface</code>, its shape is bounded by a top
 * left corner, plus a width and height; these specify the size of the oval it is cut out of,
 * not the space taken up by the arc itself. In addition, it has an angle in degrees specifying
 * at what angle the arc starts, and another saying how many degrees from there the arc extends.
 * @author Russell Zahniser (russell@zahniser.net)
 *
 */
public class FramedArc extends ObjectDrawRectangularShape implements DrawableStrokeInterface {
	private double startAngle, arcAngle;
	private BasicStroke stroke;
	
	/**
	 * Construct a <code>FramedArc</code> whose bounding rectangle has its top left corner at 
	 * (<code>x</code>, <code>y</code>) and extends the given <code>width</code> and 
	 * <code>height</code> from there. The arc will start at <code>startAngle</code>, given in 
	 * degrees in the usual angle system, so that 0 is east, <code>Math.PI / 2</code> is north, 
	 * and so on; it will extend <code>arcAngle</code> degrees from there. The sign of 
	 * <code>arcAngle</code> determines which way the arc goes from <code>startAngle</code>; 
	 * a negative sign means that the arc extends clockwise, whereas a positive sign means 
	 * that it extends counterclockwise.
	 * @param x x coordinate of top left corner of bounding rectangle
	 * @param y y coordinate of top left corner of bounding rectangle
	 * @param width width of bounding rectangle
	 * @param height height of bounding rectangle
	 * @param startAngle start angle, in degrees
	 * @param arcAngle arc angle, in degrees
	 * @param color <code>Color</code> of the arc
	 * @param canvas <code>DrawingCanvas</code> on which to place the arc
	 */
	public FramedArc(double x, double y, double width, double height, double startAngle, double arcAngle, Color color, DrawingCanvas canvas) {
		super(x, y, width, height, false, color, canvas);
		this.startAngle = startAngle;
		this.arcAngle = arcAngle;
		stroke = DEFAULT_STROKE;
		ready();
	}
	
	/**
	 * Construct a <code>FramedArc</code> whose bounding rectangle has its top left corner at 
	 * (<code>x</code>, <code>y</code>) and extends the given <code>width</code> and 
	 * <code>height</code> from there. The arc will start at <code>startAngle</code>, given in 
	 * degrees in the usual angle system, so that 0 is east, <code>Math.PI / 2</code> is north, 
	 * and so on; it will extend <code>arcAngle</code> degrees from there. The sign of 
	 * <code>arcAngle</code> determines which way the arc goes from <code>startAngle</code>; 
	 * a negative sign means that the arc extends clockwise, whereas a positive sign means 
	 * that it extends counterclockwise.
	 * @param x x coordinate of top left corner of bounding rectangle
	 * @param y y coordinate of top left corner of bounding rectangle
	 * @param width width of bounding rectangle
	 * @param height height of bounding rectangle
	 * @param startAngle start angle, in degrees
	 * @param arcAngle arc angle, in degrees
	 * @param canvas <code>DrawingCanvas</code> on which to place the arc
	 */
	public FramedArc(double x, double y, double width, double height, double startAngle, double arcAngle, DrawingCanvas canvas) {
		this(x, y, width, height, startAngle, arcAngle, null, canvas);
	}
	
	/**
	 * Construct a <code>FramedArc</code> whose bounding rectangle has its top left corner at 
	 * <code>point</code> and extends the given <code>width</code> and 
	 * <code>height</code> from there. The arc will start at <code>startAngle</code>, given in 
	 * degrees in the usual angle system, so that 0 is east, <code>Math.PI / 2</code> is north, 
	 * and so on; it will extend <code>arcAngle</code> degrees from there. The sign of 
	 * <code>arcAngle</code> determines which way the arc goes from <code>startAngle</code>; 
	 * a negative sign means that the arc extends clockwise, whereas a positive sign means 
	 * that it extends counterclockwise.
	 * @param point <code>Location</code> of the top left corner of the bounding rectangle
	 * @param width width of bounding rectangle
	 * @param height height of bounding rectangle
	 * @param startAngle start angle, in degrees
	 * @param arcAngle arc angle, in degrees
	 * @param color <code>Color</code> of the arc
	 * @param canvas <code>DrawingCanvas</code> on which to place the arc
	 */
	public FramedArc(Location point, double width, double height, double startAngle, double arcAngle, Color color, DrawingCanvas canvas) {
		this(point.getDoubleX(), point.getDoubleY(), width, height, startAngle, arcAngle, color, canvas);
	}
	
	/**
	 * Construct a <code>FramedArc</code> whose bounding rectangle has its top left corner at 
	 * <code>point</code> and extends the given <code>width</code> and 
	 * <code>height</code> from there. The arc will start at <code>startAngle</code>, given in 
	 * degrees in the usual angle system, so that 0 is east, <code>Math.PI / 2</code> is north, 
	 * and so on; it will extend <code>arcAngle</code> degrees from there. The sign of 
	 * <code>arcAngle</code> determines which way the arc goes from <code>startAngle</code>; 
	 * a negative sign means that the arc extends clockwise, whereas a positive sign means 
	 * that it extends counterclockwise.
	 * @param point <code>Location</code> of the top left corner of the bounding rectangle
	 * @param width width of bounding rectangle
	 * @param height height of bounding rectangle
	 * @param startAngle start angle, in degrees
	 * @param arcAngle arc angle, in degrees
	 * @param canvas <code>DrawingCanvas</code> on which to place the arc
	 */
	public FramedArc(Location point, double width, double height, double startAngle, double arcAngle, DrawingCanvas canvas) {
		this(point, width, height, startAngle, arcAngle, null, canvas);
	}
	
	/**
	 * Construct a <code>FramedArc</code> whose bounding rectangle stretches between <code>corner1</code> 
	 * and <code>corner2</code>. The arc will start at <code>startAngle</code>, given in 
	 * degrees in the usual angle system, so that 0 is east, <code>Math.PI / 2</code> is north, 
	 * and so on; it will extend <code>arcAngle</code> degrees from there. The sign of 
	 * <code>arcAngle</code> determines which way the arc goes from <code>startAngle</code>; 
	 * a negative sign means that the arc extends clockwise, whereas a positive sign means 
	 * that it extends counterclockwise.
	 * @param corner1 <code>Location</code> of one corner of the bounding rectangle
	 * @param corner2 <code>Location</code> of the opposite corner of the bounding rectangle
	 * @param startAngle start angle, in degrees
	 * @param arcAngle arc angle, in degrees
	 * @param color <code>Color</code> of the arc
	 * @param canvas <code>DrawingCanvas</code> on which to place the arc
	 */
	public FramedArc(Location corner1, Location corner2, double startAngle, double arcAngle, Color color, DrawingCanvas canvas) {
		this(corner1, corner2.getDoubleX() - corner1.getDoubleX(), corner2.getDoubleY() - corner1.getDoubleY(), startAngle, arcAngle, color, canvas);
	}
	
	/**
	 * Construct a <code>FramedArc</code> whose bounding rectangle stretches between <code>corner1</code> 
	 * and <code>corner2</code>. The arc will start at <code>startAngle</code>, given in 
	 * degrees in the usual angle system, so that 0 is east, <code>Math.PI / 2</code> is north, 
	 * and so on; it will extend <code>arcAngle</code> degrees from there. The sign of 
	 * <code>arcAngle</code> determines which way the arc goes from <code>startAngle</code>; 
	 * a negative sign means that the arc extends clockwise, whereas a positive sign means 
	 * that it extends counterclockwise.
	 * @param corner1 <code>Location</code> of one corner of the bounding rectangle
	 * @param corner2 <code>Location</code> of the opposite corner of the bounding rectangle
	 * @param startAngle start angle, in degrees
	 * @param arcAngle arc angle, in degrees
	 * @param canvas <code>DrawingCanvas</code> on which to place the arc
	 */
	public FramedArc(Location corner1, Location corner2, double startAngle, double arcAngle, DrawingCanvas canvas) {
		this(corner1, corner2, startAngle, arcAngle, null, canvas);
	}
	
	/* * * * Arc angle methods * * * */
	
	/**
	 * Return the angular extent of the arc, in degrees. A negative angle means that the arc
	 * goes clockwise from the start angle; a positive angle means that it extends 
	 * counterclockwise. The end angle is always at <code>(startAngle + arcAngle)</code>.
	 * @return the arc angle
	 */
	public double getArcAngle() {
		return arcAngle;
	}
	
	/**
	 * Get the angle at which the arc starts, in degrees. Angles are positioned starting at 0
	 * at the x axis and going counterclockwise from there, as is customary in math.
	 * @return the start angle
	 */
	public double getStartAngle() {
		return startAngle;
	}
	
	/**
	 * Set the start angle of the arc, in degrees. Angles are positioned starting at 0
	 * at the x axis and going counterclockwise from there, as is customary in math.
	 * @param a new start angle
	 */
	public void setStartAngle(double a) {
		startAngle = a;
		update();
	}
	
	/**
	 * Set the angular length of the arc, in degrees. A negative angle will cause the arc to 
	 * extend clockwise from the start angle; a positive angle will make it extend 
	 * counterclockwise.
	 * @param a new arc angle
	 */
	public void setArcAngle(double a) {
		arcAngle = a;
		update();
	}

	/**
	 * @deprecated for internal use only
	 */
	public Shape makeShape() {
		return new Arc2D.Double(getDoubleX(), getDoubleY(), 
				getDoubleWidth(), getDoubleHeight(), startAngle, arcAngle, Arc2D.OPEN);
	}
	
	/* * * * Methods from DrawableStrokeInterface * * * */

	/**
	 * Return the width of this object's lines.
	 * @return the width of this object's lines
	 */
	public double getLineWidth() {
		return stroke.getLineWidth();
	}

	/**
	 * Return the stroke used for this object's lines.
	 * @return the stroke used for this object's lines
	 */
	public BasicStroke getStroke() {
		return stroke;
	}

	/**
	 * Set the width of this object's lines
	 * @param width the new width of this object's lines
	 */
	public void setLineWidth(double width) {
		stroke = new BasicStroke((float)width, stroke.getEndCap(), stroke.getLineJoin(), 
					stroke.getMiterLimit(), stroke.getDashArray(), stroke.getDashPhase());
		update();
	}

	/**
	 * Set the stroke used for this object's lines
	 * @param stroke the new stroke for this object's lines
	 */
	public void setStroke(BasicStroke stroke) {
		this.stroke = stroke;
		update();
	}
	
	/* * * * Methods from Resizable2DInterface * * * */
	
	/**
	 * Set the width and height of this object. Its upper left corner will stay fixed in place,
	 * and the opposite corner will move to make it the new size.
	 * @param width new width
	 * @param height new height
	 */
	public void setSize(double width, double height) { super.setSize(width, height); }
	/**
	 * Set the width of this object, leaving its left edge in place and stretching the right.
	 * @param w new width
	 */
	public void setWidth(double w) { super.setWidth(w); }
	/**
	 * Set the height of this object, leaving its top edge in place and stretching the bottom.
	 * @param h new height
	 */
	public void setHeight(double h) { super.setHeight(h); }
	
	/* * * * Methods from LocatableInterface * * * */
	
	/**
	 * Get this object's x coordinate.
	 * @return this object's x coordinate
	 */
	public int getX() { return super.getX(); }
	/**
	 * Get this object's y coordinate.
	 * @return this object's y coordinate
	 */
	public int getY() { return super.getY(); }
	/**
	 * Get this object's x coordinate, in double precision.
	 * @return this object's x coordinate, in double precision
	 */
	public double getDoubleX() { return super.getDoubleX(); }
	/**
	 * Get this object's y coordinate, in double precision.
	 * @return this object's y coordinate, in double precision
	 */
	public double getDoubleY() { return super.getDoubleY(); }
	/**
	 * Get this object's location. Translating this <code>Location</code> object
	 * will have the same effect as moving the object. So, the following two lines
	 * of code are equivalent:
	 * <br><br><code>
	 * object.move(10, 20);<br>
	 * object.getLocation().translate(10, 20);
	 * </code><br><br>If you want to create a <code>Location</code> offset by some
	 * amount from the <code>Location</code> of this object, without actually
	 * moving this object, use the <code>offset()</code> method in <code>Location</code>,
	 * which creates a new <code>Location</code>.
	 * @return the <code>Location</code> of this object.
	 */
	public Location getLocation() { return super.getLocation(); }

	/* * * * Methods from Drawable2DInterface * * * */

	/**
	 * Return the height of the rectangle bounding this object.
	 * @return the height of the rectangle bounding this object.
	 */
	public int getWidth() { return super.getWidth(); }
	/**
	 * Return the width of the rectangle bounding this object.
	 * @return the width of the rectangle bounding this object.
	 */
	public int getHeight() { return super.getHeight(); }
	/**
	 * Return the height of the rectangle bounding this object, in double precision.
	 * @return the height of the rectangle bounding this object, in double precision.
	 */
	public double getDoubleWidth() { return super.getDoubleWidth(); }
	/**
	 * Return the height of the rectangle bounding this object, in double precision.
	 * @return the height of the rectangle bounding this object, in double precision.
	 */
	public double getDoubleHeight() { return super.getDoubleHeight(); }
	/**
	 * Return <code>true</code> if the rectangle bounding this object overlaps with the
	 * rectangle bounding another object; <code> false</code> otherwise.
	 * @param other another two dimensional object to check for overlapping with
	 * @return whether the bounding rectangles of the two objects overlap
	 */
	public boolean overlaps(Drawable2DInterface other) { return super.overlaps(other); }

	/* * * * Methods from DrawableInterface * * * */
	
	/**
	 * Make this object invisible. It stays in position on its canvas; it is simply not drawn.
	 * This method has no effect if the object was already hidden.
	 */
	public void hide() { super.hide(); }
	/**
	 * Make an object visible again after it had been hidden. This method has no effect if the
	 * object was already visible.
	 */
	public void show() { super.show(); }
	/**
	 * Return <code>true</code> if this object has been rendered invisible with a call to its
	 * <code>hide()</code> method, <code>false</code> if it is still being drawn.
	 * @return whether this object is hidden
	 */
	public boolean isHidden() { return super.isHidden(); }

	/**
	 * Place this object on the specified canvas. The object will be removed from its current
	 * canvas, if it is on a canvas that is not the one it is being moved to. The object will
	 * be at the front of the canvas, as if it had just been drawn. If the given canvas is 
	 * <code>null</code> the object is simply removed from its current canvas.
	 * @param c <code>DrawingCanvas</code> to place this object on
	 */
	public void addToCanvas(DrawingCanvas c) { super.addToCanvas(c); }
	/**
	 * Remove this object from its canvas.
	 */
	public void removeFromCanvas() { super.removeFromCanvas(); }
	/**
	 * Return the canvas that this object is on. If it has been removed from its canvas, this
	 * method will return <code>null</code>.
	 * @return the <code>DrawingCanvas</code> that this object is on.
	 */
	public DrawingCanvas getCanvas() { return super.getCanvas(); }

	/**
	 * Move the reference point of this object to the given location. For a Drawable2D object,
	 * the reference point is the top left corner; for a line, it is the start point.
	 * @param point <code>Location</code> to move to
	 */
	public void moveTo(Location point) { super.moveTo(point); }
	/**
	 * Move the reference point of this object to the given location. For a Drawable2D object,
	 * the reference point is the top left corner; for a line, it is the start point.
	 * @param x x coordinate to move to
	 * @param y y coordinate to move to
	 */
	public void moveTo(double x, double y) { super.moveTo(x, y); }
	/**
	 * Shift (translate) this object left or right by an amount <code>dx</code>, and up or
	 * down by an amount <code>dy</code>. A negative <code>dx</code> indicates a shift to the
	 * left; a negative <code>dy</code> indicates a shift up
	 * @param dx amount of x translation
	 * @param dy amount of y translation
	 */
	public void move(double dx, double dy) { super.move(dx, dy); }

	/**
	 * Return the color of this object
	 * @return the <code>Color</code> of this object
	 */
	public Color getColor() { return super.getColor(); }
	/**
	 * Change the color of this object.
	 * @param c new <code>Color</code> for this object
	 */
	public void setColor(Color c) { super.setColor(c); }

	/**
	 * Move this object one step forward in the draw order, causing it to be drawn on top of
	 * the objects that are behind it.
	 */
	public void sendForward() { super.sendForward(); }
	/**
	 * Move this object one step backward in the draw order, causing it to be drawn underneath
	 * the objects that are in front of it.
	 */
	public void sendBackward() { super.sendBackward(); }
	/**
	 * Move this object to the very front of the draw order, causing it to be drawn on top of
	 * all the other objects.
	 */
	public void sendToFront() { super.sendToFront(); }
	/**
	 * Move this object to the very back of the draw order, causing it to be drawn underneath
	 * all the other objects.
	 */
	public void sendToBack() { super.sendToBack(); }

	/**
	 * Return <code>true</code> if the given location is inside this object, <code>false</code>
	 * otherwise. Because <code>Line</code>s and <code>FramedArc</code>s have no "inside", a
	 * point is considered to be contained in one of them as long as it is within a few pixels
	 * of the line.
	 * @param point <code>Location</code> to test
	 * @return whether the given location is inside this shape
	 */
	public boolean contains(Location point) {
		double w = getDoubleWidth(), h = getDoubleHeight();
		double c = Math.PI * (3*(w + h) + Math.sqrt((3*w + h) * (w + 3*h))) / 2;
		double err = Math.PI / c;
		double cx = getDoubleX() + w / 2, cy = getDoubleY() + h / 2;
		double x = point.getDoubleX(), y = point.getDoubleY();
		
		double sa = startAngle * Math.PI / 180, da = arcAngle * Math.PI / 180;
		while(da > err) {
			double ma = sa + da / 2;
			double dx = cx + Math.cos(-ma) * w / 2 - x, dy = cy + Math.sin(-ma) * h/2 - y;
			double d = Math.sqrt(dx * dx + dy * dy);
			if(d < CLICK_SIZE) {
				return true;
			} else if(da * w * Math.sin(-ma) * dx - da * h * Math.cos(-ma) * dy < 0) {
				sa = ma;
			}
			da /= 2;
		}
		return false;
	}
	
	/**
	 * Return a string describing how this object might be constructed.
	 */
	public String toString() {
		return toString(", " + Text.formatDecimal(startAngle) + ", " + Text.formatDecimal(arcAngle));
	}
}