package objectdraw;

import java.util.HashMap;
import java.util.HashSet;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.awt.Component;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;

/**
 * <code>KeyInterpreter</code> is the class that processes key events and calls the appropriate
 * methods in <code>WindowController</code>. Additional key interpreters can be created to pipe
 * these events to other objects as well. For example, you may want an ActiveObject to be
 * controlled by keys, and you can do this by making a new <code>KeyInterpreter</code> that has
 * that object as its target and attaching it to the canvas.
 * <br><br>
 * There are three types of methods called when a key is pressed. Suppose that I press the space
 * bar, hold it down for a few seconds, and release it. When it is first pressed, the method
 * <code>onSpacePress()</code> of the target will be called, if it exists. While the key is
 * held down, periodically the method <code>onSpaceHold()</code> will be called. Finally, when
 * the key is released, the method <code>onSpaceRelease()</code> will be called.
 * <br><br>
 * In general, it is a good idea to not rely on receiving the <code>onSpaceHold()</code> event,
 * because different computers produce different sets of events when a key is held, with
 * different timings. Instead, use a variable to remember that the key is held, changing it in
 * the press and release methods.
 * <br><br>
 * There are two other method you may make use of to process key events. Sometimes you want to
 * know, not what key was pressed, but what text it typed. For example, you may want to have
 * someone type their name into the applet and have it appear on the screen. The method
 * <code>onKeyTyped(String text)</code> will be called when a key has been typed, and 
 * <code>text</code> will be whatever text that key produces, properly taking into account what
 * modifiers (such as the shift key) are down.
 * <br><br>
 * Another option for receiving key events is to implement the <code>onOtherPress(String name)</code>
 * method (and corresponding ones for release or hold). This method is called when a key was
 * pressed for which no other specific method was found to handle it. The <code>name</code> will
 * be the name of the key, such as "a" or "space". If the shift key is down, the name will be
 * capitalized. If the alt key (option, on a Mac) is down, this will be preceded by a "~" symbol.
 * If the control key is down, there will be a "^" before that. If the meta key (command, on a
 * Mac) is down, there will be a "&" symbol before that. So, a name of "^T" would mean that
 * the T key was pressed with shift and control held down. This method allows you to easily
 * react only to certain combinations of keys and modifiers.
 * 
 * @author Russell Zahniser (russell@zahniser.net)
 *
 */
public class KeyInterpreter implements KeyListener {
	//	Map of code to name
	private static HashMap codeNames;
	//	Map of name to code
	private static HashMap keyCodes = buildKeyCodes();
	//	code of this GizmoKeyCode
	private Object target;
	// Array of all keys currently pressed
	private HashSet keysDown = new HashSet();
	private Component canvas;
	
	/**
	 * Construct a <code>KeyInterpreter</code> that will listen for key events on the given
	 * <code>canvas</code> and relay them to the <code>target</code> object.
	 * @param target Object that will be notified of key events. It should have method such
	 * as <code>onSpacePress()</code> for the keys that it is interested in.
	 * @param canvas Canvas from which events will be captured.
	 */
	public KeyInterpreter(Object target, DrawingCanvas canvas) {
		this.canvas = (Component)canvas;
		this.canvas.addKeyListener(this);
		this.target = target;
	}
	
	/**
	 * Remove this <code>KeyInterpreter</code> from the canvas. It will no longer detect key events,
	 * unless you add it back to a canvas.
	 */
	public void removeFromCanvas() {
		if(canvas != null) {
			canvas.removeKeyListener(this);
			canvas = null;
		}
	}
	
	/**
	 * Add this <code>KeyInterpreter</code> to a canvas. It will automatically be removed from
	 * whatever canvas it was previously attached to.
	 * @param canvas <code>DrawingCanvas</code> to which this <code>KeyInterpreter</code>
	 * should listen.
	 */
	public void addToCanvas(DrawingCanvas canvas) {
		if(canvas != this.canvas) {
			removeFromCanvas();
			this.canvas = (Component)canvas;
			this.canvas.addKeyListener(this);
		}
	}
	
	/**
	 * Return the canvas this listener is attached to.
	 * @return the <code>DrawingCanvas</code> that this interpreter is attached to.
	 */
	public DrawingCanvas getCanvas() {
		return (DrawingCanvas)canvas;
	}

	/*
	 * In order to build our reference table, we parse the KeyEvent class
	 * and extract the name and value of any field that begins with "VK_".
	 */
	private static HashMap buildKeyCodes() {
		HashMap keyCodes = new HashMap();
		codeNames = new HashMap();
		try {
			Field[] fields = KeyEvent.class.getFields();
			for (int i = 0; i < fields.length; i++) {
				if ((fields[i].getModifiers() & Modifier.STATIC) != 0
					&& fields[i].getName().startsWith("VK_")) {
					String name =
						fields[i].getName().substring(3).toLowerCase();
					Integer value = (Integer) fields[i].get(null);
					keyCodes.put(name, value);
					codeNames.put(value, name);
				}
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
		return keyCodes;
	}

	private static Integer codeForKeyName(String key) {
		return (Integer) keyCodes.get(key.toLowerCase());
	}

	private static String nameForKeyCode(Integer code) {
		return (String) codeNames.get(code);
	}

	private static String uppercaseNameForKeyCode(Integer code) {
		return ((String) codeNames.get(code)).toUpperCase();
	}

	private static String capitalizedNameForKeyCode(Integer code) {
		String s = (String) codeNames.get(code);
		return s.substring(0, 1).toUpperCase().concat(s.substring(1));
	}
	
	private void keyEvent(KeyEvent event) {
		String type;
		Integer code = new Integer(event.getKeyCode());
				
		if(target == null) {
			return;
		}
		/*
		 * Some systems will continue to produce key pressed events as long as the key
		 * is down. We keep a list of what keys are down, and only call the "Down" method
		 * when the key was really just pressed down. Otherwise, we call the "Held" method.
		 * The "Held" method should probably not be implemented: you can't trust that all
		 * systems will keep giving key pressed events.
		 */
		if(event.getID() == KeyEvent.KEY_PRESSED) {
			if(keysDown.contains(code)) {
				type = "Hold";
			} else {
				type = "Press";
				keysDown.add(code);
			}
		} else if(event.getID() == KeyEvent.KEY_RELEASED) {
			type = "Release";
			keysDown.remove(code);
		} else {
			return;
		}
		
		String name = capitalizedNameForKeyCode(code);
		if(name == null) {
			return;
		}
		
		String methodName = "on" + name + type;
		
		ObjectDrawObject.deferUpdates();
		try {
			Method method = target.getClass().getMethod(methodName, null);
			method.invoke(target, null);
		} catch(InvocationTargetException e) {
			e.getCause().printStackTrace();
		} catch(Exception e) {
			try {
				Class[] c = new Class[] {String.class};
				
				String[] key = new String[1];
				if(event.isShiftDown()) {
					key[0] = capitalizedNameForKeyCode(code);
				} else {
					key[0] = nameForKeyCode(code);
				}
				if(event.isAltDown()) {
					key[0] = "~" + key[0];
				}
				if(event.isControlDown()) {
					key[0] = "^" + key[0];
				}
				if(event.isMetaDown()) {
					key[0] = "&" + key[0];
				}
				Method method = target.getClass().getMethod("onOther" + type, c);
				method.invoke(target, key);
			} catch(InvocationTargetException f) {
				e.getCause().printStackTrace();
			} catch(Exception f) {
			}
		} finally {
			ObjectDrawObject.runUpdates();
		}
	}
	
	/* * * KeyListener implementation * * */

	/**
	 * @deprecated for internal use only
	 */
	public void keyTyped(KeyEvent event) {
		if(target == null || event.isActionKey() || event.getKeyChar() == KeyEvent.CHAR_UNDEFINED || 
				(event.getKeyChar() != '\n' && Character.isISOControl(event.getKeyChar()))) return;
		
		ObjectDrawShape.deferUpdates();
		try {
			Class[] c = new Class[] {String.class};
			String[] key = new String[] {"" + event.getKeyChar()};
			Method method = target.getClass().getMethod("onKeyTyped", c);
			method.invoke(target, key);
		} catch(InvocationTargetException e) {
			e.getCause().printStackTrace();
		} catch(Exception f) {
		} finally {
			ObjectDrawObject.runUpdates();
		}
	}

	/**
	 * @deprecated for internal use only
	 */
	public void keyPressed(KeyEvent event) {
		keyEvent(event);
	}

	/**
	 * @deprecated for internal use only
	 */
	public void keyReleased(KeyEvent event) {
		keyEvent(event);
	}
}
