package objectdraw;

import java.awt.Color;
import java.awt.Shape;
import java.awt.BasicStroke;
import java.awt.geom.Line2D;

/**
 * A <code>Line</code> represents a line segment connecting two points on the screen. The end
 * points may be altered individually to move the line. It is also possible to set the color,
 * thickness, and stroke style of the line.
 * 
 * @author Russell Zahniser (russell@zahniser.net)
 */
public class Line extends ObjectDrawShape implements Drawable1DInterface, DrawableStrokeInterface {
	private Location end;
	private BasicStroke stroke;

	/**
	 * Create a <code>Line</code> from (<code>x1</code>, <code>y1</code>) to (<code>x2</code>,
	 * <code>y2</code>), in the given <code>color</code>.
	 * @param x1 x coordinate of the start point
	 * @param y1 y coordinate of the start point
	 * @param x2 x coordinate of the end point
	 * @param y2 y coordinate of the end point
	 * @param color <code>Color</code> of the line
	 * @param canvas <code>DrawingCanvas</code> to place the <code>Line</code>
	 */
	public Line(double x1, double y1, double x2, double y2, Color color, DrawingCanvas canvas) {
		super(new Location(x1, y1), false, color, canvas);

		end = new Location(x2, y2);
		end.depend(this);
		
		stroke = DEFAULT_STROKE;
		ready();
	}
	
	/**
	 * Create a <code>Line</code> from (<code>x1</code>, <code>y1</code>) to (<code>x2</code>,
	 * <code>y2</code>), colored black.
	 * @param x1 x coordinate of the start point
	 * @param y1 y coordinate of the start point
	 * @param x2 x coordinate of the end point
	 * @param y2 y coordinate of the end point
	 * @param canvas <code>DrawingCanvas</code> to place the <code>Line</code>
	 */
	public Line(double x1, double y1, double x2, double y2, DrawingCanvas canvas) {
		this(x1, y1, x2, y2, null, canvas);
	}
	
	/**
	 * Create a <code>Line</code> from <code>start</code> to <code>end</code>,
	 * in the given <code>color</code>.
	 * @param start <code>Location</code> of the start point
	 * @param end <code>Location</code> of the end point
	 * @param color <code>Color</code> of the line
	 * @param canvas <code>DrawingCanvas</code> to place the <code>Line</code>
	 */
	public Line(Location start, Location end, Color color, DrawingCanvas canvas) {
		this(start.getDoubleX(), start.getDoubleY(), end.getDoubleX(), end.getDoubleY(), color, canvas);
	}
	
	/**
	 * Create a <code>Line</code> from <code>start</code> to <code>end</code>,
	 * colored black.
	 * @param start <code>Location</code> of the start point
	 * @param end <code>Location</code> of the end point
	 * @param canvas <code>DrawingCanvas</code> to place the <code>Line</code>
	 */
	public Line(Location start, Location end, DrawingCanvas canvas) {
		this(start, end, null, canvas);
	}

	/**
	 * @deprecated for internal use only
	 */
	public Shape makeShape() {
		Location start = getStart();
		return new Line2D.Double(start.getDoubleX(), start.getDoubleY(), end.getDoubleX(), end.getDoubleY());
	}

	/* * * * Methods from DrawableStrokeInterface * * * */

	/**
	 * Return the width of this object's lines.
	 * @return the width of this object's lines
	 */
	public double getLineWidth() {
		return stroke.getLineWidth();
	}

	/**
	 * Return the stroke used for this object's lines.
	 * @return the stroke used for this object's lines
	 */
	public BasicStroke getStroke() {
		return stroke;
	}

	/**
	 * Set the width of this object's lines
	 * @param width the new width of this object's lines
	 */
	public void setLineWidth(double width) {
		stroke = new BasicStroke((float)width, stroke.getEndCap(), stroke.getLineJoin(), 
					stroke.getMiterLimit(), stroke.getDashArray(), stroke.getDashPhase());
		update();
	}

	/**
	 * Set the stroke used for this object's lines
	 * @param stroke the new stroke for this object's lines
	 */
	public void setStroke(BasicStroke stroke) {
		this.stroke = stroke;
		update();
	}
	
	/* * * * Methods from Drawable1DInterface * * * */
	
	/**
	 * Get the location of the start of the line. Translating this <code>Location</code> object
	 * will have the same effect as changing the start point. So, the following three lines
	 * of code are equivalent:
	 * <br><br><code>
	 * line.setStart(line.getStart().getX() + 10, line.getStart().getY() + 20);<br>
	 * line.setStart(line.getStart().offset(10, 20));<br>
	 * line.getStart().translate(10, 20);
	 * </code><br><br>If you want to create a <code>Location</code> offset by some
	 * amount from the start <code>Location</code> of this line, without actually
	 * moving this line, use the <code>offset()</code> method in <code>Location</code>,
	 * which creates a new <code>Location</code>.
	 * @return the <code>Location</code> of the start of the line.
	 */
	public Location getStart() {
		return getMyLocation();
	}
	/**
	 * Get the location of the end of the line. Translating this <code>Location</code> object
	 * will have the same effect as changing the end point. So, the following three lines
	 * of code are equivalent:
	 * <br><br><code>
	 * line.setEnd(line.getEnd().getX() + 10, line.getEnd().getY() + 20);<br>
	 * line.setEnd(line.getEnd().offset(10, 20));<br>
	 * line.getEnd().translate(10, 20);
	 * </code><br><br>If you want to create a <code>Location</code> offset by some
	 * amount from the end <code>Location</code> of this line, without actually
	 * moving this line, use the <code>offset()</code> method in <code>Location</code>,
	 * which creates a new <code>Location</code>.
	 * @return the <code>Location</code> of the end of the line.
	 */
	public Location getEnd() {
		return end;
	}
	/**
	 * Move the start of the line to a new location
	 * @param point <code>Location</code> to which the start of the line should be moved
	 */
	public void setStart(Location point) {
		setStart(point.getDoubleX(), point.getDoubleY());
	}
	/**
	 * Move the start of the line to a new location
	 * @param x x coordinate to which the start of the line should be moved
	 * @param y y coordinate to which the start of the line should be moved
	 */
	public void setStart(double x, double y) {
		Location start = getStart();
		start.translate(x - start.getDoubleX(), y - start.getDoubleY());
	}
	/**
	 * Move the end of the line to a new location
	 * @param point <code>Location</code> to which the end of the line should be moved
	 */
	public void setEnd(Location point) {
		setEnd(point.getDoubleX(), point.getDoubleY());
	}
	/**
	 * Move the end of the line to a new location
	 * @param x x coordinate to which the end of the line should be moved
	 * @param y y coordinate to which the end of the line should be moved
	 */
	public void setEnd(double x, double y) {
		end.translate(x - end.getDoubleX(), y - end.getDoubleY());
	}
	/**
	 * Move both endpoints of the line.
	 * @param start <code>Location</code> to which the start of the line should be moved
	 * @param end <code>Location</code> to which the end of the line should be moved
	 */
	public void setEndPoints(Location start, Location end) {
		setStart(start);
		setEnd(end);
	}
	/**
	 * Move both endpoints of the line.
	 * @param x1 x1 coordinate to which the start of the line should be moved
	 * @param y1 y1 coordinate to which the start of the line should be moved
	 * @param x2 x coordinate to which the end of the line should be moved
	 * @param y2 y coordinate to which the end of the line should be moved
	 */
	public void setEndPoints(double x1, double y1, double x2, double y2) {
		setStart(x1, y1);
		setEnd(x2, y2);
	}
	
	/* * * * Methods from DrawableInterface * * * */
	
	/**
	 * Make this object invisible. It stays in position on its canvas; it is simply not drawn.
	 * This method has no effect if the object was already hidden.
	 */
	public void hide() { super.hide(); }
	/**
	 * Make an object visible again after it had been hidden. This method has no effect if the
	 * object was already visible.
	 */
	public void show() { super.show(); }
	/**
	 * Return <code>true</code> if this object has been rendered invisible with a call to its
	 * <code>hide()</code> method, <code>false</code> if it is still being drawn.
	 * @return whether this object is hidden
	 */
	public boolean isHidden() { return super.isHidden(); }

	/**
	 * Place this object on the specified canvas. The object will be removed from its current
	 * canvas, if it is on a canvas that is not the one it is being moved to. The object will
	 * be at the front of the canvas, as if it had just been drawn. If the given canvas is 
	 * <code>null</code> the object is simply removed from its current canvas.
	 * @param c <code>DrawingCanvas</code> to place this object on
	 */
	public void addToCanvas(DrawingCanvas c) { super.addToCanvas(c); }
	/**
	 * Remove this object from its canvas.
	 */
	public void removeFromCanvas() { super.removeFromCanvas(); }
	/**
	 * Return the canvas that this object is on. If it has been removed from its canvas, this
	 * method will return <code>null</code>.
	 * @return the <code>DrawingCanvas</code> that this object is on.
	 */
	public DrawingCanvas getCanvas() { return super.getCanvas(); }

	/**
	 * Move the reference point of this object to the given location. For a Drawable2D object,
	 * the reference point is the top left corner; for a line, it is the start point.
	 * @param point <code>Location</code> to move to
	 */
	public void moveTo(Location point) { super.moveTo(point); }
	/**
	 * Move the reference point of this object to the given location. For a Drawable2D object,
	 * the reference point is the top left corner; for a line, it is the start point.
	 * @param x x coordinate to move to
	 * @param y y coordinate to move to
	 */
	public void moveTo(double x, double y) { super.moveTo(x, y); }
	/**
	 * Shift (translate) this object left or right by an amount <code>dx</code>, and up or
	 * down by an amount <code>dy</code>. A negative <code>dx</code> indicates a shift to the
	 * left; a negative <code>dy</code> indicates a shift up
	 * @param dx amount of x translation
	 * @param dy amount of y translation
	 */
	public void move(double dx, double dy) {
		super.move(dx, dy);
		end.translate(dx, dy);
	}

	/**
	 * Return the color of this object
	 * @return the <code>Color</code> of this object
	 */
	public Color getColor() { return super.getColor(); }
	/**
	 * Change the color of this object.
	 * @param c new <code>Color</code> for this object
	 */
	public void setColor(Color c) { super.setColor(c); }

	/**
	 * Move this object one step forward in the draw order, causing it to be drawn on top of
	 * the objects that are behind it.
	 */
	public void sendForward() { super.sendForward(); }
	/**
	 * Move this object one step backward in the draw order, causing it to be drawn underneath
	 * the objects that are in front of it.
	 */
	public void sendBackward() { super.sendBackward(); }
	/**
	 * Move this object to the very front of the draw order, causing it to be drawn on top of
	 * all the other objects.
	 */
	public void sendToFront() { super.sendToFront(); }
	/**
	 * Move this object to the very back of the draw order, causing it to be drawn underneath
	 * all the other objects.
	 */
	public void sendToBack() { super.sendToBack(); }

	/**
	 * Return <code>true</code> if the given location is inside this object, <code>false</code>
	 * otherwise. Because <code>Line</code>s and <code>FramedArc</code>s have no "inside", a
	 * point is considered to be contained in one of them as long as it is within a few pixels
	 * of the line.
	 * @param point <code>Location</code> to test
	 * @return whether the given location is inside this shape
	 */
	public boolean contains(Location point) {
		double rx = point.getDoubleX() - getStart().getDoubleX(), ry = point.getDoubleY() - getStart().getDoubleY();
		double r2x = point.getDoubleX() - end.getDoubleX(), r2y = point.getDoubleY() - end.getDoubleY();
		if(Math.max(Math.abs(rx), Math.abs(ry)) < CLICK_SIZE && Math.sqrt(rx * rx + ry * ry) < CLICK_SIZE) return true;
		if(Math.max(Math.abs(r2x), Math.abs(r2y)) < CLICK_SIZE && Math.sqrt(r2x * r2x + r2y * r2y) < CLICK_SIZE) return true;
		
		double length = getStart().distanceTo(end);
		double dx = end.getDoubleX() - getStart().getDoubleX(), dy = end.getDoubleY() - getStart().getDoubleY();
		double parameter = (rx * dx + ry * dy) / length;
		double offset = Math.abs((rx * -dy + ry * dx) / length);
		return parameter >= 0 && parameter <= length && offset <= CLICK_SIZE;
	}

	/**
	 * Return a string describing how this object might be constructed.
	 */
	public String toString() {
		return toString(getStart() + ", " + getEnd());
	}
}
