package objectdraw;

import java.awt.Point;

/**
 * A <code>Location</code> object represents an (x, y) point on the screen. It is possible to
 * interconvert between the Java <code>Point</code> object and a <code>Location</code> object
 * using the methods and constructors in <code>Location</code>. You can retrieve coordinates 
 * from a <code>Location</code>, or translate it some distance across the screen. You can also
 * use the <code>distanceTo(Location other)</code> method to find the distance between two
 * <code>Location</code>s.
 * <br><br>
 * The <code>translate()</code> and <code>offset()</code> methods are a source of some confusion
 * and need some more explanation. The <code>offset()</code> method leaves this <code>Location</code>
 * unchanged, and returns a new <code>Location</code> some distance away. The <code>translate()</code>
 * method, on the other hand, actually moves this <code>Location</code> to a new place. If this
 * <code>Location</code> belongs to a particular object, this might result in that object moving.
 * 
 * @author Russell Zahniser (russell@zahniser.net)
 *
 */
public class Location extends ObjectDrawObject {
	private double x;
	private double y;
	
	/**
	 * Construct a <code>Location</code> representing the point (<code>x</code>, <code>y</code>).
	 * @param x x coordinate
	 * @param y y coordiante
	 */
	public Location(double x, double y) {
		super();
		this.x = x;
		this.y = y;
	}
	
	/**
	 * Construct a <code>Location</code> that is a duplicate of the given <code>point</code>.
	 * @param point <code>Location</code> to duplicate
	 */
	public Location(Location point) {
		this(point.x, point.y);
	}
	
	/**
	 * Construct a <code>Location</code> representing the same coordinates as the given
	 * Java <code>Point</code>.
	 * @param point <code>Point</code> to duplicate
	 */
	public Location(Point point) {
		this(point.getX(), point.getY());
	}
	
	/**
	 * Get the <code>x</code> coordinate of this <code>Location</code>, in integer precision.
	 * @return the <code>x</code> coordinate of this <code>Location</code>, in integer precision
	 */
	public int getX() {
		return (int)(x + .5);
	}
	
	/**
	 * Get the <code>y</code> coordinate of this <code>Location</code>, in integer precision.
	 * @return the <code>y</code> coordinate of this <code>Location</code>, in integer precision
	 */
	public int getY() {
		return (int)(y + .5);
	}
	
	/**
	 * Get the <code>x</code> coordinate of this <code>Location</code>, in <code>double</code> precision.
	 * @return the <code>x</code> coordinate of this <code>Location</code>, in <code>double</code> precision
	 */
	public double getDoubleX() {
		return x;
	}
	
	/**
	 * Get the <code>y</code> coordinate of this <code>Location</code>, in <code>double</code> precision.
	 * @return the <code>y</code> coordinate of this <code>Location</code>, in <code>double</code> precision
	 */
	public double getDoubleY() {
		return y;
	}
	
	/**
	 * Translate this <code>Location</code> by the given amount in the x and y. You can think
	 * of this as adding <code>dx</code> to its <code>x</code> and adding <code>dy</code> 
	 * to its <code>y</code>. If this <code>Location</code> was retrieved by calling a shape's
	 * <code>getLocation()</code> method, translating the <code>Location</code> will cause the
	 * shape to move.
	 * @param dx distance to move in the x direction. A positive <code>dx</code> moves the point
	 * to the right; negative moves it to the left.
	 * @param dy distance to move in the y direction. A positive <code>dy</code> moves the point
	 * down; negative moves it up.
	 */
	public void translate(double dx, double dy) {
		this.x += dx;
		this.y += dy;
		update();
	}
	
	/**
	 * Create a new <code>Location</code> that is offset from this one by the given amount.
	 * So, these two lines of code do the same thing:
	 * <br><br><code>
	 * Location loc1 = new Location(loc.getX() + 10, loc.getY() + 20);<br>
	 * Location loc1 = loc.offset(10, 20);
	 * </code><br><br>
	 * You can think of this as an alternative to <code>translate()</code> that creates
	 * a new <code>Location</code>, leaving the original unchanged.
	 * @param dx distance to move in the x direction. A positive <code>dx</code> places the
	 * new <code>Location</code> to the right of this one; a negative <code>dx</code> places it to the left.
	 * @param dy distance to move in the y direction. A positive <code>dy</code> places the
	 * new <code>Location</code> below this one; a negative <code>dy</code> places it above.
	 * @return A new <code>Location</code> offset from this one by the given amounts.
	 */
	public Location offset(double dx, double dy) {
		return new Location(this.x + dx, this.y + dy);
	}
	
	/**
	 * Find the distance from this point to another point. You could also think of this as the
	 * length of the line between the two points.
	 * @param point <code>Location</code> to find the distance to
	 * @return the distance between <code>this</code> and <code>point</code>
	 */
	public double distanceTo(Location point) {
		double dx = x - point.x, dy = y - point.y;
		return Math.sqrt(dx * dx + dy * dy);
	}
	
	/**
	 * Return a Java <code>Point</code> object representing the coordinates of this <code>Location</code>.
	 * @return a Java <code>Point</code> object representing the coordinates of this <code>Location</code>
	 */
	public Point toPoint() {
		return new Point(getX(), getY());
	}
	
	/**
	 * @deprecated for internal use only
	 */
	public int hashCode() {
		return getX() + (getY() << 16);
	}
	
	/**
	 * Return <code>true</code> if the other object given is a <code>Location</code> with the
	 * same coordinates as this one, <code>false</code> otherwise
	 * @return whether the other object given matches this <code>Location</code>
	 */
	public boolean equals(Object other) {
		if(!(other instanceof Location)) return false;
		return getX() == ((Location)other).getX() && getY() == ((Location)other).getY();
	}
	
	/**
	 * Return a string describing how this <code>Location</code> might be constructed.
	 * @return a string describing how this <code>Location</code> might be constructed
	 */
	public String toString() {
		return toString(Text.formatDecimal(x) + ", " + Text.formatDecimal(y));
	}
}
