package objectdraw;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.CharBuffer;
import java.nio.channels.ClosedChannelException;
import java.nio.channels.SelectableChannel;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.nio.channels.ServerSocketChannel;
import java.nio.channels.SocketChannel;
import java.nio.charset.Charset;
import java.nio.charset.CharsetDecoder;
import java.nio.charset.CharsetEncoder;
import java.nio.charset.CoderResult;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

/**
 * I should be able to create a Server with a given String name, get a list of 
 * active Servers on the network or ask if a Server of a given name exists, and
 * get a Client connected to a particular Server (specifying a String as the 
 * Client name). I should be able to send a String through that Client, and have 
 * a callback notified when that client receives a String. I should be able to 
 * get, from the Server, a list of connected Client names, and I should be able 
 * to send a message either to a particular Client or to broadcast that message
 * to all Clients. I should be able to register a callback to be notified when
 * the server receives a message, and the client name should be specified in the
 * message.
 */
abstract class NetworkThread implements Runnable {
	private static final int MAX_MESSAGE_LENGTH = 65536;
	private static final String LOST = new String("Lost");
	private static final int IDLE_SLEEP_TIME = 10;
	static final String CHARSET_NAME = "ISO-8859-1";
	private static final Charset CHARSET = Charset.forName(CHARSET_NAME);
	private static final boolean DEBUG = false;
	
	private final CharsetEncoder encoder = CHARSET.newEncoder();
	private final CharsetDecoder decoder = CHARSET.newDecoder();
	private final List channelChanges = Collections.synchronizedList(new LinkedList());
	private final List messagesToWrite = Collections.synchronizedList(new LinkedList());
	private final ArrayList channels = new ArrayList();
	private final LinkedList POOL = new LinkedList();
	private final CharBuffer readCharBuffer = CharBuffer.allocate(MAX_MESSAGE_LENGTH);
	final Selector selector;
	private final String name;
	private final Message nameMessage;
	
	private boolean running = false;
	private int nextID = 0;
	
	private Message encodeShortMessage(String string) {
		final CharBuffer input = CharBuffer.wrap(string);
		
		final Message message = createMessage();
		final ByteBuffer buffer = message.buffer;
		buffer.position(2);
		
		encoder.reset();
		encoder.encode(input, buffer, true);
		encoder.flush(buffer);
		
		final int length = buffer.position();
		buffer.flip();
		buffer.put((byte)(length >> 8));
		buffer.put((byte)(length & 0xFF));
		buffer.rewind();
		
		return message;
	}
	
	public String getName() {
		return name;
	}
	
	void start() {
	    new Thread(this).start();
	}
	
	NetworkThread(final String name) throws IOException {
		selector = Selector.open();
		
	    this.name = name;
	    nameMessage = encodeShortMessage(name);
	    nameMessage.retainCount = -1;
	}
	
	private void release(Message message) {
		if(--message.retainCount == 0) POOL.add(message);
	}
	
	private Message createMessage() {
		final Message message = (POOL.isEmpty() ? new Message() : (Message)POOL.removeFirst());
		message.buffer.clear();
		return message;
	}
	
	/**
	 * Decode a string of messages. Each message should be rewound, with its limit
	 * set to indicate how much is written in it - this state can be accomplished
	 * simply by flipping it after writing. The first buffer should be positioned
	 * *after* the two bytes indicating the message length. The messages will be
	 * released and added to the pool once they are decoded.
	 * @param messages Set of messages to decode
	 * @return String value of the decoded message
	 */
	private String decode(ArrayList messages) {
		readCharBuffer.clear();
		
		decoder.reset();
		// Skip over the length
		((Message)messages.get(0)).buffer.position(2);
		for(int i = 0; i < messages.size(); i++) {
			final ByteBuffer buffer = ((Message)messages.get(i)).buffer;
			if(i == messages.size() - 1) {
				decoder.decode(buffer, readCharBuffer, true);
			} else {
				decoder.decode(buffer, readCharBuffer, false);
			}
		}
		decoder.flush(readCharBuffer);
		
		readCharBuffer.flip();
		if(DEBUG) System.out.println("Decoded message of length " + readCharBuffer.limit());
		return readCharBuffer.toString();
	}
	
	/**
	 * Encodes a String into an array of messages. Each message is rewound, with its
	 * limit set to indicate the amount written in it. The retain count of each
	 * message is set as they are pulled out of the pool.
	 * @param string String to encode
	 * @return and array of messages
	 */
	private ArrayList encode(final String string, final int retainCount) {
		final CharBuffer input = CharBuffer.wrap(string);
		
		final ArrayList messages = new ArrayList();
		
		final Message firstMessage = createMessage();
		// The first Message has space 
		firstMessage.buffer.position(2);
		messages.add(firstMessage);
		firstMessage.retainCount = retainCount;
		
		Message message = firstMessage;
		
		// This will make length specify the total length, including the two bytes of length code
		// (I considered starting length at -2, but this is cleaner)
		int length = 0;
		encoder.reset();
		
		CoderResult result = encoder.encode(input, message.buffer, true);
		while(result.isOverflow()) {
			length += message.buffer.position();
			message.buffer.flip();
			messages.add(message = createMessage());
			message.retainCount = retainCount;
			result = encoder.encode(input, message.buffer, true);
		}
		
		result = encoder.flush(message.buffer);
		while(result.isOverflow()) {
			// Highly unlikely that this will happen, and I think impossible 
			// for it to happen several times, but why not be safe?
			length += message.buffer.position();
			message.buffer.flip();
			messages.add(message = createMessage());
			message.retainCount = retainCount;
			result = encoder.flush(message.buffer);
		}
		length += message.buffer.position();
		message.buffer.flip();
		
		// The first message will have been flipped; just put two bytes to specify
		// the total length written, and then rewind it
		firstMessage.buffer.put((byte)(length >> 8));
		firstMessage.buffer.put((byte)(length & 0xFF));
		firstMessage.buffer.rewind();
		if(DEBUG) System.out.println("Sending message of length " + length);
		
		return messages;
	}
	
	private Channel getChannel(final int id) {
		if(channels.isEmpty()) return null;
		
		int max = Math.min(id + 1, channels.size());
		int min = 0;
		
		while(min + 1 != max) {
			final int mid = (max + min) / 2;
			if(id >= ((Channel)channels.get(mid)).id) {
				min = mid;
			} else {
				max = mid;
			}
		}
		
		final Channel found = (Channel)channels.get(min);
		return (found.id == id ? found : null);
	}
	
	interface ChannelHandler {
		public void read();
		public void write();
	}
	
	// A Channel manages the traffic on a single channel. It is always interested
	// in reading from that channel, but only interested in writing if it has
	// messages enqueued to send.
	private class Channel implements ChannelHandler {
		private String name = null;
		private final int id;
		private final ArrayList readBuffers = new ArrayList();
		private final LinkedList writeQueue = new LinkedList();
		private final SocketChannel channel;
		private int remainingReadLength;
		private int writePosition;
		
		public Channel(final SocketChannel channel) {
			this.channel = channel;
			readBuffers.add(createMessage());
			remainingReadLength = -1;
			
			writeQueue.add(nameMessage);
			writePosition = -1;
			
			id = nextID++;
			channels.add(this);
		}
		
		private void setWriteInterest(boolean interested) {
			final SelectionKey key = channel.keyFor(selector);
			key.interestOps(interested ? 
					key.interestOps() | SelectionKey.OP_WRITE : 
					key.interestOps() & ~SelectionKey.OP_WRITE);
		}

		public boolean enqueue(ArrayList messages) {
			if(!channel.isOpen()) {
				close();
				return false;
			}
			
			if(writeQueue.isEmpty()) setWriteInterest(true);
			writeQueue.addAll(messages);
			
			return true;
		}
		
		// Read as much as can fit into a single buffer, and return a boolean indicating
		// whether the buffer was filled, and thus whether we should keep reading more.
		// A new buffer will have been put in place if the old one was filled.
		public boolean readOnePage() throws IOException {
			ByteBuffer buffer = ((Message)readBuffers.get(readBuffers.size() - 1)).buffer;
			final int readBytes = channel.read(buffer);
			if(readBytes == -1) {
				// Reached end-of-stream - the channel is dead
				close();
				return false;
			} else if(readBytes == 0) {
				// No more bytes available on the channel
				return false;
			}
			
			int position = buffer.position();
			final boolean moreToRead = (position == buffer.capacity());
			
			if(DEBUG) System.out.println("Read " + readBytes + " bytes, now at " + position + " / " + remainingReadLength);
			
			if(remainingReadLength == -1) {
				// Read length has not yet been set. If we haven't finished reading the
				// length code yet, then those bytes aren't on the channel yet, since 
				// clearly we didn't run out of buffer space. In that case, stay in
				// limbo and wait for more to be read. This will almost never happen.
				if(position < 2) return false;
				
				buffer.position(0);
				remainingReadLength = ((buffer.get() & 0xFF) << 8) | (buffer.get() & 0xFF);
				if(DEBUG) System.out.println("Receiving message of length " + remainingReadLength);
				buffer.position(position);
			}
			
			if(position == remainingReadLength) {
				// Finished reading exactly at the end of a packet
				buffer.flip();
				finishRead();
				readBuffers.add(createMessage());
				remainingReadLength = -1;
			} else if(position > remainingReadLength) {
				// Read more data after this packet
				while(position >= remainingReadLength) {
					// We read into the next packet - we will have to start a new message
					// after decoding the current one. Cut the extra bytes out of this 
					// message and patch them into the new one
					
					final Message nextMessage = createMessage();
					
					buffer.flip();
					if(position > remainingReadLength) {
						// Copy remainder into new buffer
						buffer.position(remainingReadLength);
						nextMessage.buffer.put(buffer);
						buffer.rewind();
						
						// Set this buffer to stop at packet boundary
						buffer.limit(remainingReadLength);
					}
					
					// Send the read buffers, and load up the new one
					finishRead();
					readBuffers.add(nextMessage);
					
					// Check the next buffer to see if it holds an entire packet
					buffer = nextMessage.buffer;
					position = buffer.position();
					if(position >= 2) {
						// Read enough to determine length of next packet
						buffer.position(0);
						remainingReadLength = ((buffer.get() & 0xFF) << 8) | (buffer.get() & 0xFF);
						if(DEBUG) System.out.println("Receiving another message of length " + remainingReadLength);
						buffer.position(position);
					} else {
						// Can't yet determine next packet length
						remainingReadLength = -1;
						break;
					}
				}
			} else if(moreToRead) {
				// Reached the end of a buffer
				buffer.flip();
				remainingReadLength -= position;
				
				readBuffers.add(createMessage());
			}
			return moreToRead;
		}
		
		private void finishRead() {
			// Messages read in are queued, rather than notifying the listeners
			// directly, for two reasons: First, listeners might manipulate the
			// state of the channel in a way that would confuse the ongoing read
			// operation, and second, we want to do all the reading and writing
			// as soon as possible after the selection, and listeners might do
			// computationally intensive tasks (though hopefully they spawn a
			// worker thread for that).
			if(name == null) {
				name = decode(readBuffers);
				channelReady(id, name);
			} else {
				messageReceived(id, name, decode(readBuffers));
			}
			POOL.addAll(readBuffers);
			readBuffers.clear();
		}

		public void read() {
			// This may be telling us to continue an existing read operation, or
			// to start a new one. In either case, we will read until there are no
			// more bytes available on the socket, swapping in new buffers when 
			// needed to hold all those bytes; the last buffer will be assumed
			// to have space left in it.
			
			try {
				while(readOnePage()) { }
			} catch (IOException e) {
				// Anything that could cause this would mean the stream is hopelessly munged, so just close it.
				close();
				e.printStackTrace();
				return;
			}
		}
		
		public void close() {
			channelClosed(id, name);
			
			channels.remove(this);
			
			// Release any messages we had scheduled to send
			while(!writeQueue.isEmpty()) release((Message)writeQueue.removeFirst());
			
			try {
				channel.close();
			} catch (IOException e) {
				// ignore
			}
		}

		public void write() {
			try {
				// Write as many buffers as we can before the channel blocks
				while(!writeQueue.isEmpty()) {
					final ByteBuffer buffer = ((Message)writeQueue.getFirst()).buffer;
					
					if(writePosition >= 0) {
						buffer.position(writePosition);
						writePosition = -1;
					} else {
						buffer.rewind();
					}
					
					try {
						channel.write(buffer);
					} catch(ClosedChannelException e) {
						close();
						return;
					}
					
					if(buffer.remaining() > 0) {
						// Writing has blocked, so we're done for now.
						writePosition = buffer.position();
						return;
					}
					release((Message)writeQueue.removeFirst());
				}
				// We exhausted the write queue. Set us not interested in writing.
				setWriteInterest(false);
			} catch (IOException e) {
				// Anything that could cause this would mean the stream is hopelessly munged, so just close it.
				close();
				e.printStackTrace();
				return;
			}
		}
	}
	
	abstract void messageReceived(int channel, String channelName, String message);
	abstract void channelClosed(int channel, String channelName);
	abstract void channelReady(int channel, String channelName);
	
	private class Message {
		final ByteBuffer buffer = ByteBuffer.allocateDirect(1024);
		int retainCount = 0;
	}
	
	private static class PendingMessage {
		public final String message;
		public final int channel;
		
		public PendingMessage(final String message, final int channel) {
			this.message = message;
			this.channel = channel;
		}
	}
	
	private class ChannelUpdate {
		private final SelectionKey key;
		private final int interestOps, changeMask;
		
		public ChannelUpdate(final SelectionKey key, final int interestOps, final int changeMask) {
			this.key = key;
			this.interestOps = interestOps;
			this.changeMask = changeMask;
		}
	}
	
	void updateChannelLater(final SelectableChannel channel, final int interestOps, final int changeMask) {
		final SelectionKey key = channel.keyFor(selector);
		channelChanges.add(new ChannelUpdate(key, interestOps, changeMask));
	}
	
	public void shutDown() {
		running = false;
	}
	
	public boolean isRunning() {
		return running;
	}
	
	void enqueueMessage(final String message, final int channelID) {
		messagesToWrite.add(new PendingMessage(message, channelID));
		selector.wakeup();
	}
	
	private void handlePendingMessages() {
		while(!channelChanges.isEmpty()) {
			final ChannelUpdate update = (ChannelUpdate)channelChanges.remove(0);
			update.key.interestOps(
					(update.key.interestOps() & ~update.changeMask) | 
					(update.interestOps & update.changeMask));
		}
		
		while(!messagesToWrite.isEmpty()) {
			final PendingMessage message = (PendingMessage)messagesToWrite.remove(0);
			if(message.channel < 0) {
				// Send message over all channels
				if(channels.isEmpty()) continue;
				
				final ArrayList messages = encode(message.message, channels.size());
				
				for(int i = 0; i < channels.size(); i++) {
					final Channel channel = (Channel)channels.get(i);
					
					// Enqueue returns false if the channel has been closed. This will
					// have removed the channel from the list, but we need to skip back
					// one as a result.
					if(!channel.enqueue(messages)) i--;
				}
			} else {
				// Send message just over one channel
				final Channel channel = getChannel(message.channel);
				
				if(channel == null) continue;
				
				channel.enqueue(encode(message.message, 1));
			}
		}
	}
	
	private void newChannel(final SocketChannel socketChannel) {
		try {
			socketChannel.configureBlocking(false);
			final Channel channel = new Channel(socketChannel);
			
			final SelectionKey key = socketChannel.keyFor(selector);
			if(key == null) {
				socketChannel.register(selector, SelectionKey.OP_READ | SelectionKey.OP_WRITE, channel);
			} else {
				key.attach(channel);
				key.interestOps(SelectionKey.OP_READ | SelectionKey.OP_WRITE);
			}
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	private void performSelection() {
		try {
			if(selector.select() == 0) {
				// If the server has nothing to do, sleep a bit; don't just keep the thread churning
				try {
					Thread.sleep(IDLE_SLEEP_TIME);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
				return;
			}
		} catch (IOException e) {
			e.printStackTrace();
			return;
		}
		
		for(final Iterator iter = selector.selectedKeys().iterator(); iter.hasNext(); ) {
			final SelectionKey key = (SelectionKey)iter.next();
			iter.remove();
			
			if(key.isAcceptable()) {
				try {
					newChannel(((ServerSocketChannel)key.channel()).accept());
				} catch (IOException e) {
					e.printStackTrace();
				}
			} else if(key.isConnectable()) {
				final SocketChannel channel = (SocketChannel)key.channel();
				try {
					if(!channel.finishConnect()) {
						throw new RuntimeException("Connection failed to finish.");
					}
					newChannel(channel);
				} catch (IOException e) {
					e.printStackTrace();
				}
			} else {
				final ChannelHandler channel = (ChannelHandler)key.attachment();
				
				if(channel == null) continue;
				
				if(key.isReadable()) channel.read();
				if(key.isValid() && key.isWritable()) channel.write();
			}
		}
	}
	
	void closeAll() {
		try {
			selector.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		while(!channels.isEmpty()) ((Channel)channels.get(channels.size() - 1)).close();
	}
	
	boolean setup() {
		return true;
	}
	
	void tearDown() {
	}

	/**
	 * @deprecated for internal use only
	 */
	public void run() {
		running = setup();
		while(running) {
			try {
				handlePendingMessages();
				performSelection();
			} catch(Throwable e) {
				e.printStackTrace();
			}
		}
		tearDown();
		closeAll();
	}
}
