package objectdraw;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.util.ArrayList;
import java.util.HashMap;
import java.awt.Shape;
import java.lang.reflect.Field;

/**
 * @deprecated for internal use only
 * ObjectDrawShape is the superclass of all shapes that appear on the screen, and implements
 * the DrawableInterface. It takes care of color, draw order, and drawing the shapes, so that
 * all that a subclass needs to do is to implement a method to return its shape. This class
 * also takes care of invalidating the stored shape when a change is made to the object and
 * requesting the generation of a new shape.
 * 
 * @author Russell Zahniser (russell@zahniser.net)
 *
 */
public abstract class ObjectDrawShape extends ObjectDrawObject implements DrawableInterface {
	private Color color;
	private boolean filled, hidden;
	private DrawingCanvas canvas;
	private Location location;
	private Shape shape;
	private boolean ready;
	
	static final Color DEFAULT_COLOR = Color.BLACK;
	
	// Map of canvas to ArrayList of objects
	private static HashMap canvases = new HashMap();
	
	// RI:
	// If canvas == null, then no canvas contains this object
	// If canvas != null, then this object is in the list only of that canvas, and appears only once in that list
	// color and location are not null
	// shape is null only if it has been invalidated and not yet regenerated
	// shape and location agree
	
	public ObjectDrawShape(Location l, boolean f, Color c, DrawingCanvas canvas) {
		filled = f;
		hidden = false;
		setMyLocation(l);
		color = (c == null ? DEFAULT_COLOR : c);
		shape = null;
		
		this.canvas = canvas;
		ready = false;
	}
	
	/**
	 * @deprecated for internal use only
	 */
	protected synchronized void ready() {
		if(ready) return;
		ready = true;
		
		if(canvas == null) return;
		
		ArrayList list = (ArrayList)canvases.get(canvas);
		if(list == null) {
			list = new ArrayList();
			canvases.put(canvas, list);
		}
		list.add(this);
		depend(canvas);

		update();
	}
	
	/**
	 * @deprecated for internal use only
	 */
	protected void setMyLocation(Location loc) {
		if(location != null) location.undepend(this);
		
		location = loc;
		
		if(location != null) location.depend(this);
	}
	
	/**
	 * @deprecated for internal use only
	 */
	public Shape getShape() {
		ObjectDrawObject.runUpdates();
		
		if(shape != null) return shape;
		
		shape = makeShape();
		
		return shape;
	}

	/**
	 * @deprecated for internal use only
	 */
	public void setMyShape(Shape shape) {
		this.shape = shape;
	}
	
	/**
	 * @deprecated for internal use only
	 */
	public abstract Shape makeShape();
	
	/**
	 * @deprecated for internal use only
	 */
	public void update() {
		shape = null;
		super.update();
	}
	
	/**
	 * @deprecated for internal use only
	 */
	protected Location getMyLocation() {
		return location;
	}
	
	/**
	 * @deprecated for internal use only
	 */
	public static DrawableIterator getDrawables(DrawingCanvas canvas) {
		return new DrawableIterator((ArrayList)canvases.get(canvas));
	}
	
	/**
	 * @deprecated for internal use only
	 */
	public static void draw(DrawingCanvas canvas, Graphics2D g) {
		g.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
		g.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
        g.setRenderingHint(RenderingHints.KEY_FRACTIONALMETRICS, RenderingHints.VALUE_FRACTIONALMETRICS_ON);
		g.setRenderingHint(RenderingHints.KEY_STROKE_CONTROL, RenderingHints.VALUE_STROKE_PURE);
		
		g.setColor(canvas.getBackground());
		g.fill(canvas.getBounds());
		
		for(DrawableIterator d = getDrawables(canvas); d.hasNext(); ) {
			DrawableInterface di = d.next();
			if(di != null && !di.isHidden()) {
				try {
					di.draw(g);
				} catch(Exception e) {
					e.printStackTrace();
				}
			}
		}
	}
	
	/**
	 * @deprecated for internal use only
	 */
	public void draw(Graphics2D g) {
		g.setColor(color);
		
		Shape s = getShape();
		if(s == null) return;
		
		if(filled) {
			g.fill(s);
		} else {
			if(this instanceof DrawableStrokeInterface) {
				g.setStroke(((DrawableStrokeInterface)this).getStroke());
			}
			g.draw(s);
		}
	}
	
	/**
	 * @deprecated for internal use only
	 */
	public static void clearCanvas(DrawingCanvas canvas) {
		for(DrawableIterator d = getDrawables(canvas); d.hasNext(); ) {
			ObjectDrawShape di = (ObjectDrawShape)d.next();
			di.setCanvas(null);
		}
	}

	/**
	 * @deprecated for internal use only
	 */
	public static void removeCanvas(DrawingCanvas canvas) {
		clearCanvas(canvas);
		canvases.remove(canvas);
	}
	
	/**
	 * @deprecated for internal use only
	 */
	private void setCanvas(DrawingCanvas c) {
		if(canvas != null) {
			undepend(canvas);
			ArrayList list = (ArrayList)canvases.get(canvas);
			if(list != null) list.remove(this);
			canvas.update();
		}
		if(c != null) {
			ArrayList list = (ArrayList)canvases.get(c);
			if(list == null) {
				list = new ArrayList();
				canvases.put(c, list);
			}
			list.add(this);
			depend(c);
		}
		canvas = c;
		update();
	}
	
	public void addToCanvas(DrawingCanvas c) {
		if(canvas != null) throw new IllegalStateException("An object that is already on a canvas cannot be added to another.");
		setCanvas(c);
	}
	
	public DrawingCanvas getCanvas() {
		return canvas;
	}
	
	public Color getColor() {
		return color;
	}
	
	public void hide() {
		hidden = true;
		update();
	}
	
	public boolean isHidden() {
		return hidden;
	}
	
	public void move(double dx, double dy) {
		location.translate(dx, dy);
	}
	
	public void moveTo(double x, double y) {
		move(x - location.getDoubleX(), y - location.getDoubleY());
	}
	
	public void moveTo(Location point) {
		moveTo(point.getDoubleX(), point.getDoubleY());
	}
	
	public void removeFromCanvas() {
		setCanvas(null);
	}
	
	private int getDrawOrder() {
		if(canvas == null) return -1;
		return ((ArrayList)canvases.get(canvas)).indexOf(this);
	}
	
	private void setDrawOrder(int pos) {
		if(canvas == null) return;
		
		ArrayList list = (ArrayList)canvases.get(canvas);
		
		if(pos == -2 || pos >= list.size()) pos = list.size() - 1;
		else if(pos < 0) pos = 0;
		
		list.remove(this);
		list.add(pos, this);
		
		update();
	}
	
	public void sendBackward() {
		setDrawOrder(getDrawOrder() - 1);
	}
	
	public void sendForward() {
		setDrawOrder(getDrawOrder() + 1);
	}
	
	public void sendToBack() {
		setDrawOrder(0);
	}
	
	public void sendToFront() {
		setDrawOrder(-2);
	}
	
	public void setColor(Color c) {
		color = c;
		update();
	}
	
	public void show() {
		hidden = false;
		update();
	}
	
	private String nameColor() {
		Color c = getColor();
		try {
			Field[] fields = Color.class.getDeclaredFields();
			for(int i = 0; i < fields.length; i++) {
				Field f = fields[i];
				char start = f.getName().charAt(0);
				if(start >= 'A' && start <= 'Z' && c.equals(f.get(null))) {
					return "Color." + f.getName();
				}
			}
		} catch (IllegalArgumentException e) {
		} catch (IllegalAccessException e) {
		}
		return "new Color(" + c.getRed() + ", " + c.getGreen() + ", " +c.getBlue() + 
				(c.getAlpha() == 255 ? "" : ", " + getColor().getAlpha()) + ")";
	}
	
	/**
	 * @deprecated for internal use only
	 */
	protected String toString(String middle) {
		return super.toString(middle + ", " + nameColor() + ", " + getCanvas());
	}
}
