package objectdraw;

import java.awt.Shape;
import java.awt.geom.PathIterator;
import java.util.ArrayList;
import java.util.HashMap;

class PathMeasurer {
	private static HashMap segments = new HashMap();
	private static final double ACCURACY = .1;
	
	static {
		segments.put(new Integer(PathIterator.SEG_CLOSE), new CloseSegment());
		segments.put(new Integer(PathIterator.SEG_MOVETO), new MoveSegment());
		segments.put(new Integer(PathIterator.SEG_LINETO), new LineSegment());
		segments.put(new Integer(PathIterator.SEG_QUADTO), new QuadraticSplineSegment());
		segments.put(new Integer(PathIterator.SEG_CUBICTO), new CubicSplineSegment());
	}
	
	private static Segment getSegment(int type) {
		return (Segment)segments.get(new Integer(type));
	}
	
	public static class Point {
		public final double x;
		public final double y;
		public final double nx;
		public final double ny;
		
		private Point(final double x, final double y, final double nx, final double ny) {
			this.x = x;
			this.y = y;
			this.nx = nx;
			this.ny = ny;
		}
	}
	
	static Point[] getPoints(final Shape shape, final double separation) {
		return new StrokeData(shape.getPathIterator(null)).measure(separation);
	}
	
	private static class StrokeData {
		private PathIterator path;
		// Location of most recent MOVETO
		private double startX, startY;
		// End of previous segment and thus start of next
		private double currentX, currentY;
		// How far into the next separation we have gone.
		private double currentOffset;
		// Data retrieved from path:
		private double[] coordinates = new double[6];
		// Used to store coefficients of parameterized curve:
		private double linearX, linearY;
		private double quadraticX, quadraticY;
		private double cubicX, cubicY;
		
		public StrokeData(PathIterator path) {
			this.path = path;
		}
		
		public Point[] measure(final double separation) {
			startX = startY = currentX = currentY = currentOffset = 0;
			
			final ArrayList points = new ArrayList();
			while(!path.isDone()) {
				Segment seg = getSegment(path.currentSegment(coordinates));
				if(seg != null) seg.measure(points, this, separation);
			
				path.next();
			}
			return (Point[])points.toArray(new Point[points.size()]);
		}
	}
	
	// Static singletons represent each of the stroke types, but they store their data within the
	// ShapeTrailStroke currently being evaluated, thus making them threadsafe.
	private static abstract class Segment {
		public abstract void measure(ArrayList points, StrokeData data, double separation);
	}
	
	private static class LineSegment extends Segment {
		private double length(StrokeData data) {
			// p(t) = p0 * (1 - t) + p1 * t
			// p(t) = p0 + (p1 - p0) * t
			data.linearX = getEndX(data) - data.currentX;
			data.linearY = getEndY(data) - data.currentY;
			
			return Math.sqrt(data.linearX * data.linearX + data.linearY * data.linearY);
		}

		protected double getEndX(StrokeData data) {
			return data.coordinates[0];
		}

		protected double getEndY(StrokeData data) {
			return data.coordinates[1];
		}
		
		public void measure(final ArrayList points, final StrokeData data, final double separation) {
			double length = length(data);
			if(length <= data.currentOffset) {
				data.currentX = getEndX(data);
				data.currentY = getEndY(data);
				data.currentOffset -= length;
				return;
			}
			final double dt = separation / length;
			double t = data.currentOffset / length;
			
			final double dx = data.linearX / length;
			final double dy = data.linearY / length;
			
			while(t <= 1) {
				points.add(new Point(data.currentX + t * data.linearX, data.currentY + t * data.linearY, dy, -dx));
				t += dt;
			}
			data.currentOffset = (t - 1) * length;
			data.currentX = getEndX(data);
			data.currentY = getEndY(data);
		}
	}
	
	private static class MoveSegment extends LineSegment {
		public void measure(final ArrayList points, final StrokeData data, final double separation) {
			data.currentX = data.startX = getEndX(data);
			data.currentY = data.startY = getEndY(data);
		}
	}
	
	private static class CloseSegment extends LineSegment {
		protected double getEndX(StrokeData data) {
			return data.startX;
		}

		protected double getEndY(StrokeData data) {
			return data.startY;
		}
	}
	
	private static abstract class SplineSegment extends Segment {
		protected abstract double getEndX(StrokeData data);
		protected abstract double getEndY(StrokeData data);
		protected abstract double getX(StrokeData data, double t);
		protected abstract double getY(StrokeData data, double t);
		protected abstract Point getPoint(StrokeData data, double t);
		protected abstract void prepare(StrokeData data);
		
		public void measure(final ArrayList points, final StrokeData data, final double sep) {
			// We want to split the path into segments of equal length. So:
			//  - We will keep performing binary subdivision of the section under consideration until
			//    its chord length is less than ACCURACY * separation
			//  - Once we get down to segments whose length is small enough:
			//       - If we determine that the next shape should be placed within it, we will find 
			//         that position by linear interpolation in t and place the shape
			//       - Otherwise, we simply subtract the chord length from the offset-to-next-spot
			prepare(data);
			double t = 0, x = data.currentX, y = data.currentY;
			final double error = sep * sep * ACCURACY * ACCURACY;
			
			final double[] stack = new double[300];
			
			int top = 0;
			stack[0] = 1;
			stack[1] = getEndX(data);
			stack[2] = getEndY(data);
			
			// Initially break the segment into 4 pieces. This is to avoid skipping the
			// whole segment if it happens to loop back on itself. We will assume that
			// looping a segment is not entirely uncommon, but looping a small bit of a
			// segment probably is.
			for(int i = 1; i < 4; i++) {
				top = i * 3;
				stack[top] = 1 - .25 * i;
				stack[top + 1] = getX(data, stack[top]);
				stack[top + 2] = getY(data, stack[top]);
			}
			
			while(top >= 0) {
				// Find the distance from the current point to the top one
				double dx = stack[top + 1] - x;
				double dy = stack[top + 2] - y;
				double length = dx * dx + dy * dy;
				
				if(length < error) {
					// Find the length of the current segment, then pop the top point
					length = Math.sqrt(length);
					if(length > data.currentOffset) {
						// Next placement is within this segment
						double tt = ((length - data.currentOffset) * t + data.currentOffset * stack[top]) / length;
						points.add(getPoint(data, tt));
						data.currentOffset += sep;
					}
					data.currentOffset -= length;
					
					t = stack[top];
					x = stack[top + 1];
					y = stack[top + 2];
					top -= 3;
				} else {
					// Find and push a point halfway from the current point to the top one
					double t2 = (stack[top] + t) / 2;
					top += 3;
					stack[top] = t2;
					stack[top + 1] = getX(data, t2);
					stack[top + 2] = getY(data, t2);
				}
			}
			data.currentX = getEndX(data);
			data.currentY = getEndY(data);
		}
	}
	
	private static class QuadraticSplineSegment extends SplineSegment {
		protected void prepare(StrokeData data) {
			// p(t) = (1 - t)((1 - t)p0 + (t)p1) + (t)((1 - t)p1 + (t)p2)
			// p(t) = p0 + (2 p1 - 2 p0)t + (p0 - 2 p1 + p2)t^2
			data.linearX = 2 * (data.coordinates[0] - data.currentX);
			data.linearY = 2 * (data.coordinates[1] - data.currentY);
			data.quadraticX = data.currentX - 2 * data.coordinates[0] + data.coordinates[2];
			data.quadraticY = data.currentY - 2 * data.coordinates[1] + data.coordinates[3];
		}

		protected double getEndX(StrokeData data) {
			return data.coordinates[2];
		}

		protected double getEndY(StrokeData data) {
			return data.coordinates[3];
		}

		protected double getX(StrokeData data, double t) {
			return data.currentX + t * (data.linearX + t * data.quadraticX);
		}

		protected double getY(StrokeData data, double t) {
			return data.currentY + t * (data.linearY + t * data.quadraticY);
		}

		protected Point getPoint(StrokeData data, double t) {
			double dx = data.linearX + 2 * t * data.quadraticX;
			double dy = data.linearY + 2 * t * data.quadraticY;
			double d = Math.sqrt(dx * dx + dy * dy);
			dx /= d;
			dy /= d;
			
			return new Point(
					data.currentX + t * (data.linearX + t * data.quadraticX),
					data.currentY + t * (data.linearY + t * data.quadraticY),
					dy, -dx);
		}
	}
	
	private static class CubicSplineSegment extends SplineSegment {
		protected void prepare(StrokeData data) {
			// p(t) = (1 - t)((1 - t)((1 - t)p0 + (t)p1) + (t)((1 - t)p1 + (t)p2) 
			//          + (t)((1 - t)((1 - t)p1 + (t)p2) + (t)((1 - t)p2 + (t)p3)
			// p(t) = p0 + (-3 p0 + 3 p1)t + (3 p0 - 6 p1 + 3 p2)t^2 + (-p0 + 3 p1 - 3 p2 + p3)t^3
			data.linearX = 3 * (data.coordinates[0] - data.currentX);
			data.linearY = 3 * (data.coordinates[1] - data.currentY);
			data.quadraticX = 3 * data.currentX - 6 * data.coordinates[0] + 3 * data.coordinates[2];
			data.quadraticY = 3 * data.currentY - 6 * data.coordinates[1] + 3 * data.coordinates[3];
			data.cubicX = -data.currentX + 3 * data.coordinates[0] - 3 * data.coordinates[2] + data.coordinates[4];
			data.cubicY = -data.currentY + 3 * data.coordinates[1] - 3 * data.coordinates[3] + data.coordinates[5];
		}

		protected double getEndX(StrokeData data) {
			return data.coordinates[4];
		}

		protected double getEndY(StrokeData data) {
			return data.coordinates[5];
		}

		protected double getX(StrokeData data, double t) {
			return data.currentX + t * (data.linearX + t * (data.quadraticX + t * data.cubicX));
		}

		protected double getY(StrokeData data, double t) {
			return data.currentY + t * (data.linearY + t * (data.quadraticY + t * data.cubicY));
		}

		protected Point getPoint(StrokeData data, double t) {
			double dx = data.linearX + t * (2 * data.quadraticX + 3 * t * data.cubicX);
			double dy = data.linearY + t * (2 * data.quadraticY + 3 * t * data.cubicY);
			double d = Math.sqrt(dx * dx + dy * dy);
			dx /= d;
			dy /= d;
			
			return new Point(
					data.currentX + t * (data.linearX + t * (data.quadraticX + t * data.cubicX)),
					data.currentY + t * (data.linearY + t * (data.quadraticY + t * data.cubicY)),
					dy, -dx);
		}
	}
}
