package objectdraw;

/**
 * The <code>ServerListener</code> interface specifies what methods an object needs in
 * order to be able to function as a server listener, the object managing a server.
 * The listener will be notified when the server starts and stops, when a channel is opened
 * to a client, and when a message is received from a client. Each channel is given a
 * unique ID number in case you want to send a message just to one client.
 * <br><br>
 * To specify what object will be your <code>ServerListener</code>, use the
 * <code>Network.configure()</code> method.
 * 
 * @see Network#configure(String, ServerListener, ClientListener)
 * @author Russell Zahniser (russell@zahniser.net)
 */
public interface ServerListener {
	/**
	 * If implementing your own server listener is too daunting a task, this one may
	 * be all you need. Every time it gets a message, it sends back that same message
	 * to <b>all</b> clients. This means that you can essentially ignore the server
	 * and assume that sending a message from any client will result in all the 
	 * client computers (including the sender) getting a <code>messageReceived()</code>
	 * event.
	 */
	public static final ServerListener ECHO = new ServerListener() {
		public void channelClosed(Server server, int channelID, String channelName) { }
		public void channelOpened(Server server, int channelID, String channelName) { }
		public void serverStarted(Server server) { }
		public void serverStopped(Server server) { }
		public void messageReceived(Server server, int channelID, String channelName, String message) {
			server.sendMessage(message);
		}
	};
	
	/**
	 * This method will be called when the server starts
	 * @param server The server
	 */
	public void serverStarted(Server server);
	/**
	 * This method will be called when a client logs in to the server. The ID numbers
	 * assigned for clients will always count up from zero, although occasionally one
	 * will be skipped if a client quit the program before fully logging in.
	 * @param server The server
	 * @param channelID Unique ID that will always be used for this client
	 * @param channelName Client's name
	 */
	public void channelOpened(Server server, int channelID, String channelName);
	/**
	 * This method will be called when a client sends a message to the server.
	 * @param server The server
	 * @param channelID Unique ID that will always be used for this client
	 * @param channelName Client's name
	 * @param message The message that was sent
	 */
	public void messageReceived(Server server, int channelID, String channelName, String message);
	/**
	 * This method will be called when a client disconnects from the server.
	 * @param server The server
	 * @param channelID Unique ID that will always be used for this client
	 * @param channelName Client's name
	 */
	public void channelClosed(Server server, int channelID, String channelName);
	/**
	 * This method will be called when the server shuts down.
	 * @param server The server
	 */
	public void serverStopped(Server server);
}
