package objectdraw;

import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.Shape;
import java.awt.geom.AffineTransform;
import java.awt.geom.GeneralPath;
import java.awt.geom.Rectangle2D;

import objectdraw.TextWrappingLayout.TextRun;

/**
 * A <code>Text</code> object displays some text on the screen. The text can have multiple lines
 * separated by "\n" line breaks, and can be aligned in various different ways. It is also 
 * possible to wrap text within another shape, or align it along a line. The shape that a
 * <code>Text</code> is aligned with or wrapped within is called its <b>basis</b>.
 * <br><br>
 * The color, font, size, and style of text can be altered. If you want to have text 
 * containing runs of several different styles, you can create many <code>Text</code> objects 
 * and chain them together by passing along the previous segment as the <code>basis</code>
 * when you create a new segment.
 * 
 * @author Russell Zahniser (russell@zahniser.net)
 *
 */
public class Text extends ObjectDrawShape implements Drawable2DInterface, TextWrappingLayout.TextRun {
	private DrawableInterface basis;
	private double alignH = LEFT, alignV = BOTTOM;
	private String text;
	private Font font = DEFAULT_FONT;
	private Text next;
	
	/**
	 * Alignment value to align the top of text with the basis.
	 */
	public static final int TOP = -1;
	/**
	 * Alignment value to align the bottom of text with the basis
	 */
	public static final int BOTTOM = 1;
	/**
	 * Alignment value to align the left edge of text with the basis.
	 */
	public static final int LEFT = -1;
	/**
	 * Alignment value to align the right edge of text with the basis.
	 */
	public static final int RIGHT = 1;
	/**
	 * Alignment value to center text with respect to the basis.
	 */
	public static final int CENTER = 0;
	
	/**
	 * <code>Font</code> used for newly created <code>Text</code> objects
	 */
	public static final Font DEFAULT_FONT = Font.decode("Times-PLAIN-12");
	
	/**
	 * Format a decimal value as a string, using a maximum of 4 decimal points.
	 * @param num number to format
	 * @return the <code>String</code> representation of that number
	 */
	public static String formatDecimal(double num) {
		return formatDecimal(num, 4);
	}
	
	/**
	 * Format a decimal value as a string, using a maximum of <code>places</code> decimal points.
	 * @param num number to format
	 * @param places maximum number of decimal places
	 * @return the <code>String</code> representation of that number
	 */
	public static String formatDecimal(double num, int places) {
		String sign = (num < 0 ? "-" : "");
		String tail = "";
		
		double min = 1;
		for(int i = 0; i < places; i++) min /= 10;
		
		if(num < 0) num = -num;
		
		if(num >= 1e9) {
			int exponent = 0;
			while(num >= 10) {
				num /= 10;
				exponent++;
			}
			// We don't need a decimal point now
			tail = "e" + exponent + tail;
		} else if(num <= 1e-10) {
			return "0";
		} else if(num <= min) {
			int exponent = 0;
			while(num < 1) {
				num *= 10;
				exponent++;
			}
			// We don't need a decimal point now
			tail = "e-" + exponent + tail;
		}
		int ipart = (int)num;
		num -= ipart;
		
		int place = 1;
		double adjust = .5;
		for(int i = 0; i < places; i++) {
			place *= 10;
			adjust /= 10;
		}
	
		int d = (int)((num + adjust) * place) % 10;
		while(d == 0 && place > 1) {
			place /= 10;
			adjust *= 10;
			d = (int)((num + adjust) * place) % 10;
		}
		if(place > 1) {
			while(place > 1) {
				tail = d + tail;
				place /= 10;
				d = (int)((num + adjust) * place) % 10;
			}
			return sign + ipart + "." + tail;
		} else {
			return sign + ipart + tail;
		}
	}
	private Text(Object text, double x, double y, DrawableInterface basis, Color color, DrawingCanvas canvas) {
		super(new Location(x, y), true, color, canvas);
		this.basis = basis;
		this.text = text + "";
		
		if(basis instanceof ObjectDrawShape) {
			((ObjectDrawObject)basis).depend(this);
			if(basis instanceof Text) {
				Text b = (Text)basis;
				if(b.next != null) {
					next = b.next;
					next.basis = this;
				}
				b.next = this; 
			}
		}
		update();
		ready();
	}
	
	/**
	 * Create a <code>Text</code> object at (<code>x</code>, <code>y</code>), displaying the
	 * given <code>text</code> in the given <code>color</code>. It will start out using the
	 * default font and default alignment (bottom left).
	 * @param text value whose <code>String</code> form will be displayed.
	 * @param x x coordinate of the <code>Text</code>'s position
	 * @param y y coordinate of the <code>Text</code>'s position
	 * @param color <code>Color</code> of the <code>Text</code>
	 * @param canvas <code>DrawingCanvas</code> to place this <code>text</code> on
	 */
	public Text(Object text, double x, double y, Color color, DrawingCanvas canvas) {
		this(text, x, y, null, color, canvas);
	}
	
	private static ObjectDrawShape checkBasis(DrawableInterface basis) {
		if(!(basis instanceof ObjectDrawShape)) {
			String name = basis.getClass().getName();
			name = name.substring(name.indexOf(".") + 1);
			throw new IllegalArgumentException("Text is not able to lay itself out in a " + name + ".");
		}
		return (ObjectDrawShape)basis;
	}
	/**
	 * Create a <code>Text</code> object attached to the given <code>basis</code>, displaying the
	 * given <code>text</code> in the given <code>color</code>. It will start out using the
	 * default font and default alignment (bottom left).
	 * @param text value whose <code>String</code> form will be displayed.
	 * @param basis shape that this <code>text</code> will be attached to
	 * @param color <code>Color</code> of the <code>Text</code>
	 * @param canvas <code>DrawingCanvas</code> to place this <code>text</code> on
	 */
	public Text(Object text, DrawableInterface basis, Color color, DrawingCanvas canvas) {
		this(text, checkBasis(basis).getMyLocation().getX(), checkBasis(basis).getMyLocation().getY(), basis, color, canvas);
	}
	/**
	 * Create a <code>Text</code> object at <code>origin</code>, displaying the
	 * given <code>text</code> in the given <code>color</code>. It will start out using the
	 * default font and default alignment (bottom left).
	 * @param text value whose <code>String</code> form will be displayed.
	 * @param origin <code>Location</code> at which this <code>Text</code> will be positioned
	 * @param color <code>Color</code> of the <code>Text</code>
	 * @param canvas <code>DrawingCanvas</code> to place this <code>text</code> on
	 */
	public Text(Object text, Location origin, Color color, DrawingCanvas canvas) {
		this(text, origin.getDoubleX(), origin.getDoubleY(), color, canvas);
	}
	/**
	 * Create a <code>Text</code> object at (<code>x</code>, <code>y</code>), displaying the
	 * given <code>text</code> in the given <code>color</code>. It will start out using the
	 * default font and default alignment (bottom left).
	 * @param text value whose <code>String</code> form will be displayed.
	 * @param x x coordinate of the <code>Text</code>'s position
	 * @param y y coordinate of the <code>Text</code>'s position
	 * @param color <code>Color</code> of the <code>Text</code>
	 * @param canvas <code>DrawingCanvas</code> to place this <code>text</code> on
	 */
	public Text(boolean text, double x, double y, Color color, DrawingCanvas canvas) {
		this("" + text, x, y, color, canvas);
	}
	/**
	 * Create a <code>Text</code> object attached to the given <code>basis</code>, displaying the
	 * given <code>text</code> in the given <code>color</code>. It will start out using the
	 * default font and default alignment (bottom left).
	 * @param text value whose <code>String</code> form will be displayed.
	 * @param basis shape that this <code>text</code> will be attached to
	 * @param color <code>Color</code> of the <code>Text</code>
	 * @param canvas <code>DrawingCanvas</code> to place this <code>text</code> on
	 */
	public Text(boolean text, DrawableInterface basis, Color color, DrawingCanvas canvas) {
		this("" + text, basis, color, canvas);
	}
	/**
	 * Create a <code>Text</code> object at <code>origin</code>, displaying the
	 * given <code>text</code> in the given <code>color</code>. It will start out using the
	 * default font and default alignment (bottom left).
	 * @param text value whose <code>String</code> form will be displayed.
	 * @param origin <code>Location</code> at which this <code>Text</code> will be positioned
	 * @param color <code>Color</code> of the <code>Text</code>
	 * @param canvas <code>DrawingCanvas</code> to place this <code>text</code> on
	 */
	public Text(boolean text, Location origin, Color color, DrawingCanvas canvas) {
		this("" + text, origin, color, canvas);
	}
	/**
	 * Create a <code>Text</code> object at (<code>x</code>, <code>y</code>), displaying the
	 * given <code>text</code> in the given <code>color</code>. It will start out using the
	 * default font and default alignment (bottom left).
	 * @param text value whose <code>String</code> form will be displayed.
	 * @param x x coordinate of the <code>Text</code>'s position
	 * @param y y coordinate of the <code>Text</code>'s position
	 * @param color <code>Color</code> of the <code>Text</code>
	 * @param canvas <code>DrawingCanvas</code> to place this <code>text</code> on
	 */
	public Text(char text, double x, double y, Color color, DrawingCanvas canvas) {
		this("" + text, x, y, color, canvas);
	}
	/**
	 * Create a <code>Text</code> object attached to the given <code>basis</code>, displaying the
	 * given <code>text</code> in the given <code>color</code>. It will start out using the
	 * default font and default alignment (bottom left).
	 * @param text value whose <code>String</code> form will be displayed.
	 * @param basis shape that this <code>text</code> will be attached to
	 * @param color <code>Color</code> of the <code>Text</code>
	 * @param canvas <code>DrawingCanvas</code> to place this <code>text</code> on
	 */
	public Text(char text, DrawableInterface basis, Color color, DrawingCanvas canvas) {
		this("" + text, basis, color, canvas);
	}
	/**
	 * Create a <code>Text</code> object at <code>origin</code>, displaying the
	 * given <code>text</code> in the given <code>color</code>. It will start out using the
	 * default font and default alignment (bottom left).
	 * @param text value whose <code>String</code> form will be displayed.
	 * @param origin <code>Location</code> at which this <code>Text</code> will be positioned
	 * @param color <code>Color</code> of the <code>Text</code>
	 * @param canvas <code>DrawingCanvas</code> to place this <code>text</code> on
	 */
	public Text(char text, Location origin, Color color, DrawingCanvas canvas) {
		this("" + text, origin, color, canvas);
	}
	/**
	 * Create a <code>Text</code> object at (<code>x</code>, <code>y</code>), displaying the
	 * given <code>text</code> in the given <code>color</code>. It will start out using the
	 * default font and default alignment (bottom left).
	 * @param text value whose <code>String</code> form will be displayed.
	 * @param x x coordinate of the <code>Text</code>'s position
	 * @param y y coordinate of the <code>Text</code>'s position
	 * @param color <code>Color</code> of the <code>Text</code>
	 * @param canvas <code>DrawingCanvas</code> to place this <code>text</code> on
	 */
	public Text(double text, double x, double y, Color color, DrawingCanvas canvas) {
		this(formatDecimal(text), x, y, color, canvas);
	}
	/**
	 * Create a <code>Text</code> object attached to the given <code>basis</code>, displaying the
	 * given <code>text</code> in the given <code>color</code>. It will start out using the
	 * default font and default alignment (bottom left).
	 * @param text value whose <code>String</code> form will be displayed.
	 * @param basis shape that this <code>text</code> will be attached to
	 * @param color <code>Color</code> of the <code>Text</code>
	 * @param canvas <code>DrawingCanvas</code> to place this <code>text</code> on
	 */
	public Text(double text, DrawableInterface basis, Color color, DrawingCanvas canvas) {
		this(formatDecimal(text), basis, color, canvas);
	}
	/**
	 * Create a <code>Text</code> object at <code>origin</code>, displaying the
	 * given <code>text</code> in the given <code>color</code>. It will start out using the
	 * default font and default alignment (bottom left).
	 * @param text value whose <code>String</code> form will be displayed.
	 * @param origin <code>Location</code> at which this <code>Text</code> will be positioned
	 * @param color <code>Color</code> of the <code>Text</code>
	 * @param canvas <code>DrawingCanvas</code> to place this <code>text</code> on
	 */
	public Text(double text, Location origin, Color color, DrawingCanvas canvas) {
		this(formatDecimal(text), origin, color, canvas);
	}
	/**
	 * Create a <code>Text</code> object at (<code>x</code>, <code>y</code>), displaying the
	 * given <code>text</code> in the given <code>color</code>. It will start out using the
	 * default font and default alignment (bottom left).
	 * @param text value whose <code>String</code> form will be displayed.
	 * @param x x coordinate of the <code>Text</code>'s position
	 * @param y y coordinate of the <code>Text</code>'s position
	 * @param color <code>Color</code> of the <code>Text</code>
	 * @param canvas <code>DrawingCanvas</code> to place this <code>text</code> on
	 */
	public Text(long text, double x, double y, Color color, DrawingCanvas canvas) {
		this("" + text, x, y, color, canvas);
	}
	/**
	 * Create a <code>Text</code> object attached to the given <code>basis</code>, displaying the
	 * given <code>text</code> in the given <code>color</code>. It will start out using the
	 * default font and default alignment (bottom left).
	 * @param text value whose <code>String</code> form will be displayed.
	 * @param basis shape that this <code>text</code> will be attached to
	 * @param color <code>Color</code> of the <code>Text</code>
	 * @param canvas <code>DrawingCanvas</code> to place this <code>text</code> on
	 */
	public Text(long text, DrawableInterface basis, Color color, DrawingCanvas canvas) {
		this("" + text, basis, color, canvas);
	}
	/**
	 * Create a <code>Text</code> object at <code>origin</code>, displaying the
	 * given <code>text</code> in the given <code>color</code>. It will start out using the
	 * default font and default alignment (bottom left).
	 * @param text value whose <code>String</code> form will be displayed.
	 * @param origin <code>Location</code> at which this <code>Text</code> will be positioned
	 * @param color <code>Color</code> of the <code>Text</code>
	 * @param canvas <code>DrawingCanvas</code> to place this <code>text</code> on
	 */
	public Text(long text, Location origin, Color color, DrawingCanvas canvas) {
		this("" + text, origin, color, canvas);
	}
	/**
	 * Create a <code>Text</code> object at (<code>x</code>, <code>y</code>), displaying the
	 * given <code>text</code>. It will start out drawn in black, using the
	 * default font and default alignment (bottom left).
	 * @param text value whose <code>String</code> form will be displayed.
	 * @param x x coordinate of the <code>Text</code>'s position
	 * @param y y coordinate of the <code>Text</code>'s position
	 * @param canvas <code>DrawingCanvas</code> to place this <code>text</code> on
	 */
	public Text(Object text, double x, double y, DrawingCanvas canvas) {
		this(text, x, y, null, canvas);
	}
	/**
	 * Create a <code>Text</code> object attached to the given <code>basis</code>, displaying the
	 * given <code>text</code>. It will start out drawn in black, using the
	 * default font and default alignment (bottom left).
	 * @param text value whose <code>String</code> form will be displayed.
	 * @param basis shape that this <code>text</code> will be attached to
	 * @param canvas <code>DrawingCanvas</code> to place this <code>text</code> on
	 */
	public Text(Object text, DrawableInterface basis, DrawingCanvas canvas) {
		this(text, basis, null, canvas);		
	}
	/**
	 * Create a <code>Text</code> object at <code>origin</code>, displaying the
	 * given <code>text</code>. It will start out drawn in black, using the
	 * default font and default alignment (bottom left).
	 * @param text value whose <code>String</code> form will be displayed.
	 * @param origin <code>Location</code> at which this <code>Text</code> will be positioned
	 * @param canvas <code>DrawingCanvas</code> to place this <code>text</code> on
	 */
	public Text(Object text, Location origin, DrawingCanvas canvas) {
		this(text, origin.getDoubleX(), origin.getDoubleY(), canvas);
	}
	/**
	 * Create a <code>Text</code> object at (<code>x</code>, <code>y</code>), displaying the
	 * given <code>text</code>. It will start out drawn in black, using the
	 * default font and default alignment (bottom left).
	 * @param text value whose <code>String</code> form will be displayed.
	 * @param x x coordinate of the <code>Text</code>'s position
	 * @param y y coordinate of the <code>Text</code>'s position
	 * @param canvas <code>DrawingCanvas</code> to place this <code>text</code> on
	 */
	public Text(boolean text, double x, double y, DrawingCanvas canvas) {
		this(text, x, y, null, canvas);
	}
	/**
	 * Create a <code>Text</code> object attached to the given <code>basis</code>, displaying the
	 * given <code>text</code>. It will start out drawn in black, using the
	 * default font and default alignment (bottom left).
	 * @param text value whose <code>String</code> form will be displayed.
	 * @param basis shape that this <code>text</code> will be attached to
	 * @param canvas <code>DrawingCanvas</code> to place this <code>text</code> on
	 */
	public Text(boolean text, DrawableInterface basis, DrawingCanvas canvas) {
		this(text, basis, null, canvas);		
	}
	/**
	 * Create a <code>Text</code> object at <code>origin</code>, displaying the
	 * given <code>text</code>. It will start out drawn in black, using the
	 * default font and default alignment (bottom left).
	 * @param text value whose <code>String</code> form will be displayed.
	 * @param origin <code>Location</code> at which this <code>Text</code> will be positioned
	 * @param canvas <code>DrawingCanvas</code> to place this <code>text</code> on
	 */
	public Text(boolean text, Location origin, DrawingCanvas canvas) {
		this(text, origin, null, canvas);
	}
	/**
	 * Create a <code>Text</code> object at (<code>x</code>, <code>y</code>), displaying the
	 * given <code>text</code>. It will start out drawn in black, using the
	 * default font and default alignment (bottom left).
	 * @param text value whose <code>String</code> form will be displayed.
	 * @param x x coordinate of the <code>Text</code>'s position
	 * @param y y coordinate of the <code>Text</code>'s position
	 * @param canvas <code>DrawingCanvas</code> to place this <code>text</code> on
	 */
	public Text(char text, double x, double y, DrawingCanvas canvas) {
		this(text, x, y, null, canvas);
	}
	/**
	 * Create a <code>Text</code> object attached to the given <code>basis</code>, displaying the
	 * given <code>text</code>. It will start out drawn in black, using the
	 * default font and default alignment (bottom left).
	 * @param text value whose <code>String</code> form will be displayed.
	 * @param basis shape that this <code>text</code> will be attached to
	 * @param canvas <code>DrawingCanvas</code> to place this <code>text</code> on
	 */
	public Text(char text, DrawableInterface basis, DrawingCanvas canvas) {
		this(text, basis, null, canvas);		
	}
	/**
	 * Create a <code>Text</code> object at <code>origin</code>, displaying the
	 * given <code>text</code>. It will start out drawn in black, using the
	 * default font and default alignment (bottom left).
	 * @param text value whose <code>String</code> form will be displayed.
	 * @param origin <code>Location</code> at which this <code>Text</code> will be positioned
	 * @param canvas <code>DrawingCanvas</code> to place this <code>text</code> on
	 */
	public Text(char text, Location origin, DrawingCanvas canvas) {
		this(text, origin, null, canvas);
	}
	/**
	 * Create a <code>Text</code> object at (<code>x</code>, <code>y</code>), displaying the
	 * given <code>text</code>. It will start out drawn in black, using the
	 * default font and default alignment (bottom left).
	 * @param text value whose <code>String</code> form will be displayed.
	 * @param x x coordinate of the <code>Text</code>'s position
	 * @param y y coordinate of the <code>Text</code>'s position
	 * @param canvas <code>DrawingCanvas</code> to place this <code>text</code> on
	 */
	public Text(double text, double x, double y, DrawingCanvas canvas) {
		this(text, x, y, null, canvas);
	}
	/**
	 * Create a <code>Text</code> object attached to the given <code>basis</code>, displaying the
	 * given <code>text</code>. It will start out drawn in black, using the
	 * default font and default alignment (bottom left).
	 * @param text value whose <code>String</code> form will be displayed.
	 * @param basis shape that this <code>text</code> will be attached to
	 * @param canvas <code>DrawingCanvas</code> to place this <code>text</code> on
	 */
	public Text(double text, DrawableInterface basis, DrawingCanvas canvas) {
		this(text, basis, null, canvas);		
	}
	/**
	 * Create a <code>Text</code> object at <code>origin</code>, displaying the
	 * given <code>text</code>. It will start out drawn in black, using the
	 * default font and default alignment (bottom left).
	 * @param text value whose <code>String</code> form will be displayed.
	 * @param origin <code>Location</code> at which this <code>Text</code> will be positioned
	 * @param canvas <code>DrawingCanvas</code> to place this <code>text</code> on
	 */
	public Text(double text, Location origin, DrawingCanvas canvas) {
		this(text, origin, null, canvas);
	}
	/**
	 * Create a <code>Text</code> object at (<code>x</code>, <code>y</code>), displaying the
	 * given <code>text</code>. It will start out drawn in black, using the
	 * default font and default alignment (bottom left).
	 * @param text value whose <code>String</code> form will be displayed.
	 * @param x x coordinate of the <code>Text</code>'s position
	 * @param y y coordinate of the <code>Text</code>'s position
	 * @param canvas <code>DrawingCanvas</code> to place this <code>text</code> on
	 */
	public Text(long text, double x, double y, DrawingCanvas canvas) {
		this(text, x, y, null, canvas);
	}
	/**
	 * Create a <code>Text</code> object attached to the given <code>basis</code>, displaying the
	 * given <code>text</code>. It will start out drawn in black, using the
	 * default font and default alignment (bottom left).
	 * @param text value whose <code>String</code> form will be displayed.
	 * @param basis shape that this <code>text</code> will be attached to
	 * @param canvas <code>DrawingCanvas</code> to place this <code>text</code> on
	 */
	public Text(long text, DrawableInterface basis, DrawingCanvas canvas) {
		this(text, basis, null, canvas);		
	}
	/**
	 * Create a <code>Text</code> object at <code>origin</code>, displaying the
	 * given <code>text</code>. It will start out drawn in black, using the
	 * default font and default alignment (bottom left).
	 * @param text value whose <code>String</code> form will be displayed.
	 * @param origin <code>Location</code> at which this <code>Text</code> will be positioned
	 * @param canvas <code>DrawingCanvas</code> to place this <code>text</code> on
	 */
	public Text(long text, Location origin, DrawingCanvas canvas) {
		this(text, origin, null, canvas);
	}
	
	private Shape setEmptyShapes() {
		Shape shape = new GeneralPath();
		setMyShape(shape);
		if(next != null) next.setEmptyShapes();
		return shape;
	}
	
	/**
	 * @deprecated for internal use only
	 */
	public Shape makeShape() {
		if(getCanvas() == null) return null;
		
		Graphics2D g2 = (Graphics2D)((JDrawingCanvas)getCanvas()).getGraphics();
		g2.setRenderingHint(RenderingHints.KEY_FRACTIONALMETRICS, RenderingHints.VALUE_FRACTIONALMETRICS_ON);
		
		if(basis instanceof Text) {
			((Text)basis).makeShape();
		} else if(basis == null || basis instanceof Line) {
			new TextWrappingLayout(this, alignH, alignV, null, g2.getFontRenderContext()).doLayout();
		} else {
			Shape basisShape = basis.getShape();
			new TextWrappingLayout(this, alignH, alignV, basisShape, g2.getFontRenderContext()).doLayout();
		}
		return getShape();
	}
	
	/**
	 * Set the horizontal and vertical alignment of this text. The possible alignment values
	 * are the constants <code>Text.TOP</code>, <code>Text.BOTTOM</code>, and so on.
	 * @param horizontalAlign horizontal alignment value, ranging from -1 (align the <b>left</b>
	 * of the text with the basis) to 1 (align the <b>right</b>) of the text with the basis.
	 * @param verticalAlign vertical alignment value, ranging from -1 (align the <b>top</b>
	 * of the text with the basis) to 1 (align the <b>bottom</b>) of the text with the basis.
	 */
	public void setAlignment(double horizontalAlign, double verticalAlign) {
		if(basis instanceof Text) {
			((Text)basis).setAlignment(horizontalAlign, verticalAlign);
		} else {
			alignH = horizontalAlign;
			alignV = verticalAlign;
			update();
		}
	}
	
	/**
	 * Retrieve the horizontal alignment value
	 * @return the horizontal alignment value, ranging from -1 (align the <b>left</b>
	 * of the text with the basis) to 1 (align the <b>right</b>) of the text with the basis.
	 */
	public double getHorizontalAlignment() {
		if(basis instanceof Text) {
			return ((Text)basis).getHorizontalAlignment();
		} else {
			return alignH;
		}
	}
	
	/**
	 * Retrieve the vertical alignment value
	 * @return the vertical alignment value, ranging from -1 (align the <b>top</b>
	 * of the text with the basis) to 1 (align the <b>bottom</b>) of the text with the basis.
	 */
	public double getVerticalAlignment() {
		if(basis instanceof Text) {
			return ((Text)basis).getVerticalAlignment();
		} else {
			return alignV;
		}
	}
	
	/**
	 * Return the <code>Font</code> object representing the font, font size, and style of this
	 * <code>Text</code>.
	 * @return the font of this text
	 */
	public Font getFont() {
		return font;
	}
	
	/**
	 * Set the <code>Font</code> of this <code>Text</code>. The <code>Font</code> object
	 * contains information on the font used, and also the size and style of the text.
	 * @param f new <code>Font</code> to use
	 */
	public void setFont(Font f) {
		font = f;
		update();
	}
	
	/**
	 * Return the text string displayed by this <code>Text</code>.
	 * @return the text string displayed by this <code>Text</code>
	 */
	public String getText() {
		return text;
	}
	
	/**
	 * Change the text string displayed by this <code>Text</code>.
	 * @param t new text to display
	 */
	public void setText(String t) {
		text = (t == null ? "" : t);
		update();
	}
	
	/**
	 * Change the font name being used by this <code>Text</code>. The new font name should
	 * be a common font, like "Times" or "Courier"; not all computers have all fonts. If that
	 * font does not exist on this computer, another font will be substituted for it.
	 * @param fname name of new font to use
	 */
	public void setFont(String fname) {
		setFont(new Font(fname, font.getStyle(), font.getSize()));
	}
	
	/**
	 * Set the style of this text to plain (not bold or italic)
	 */
	public void setPlain() {
		setFont(font.deriveFont(Font.PLAIN));
	}
	
	/**
	 * Set this font to be bold or not, depending on the value of <code>b</code>.
	 * @param b whether to become bold (<code>true</code>) or not bold (<code>false</code>)
	 */
	public void setBold(boolean b) {
		setFont(font.deriveFont(b ? font.getStyle() | Font.BOLD : font.getStyle() & ~Font.BOLD));
	}
	
	/**
	 * Set this font to be italic or not, depending on the value of <code>i</code>.
	 * @param i whether to become bold (<code>true</code>) or not bold (<code>false</code>)
	 */
	public void setItalic(boolean i) {
		setFont(font.deriveFont(i ? font.getStyle() | Font.ITALIC : font.getStyle() & ~Font.ITALIC));
	}
	
	/**
	 * Change the font size of this text.
	 * @param size new font size
	 */
	public void setFontSize(int size) {
		setFont(font.deriveFont((float)size));
	}
	
	/**
	 * Set this text to be bold
	 */
	public void setBold() {
		setBold(true);
	}
	
	/**
	 * Set this text to be italic
	 */
	public void setItalic() {
		setItalic(true);
	}
	
	/**
	 * @deprecated for internal use only
	 */
	public Rectangle2D getBounds() {
		return getShape().getBounds();
	}
	
	/**
	 * Get the next <code>Text</code> object in a chain, or <code>null</code> if there is no
	 * more text after this one.
	 * @return the next <code>Text</code> in the chain
	 */
	public Text getNext() {
		return next;
	}

	/**
	 * @deprecated for internal use only
	 */
	public TextRun getNextRun() {
		return next;
	}
	
	private Line getBasisLine() {
		if(basis instanceof Line) {
			return (Line)basis;
		} else if(basis instanceof Text) {
			return ((Text)basis).getBasisLine();
		} else {
			return null;
		}
	}
	
	private Location getBasisLocation() {
		if(basis == null) {
			return getMyLocation();
		} else if(basis instanceof Text) {
			return ((Text)basis).getBasisLocation();
		} else {
			return null;
		}
	}

	/**
	 * @deprecated for internal use only
	 */
	public void setShape(Shape shape) {
		Line line = getBasisLine();
		Location loc;
		if(line != null) {
			// transform shape to lie along line
			double x = line.getStart().getDoubleX(), y = line.getStart().getDoubleY();
			double dx = line.getEnd().getDoubleX() - x, dy = line.getEnd().getDoubleY() - y;
			double length = line.getStart().distanceTo(line.getEnd());
			
			AffineTransform transform = AffineTransform.getTranslateInstance(x, y);
			transform.rotate(Math.atan2(dy, dx));
			transform.translate(length * (getHorizontalAlignment() + 1) / 2, 0);
			
			shape = transform.createTransformedShape(shape);
		} else if((loc = getBasisLocation()) != null) {
			AffineTransform transform = AffineTransform.getTranslateInstance(loc.getDoubleX(), loc.getDoubleY());
			shape = transform.createTransformedShape(shape);
		}
		setMyShape(shape);
	}
	
	private DrawingCanvas getBasisCanvas() {
		if(basis == null) {
			return getCanvas();
		} else if(basis instanceof Text) {
			return ((Text)basis).getBasisCanvas();
		} else {
			return basis.getCanvas();
		}
	}
	
	// If basis is removed, remove me as well.
	/**
	 * @deprecated for internal use only
	 */
	public void update() {
		DrawingCanvas c = getBasisCanvas();
		if(c != getCanvas()) {
			if(c == null) {
				removeFromCanvas();
			} else {
				addToCanvas(c);
			}
		}
		if(next != null) updateForward();
		updateBackward();
	}
	
	private void updateForward() {
		if(next != null) {
			next.updateForward();
		}
		super.update();
	}
	
	private void updateBackward() {
		if(basis instanceof Text) {
			((Text)basis).updateBackward();
		}
		super.update();
	}

	/**
	 * Get the previous <code>Text</code> object in a chain, or <code>null</code> if there is no
	 * more text before this one.
	 * @return the previous <code>Text</code> in the chain
	 */
	public Text getPrevious() {
		if(basis instanceof Text) {
			return (Text)basis;
		} else {
			return null;
		}
	}
	
	/**
	 * Get the shape that is the basis of this text, or <code>null</code> if this text is simply
	 * placed at a point, with no basis.
	 * @return the basis object for this text
	 */
	public DrawableInterface getBasis() {
		if(basis instanceof Text) {
			return ((Text)basis).getBasis();
		} else {
			return basis;
		}
	}
	
	private String escapeText() {
		String s = "";
		for(int i = 0; i < text.length(); i++) {
			char c = text.charAt(i);
			if(c == '\\') {
				s += "\\\\";
			} else if(c == '"') {
				s += "\\\"";
			} else if(c == '\n') {
				s += "\\n";
			} else if(c == '\t') {
				s += "\\t";
			} else if(c == '\r') {
				s += "\\r";
			} else {
				s += c;
			}
		}
		return "\"" + s + "\"";
	}
	
	/* * * * Methods from LocatableInterface * * * */
	
	/**
	 * Get this object's x coordinate.
	 * @return this object's x coordinate
	 */
	public int getX() {
		return getLocation().getX();
	}
	/**
	 * Get this object's y coordinate.
	 * @return this object's y coordinate
	 */
	public int getY() {
		return getLocation().getY();
	}
	/**
	 * Get this object's x coordinate, in double precision.
	 * @return this object's x coordinate, in double precision
	 */
	public double getDoubleX() {
		return getLocation().getDoubleX();
	}
	/**
	 * Get this object's y coordinate, in double precision.
	 * @return this object's y coordinate, in double precision
	 */
	public double getDoubleY() {
		return getLocation().getDoubleY();
	}
	/**
	 * Get this object's location. Translating this <code>Location</code> object
	 * will have the same effect as moving the object. So, the following two lines
	 * of code are equivalent:
	 * <br><br><code>
	 * object.move(10, 20);<br>
	 * object.getLocation().translate(10, 20);
	 * </code><br><br>If you want to create a <code>Location</code> offset by some
	 * amount from the <code>Location</code> of this object, without actually
	 * moving this object, use the <code>offset()</code> method in <code>Location</code>,
	 * which creates a new <code>Location</code>.
	 * @return the <code>Location</code> of this object.
	 */
	public Location getLocation() {
		if(basis == null) {
			return getMyLocation();
		} else if(basis instanceof Text) {
			return ((Text)basis).getLocation();
		} else {
			return ((ObjectDrawShape)basis).getMyLocation();
		}
	}
	
	/* * * * Methods from Drawable2DInterface * * * */

	/**
	 * Return the height of the rectangle bounding this object.
	 * @return the height of the rectangle bounding this object.
	 */
	public int getWidth() {
		return (int)(getBounds().getWidth() + .5);
	}
	/**
	 * Return the width of the rectangle bounding this object.
	 * @return the width of the rectangle bounding this object.
	 */
	public int getHeight() {
		return (int)(getBounds().getHeight() + .5);
	}
	/**
	 * Return the height of the rectangle bounding this object, in double precision.
	 * @return the height of the rectangle bounding this object, in double precision.
	 */
	public double getDoubleWidth() {
		return getBounds().getWidth();
	}
	/**
	 * Return the height of the rectangle bounding this object, in double precision.
	 * @return the height of the rectangle bounding this object, in double precision.
	 */
	public double getDoubleHeight() {
		return getBounds().getHeight();
	}
	/**
	 * Return <code>true</code> if the rectangle bounding this object overlaps with the
	 * rectangle bounding another object; <code> false</code> otherwise.
	 * @param other another two dimensional object to check for overlapping with
	 * @return whether the bounding rectangles of the two objects overlap
	 */
	public boolean overlaps(Drawable2DInterface other) {
		return getBounds().intersects(other.getBounds());
	}

	/* * * * Methods from DrawableInterface * * * */
	
	/**
	 * Make this object invisible. It stays in position on its canvas; it is simply not drawn.
	 * This method has no effect if the object was already hidden.
	 */
	public void hide() { super.hide(); }
	/**
	 * Make an object visible again after it had been hidden. This method has no effect if the
	 * object was already visible.
	 */
	public void show() { super.show(); }
	/**
	 * Return <code>true</code> if this object has been rendered invisible with a call to its
	 * <code>hide()</code> method, <code>false</code> if it is still being drawn.
	 * @return whether this object is hidden
	 */
	public boolean isHidden() { return super.isHidden(); }

	/**
	 * Place this object on the specified canvas. The object will be removed from its current
	 * canvas, if it is on a canvas that is not the one it is being moved to. The object will
	 * be at the front of the canvas, as if it had just been drawn. If the given canvas is 
	 * <code>null</code> the object is simply removed from its current canvas.
	 * @param c <code>DrawingCanvas</code> to place this object on
	 */
	public void addToCanvas(DrawingCanvas c) { super.addToCanvas(c); }
	/**
	 * Remove this object from its canvas.
	 */
	public void removeFromCanvas() {
		// Bow out of the line. Be sure that your alignment is passed on
		// to your next, if you have any.
		if(basis instanceof Text) {
			((Text)basis).next = next;
		}
		if(next != null) {
			next.alignH = alignH;
			next.alignV = alignV;
			next.basis = basis;
			next.update();
		} else if(basis instanceof Text) {
			((Text)basis).update();
		}
		basis = null;
		next = null;
		super.removeFromCanvas();
			
	}
	
	/**
	 * Return the canvas that this object is on. If it has been removed from its canvas, this
	 * method will return <code>null</code>.
	 * @return the <code>DrawingCanvas</code> that this object is on.
	 */
	public DrawingCanvas getCanvas() { return super.getCanvas(); }

	/**
	 * Move the reference point of this object to the given location. For a Drawable2D object,
	 * the reference point is the top left corner; for a line, it is the start point.
	 * @param point <code>Location</code> to move to
	 */
	public void moveTo(Location point) { super.moveTo(point); }
	/**
	 * Move the reference point of this object to the given location. For a Drawable2D object,
	 * the reference point is the top left corner; for a line, it is the start point.
	 * @param x x coordinate to move to
	 * @param y y coordinate to move to
	 */
	public void moveTo(double x, double y) { super.moveTo(x, y); }
	/**
	 * Shift (translate) this object left or right by an amount <code>dx</code>, and up or
	 * down by an amount <code>dy</code>. A negative <code>dx</code> indicates a shift to the
	 * left; a negative <code>dy</code> indicates a shift up
	 * @param dx amount of x translation
	 * @param dy amount of y translation
	 */
	public void move(double dx, double dy) {
		if(basis == null) {
			getMyLocation().translate(dx, dy);
		} else {
			((ObjectDrawShape)basis).move(dx, dy);
		}
	}
	
	/**
	 * Return the color of this object
	 * @return the <code>Color</code> of this object
	 */
	public Color getColor() { return super.getColor(); }
	/**
	 * Change the color of this object.
	 * @param c new <code>Color</code> for this object
	 */
	public void setColor(Color c) { super.setColor(c); }

	/**
	 * Move this object one step forward in the draw order, causing it to be drawn on top of
	 * the objects that are behind it.
	 */
	public void sendForward() { super.sendForward(); }
	/**
	 * Move this object one step backward in the draw order, causing it to be drawn underneath
	 * the objects that are in front of it.
	 */
	public void sendBackward() { super.sendBackward(); }
	/**
	 * Move this object to the very front of the draw order, causing it to be drawn on top of
	 * all the other objects.
	 */
	public void sendToFront() { super.sendToFront(); }
	/**
	 * Move this object to the very back of the draw order, causing it to be drawn underneath
	 * all the other objects.
	 */
	public void sendToBack() { super.sendToBack(); }

	/**
	 * Return <code>true</code> if the given location is inside this object, <code>false</code>
	 * otherwise. Because <code>Line</code>s and <code>FramedArc</code>s have no "inside", a
	 * point is considered to be contained in one of them as long as it is within a few pixels
	 * of the line.
	 * @param point <code>Location</code> to test
	 * @return whether the given location is inside this shape
	 */
	public boolean contains(Location point) {
		if(getShape() == null) return false;
		return getShape().getBounds2D().contains(point.getX(), point.getY());
	}
	
	/**
	 * Return a string describing how this object might be constructed.
	 */
	public String toString() {
		return toString(escapeText() + ", " + basis);
	}
}
